!
! Module:dss_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains auxiliary routines to do with the data subspace (DSS).
! 
! Comments: TODO:
!	- change status values to error calls.
! 
! $Id: dss_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_aux

  use dal
  use errorHandling
  use dss
  use caltypes	! for IntervalT

  use array_utils	! for smallestTrueIndex, largestTrueIndex
			! and trapezoidSum.
  use constants		! for lnTwo.
  use dal_aux
  use dss_ranges_aux	! for valueIsWithinRange
  use intervals_aux	! for andIntervals
  use sort_mod		! for quickSortStatic.
  use ssc_misc

  implicit none 

  private 

  public hasDss
  public dumpDss

  public dumpDssComponent
  public getNum1stDssCompWithValue
  private dssCompObeys
  public numDssCompsOfAxis
  public numDssCompsWithValue
  public valueIsPassedByDssComp
  public dssCompsIdentical
public dssCompsIdenticalList
  public dssCompNumsWhichPassThisValue

!!!  public determineUniqueFilterCombos

  public filterTypeToStr
  private filterOverlaps
  public anyFilterOfThisAxis
  public numFiltersOfAxis
  public get1stFilterThisAxis
  public allFiltersPassValue
  public valueIsPassedByFilter
  public filtersAreIdentical

  public getRangesSpan
  public readRanges
  private addNewCompSubset
  private addNewCompNumToSubset
  private copyCompSubsets
  private copyCompSubset
  private compSubsetsIdentical

!  public readGtis
  public readGtisFromDss
  public getGtiFromFilter
!  public andGtis

  public decomposeRegionXyNameString

  interface hasDss
    module procedure hasDssSet, hasDssArray, hasDssTab, hasDssBlock
  end interface

  interface readRanges
    module procedure readRangesDss
    module procedure readRangesComp
  end interface

  interface dssCompsIdentical
    module procedure dssCompsIdenticalInt
    module procedure dssCompsIdenticalComp
!    module procedure dssCompsIdenticalList, dssCompsIdenticalSimple
    module procedure dssCompsIdenticalSimple
    module procedure dssCompsIdentNumVecNameScal
    module procedure dssCompsIdentNumVecNameVec
  end interface

  interface dumpDss
    module procedure dumpDssBlock
    module procedure dumpDssPointer
  end interface

  interface readGtisFromDss
    module procedure readGtisFromDssPointer, readGtisFromDssBlock
  end interface


  type, public :: dssConstraintType
    character(20)         :: name
    logical(bool)         :: ignore
    type(RangeT), pointer :: ranges(:)
  end type dssConstraintType

  type, public :: dssCompSubsetType
    type(RangeT), pointer :: ranges(:)
    integer, pointer :: compNums(:)
  end type dssCompSubsetType

contains 

  !---------------------------------------------------------------------
  logical(bool) function hasDssSet(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    !-------------------------------------------------------------------
    ! Start

    hasDssSet = .false. ! default
    if (hasAttribute(set, 'DSTYP1')) hasDssSet = .true.

  end function hasDssSet

  !---------------------------------------------------------------------
  logical(bool) function hasDssArray(inArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray

    !-------------------------------------------------------------------
    ! Start

    hasDssArray = .false. ! default
    if (hasAttribute(inArray, 'DSTYP1')) hasDssArray = .true.

  end function hasDssArray

  !---------------------------------------------------------------------
  logical(bool) function hasDssTab(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab

    !-------------------------------------------------------------------
    ! Start

    hasDssTab = .false. ! default
    if (hasAttribute(tab, 'DSTYP1')) hasDssTab = .true.

  end function hasDssTab

  !---------------------------------------------------------------------
  logical(bool) function hasDssBlock(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock

    !-------------------------------------------------------------------
    ! Start

    hasDssBlock = .false. ! default
    if (hasAttribute(inBlock, 'DSTYP1')) hasDssBlock = .true.

  end function hasDssBlock

  !---------------------------------------------------------------------
  subroutine dumpDssBlock(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock

    type(DssT) :: dssPointer

    !-------------------------------------------------------------------
    ! Start:

    dssPointer = dataSubspace(inBlock)
    call dumpDssPointer(dssPointer)
    call dssRelease(dssPointer)

  end subroutine dumpDssBlock

  !---------------------------------------------------------------------
  subroutine dumpDssPointer(dssPointer)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) ::&
      dssPointer

    type(DScompT) :: dssComp
    integer :: numComps, compNum

    !-------------------------------------------------------------------
    ! Start:

    numComps = dssComponents(dssPointer)
    if (numComps <= 0) then
      print *, 'DSS contains no components.'
  return
    end if

    do compNum = 0, numComps - 1
      print *, 'DSS component number', compNum
      dssComp = dssComponent(dssPointer, compNum)
      call dumpDssComponent(dssComp)
    end do

  end subroutine dumpDssPointer

  !---------------------------------------------------------------------
  subroutine dumpDssComponent(dssComp)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) ::&
      dssComp

    type(DSfilterT) ::&
      filter
    type(RangeT), dimension(:), pointer :: range
    character(255) ::&
      blockName
    integer(kind=int16) ::&
      i
!    integer(int32) ::&
!      bitmask
integer(int32), pointer ::&
  onBitMasks(:),&
  offBitMasks(:)
    integer :: filtNum, filterType
    logical :: typeNotFound
    character(len=20) :: columnName, filterTypeStr
    character(len=80) :: axisNames
!    character(8) ::&
character(18) ::&
      bitMaskStr
    type filtInfo
      integer :: index
      character(len=20) :: name
    end type
    type(filtInfo) :: myFiltInfo(7)

    !-------------------------------------------------------------------
    ! Start:

    myFiltInfo(1)%index = RANGEFILTERT
    myFiltInfo(2)%index = REGIONFILTERT
    myFiltInfo(3)%index = BITMASKFILTERT
    myFiltInfo(4)%index = RANGEBLOCKFILTERT
    myFiltInfo(5)%index = GTIFILTERT
    myFiltInfo(6)%index = MASKFILTERT
    myFiltInfo(7)%index = REGIONBLOCKFILTERT
    myFiltInfo(1)%name = 'RANGEFILTERT'
    myFiltInfo(2)%name = 'REGIONFILTERT'
    myFiltInfo(3)%name = 'BITMASKFILTERT'
    myFiltInfo(4)%name = 'RANGEBLOCKFILTERT'
    myFiltInfo(5)%name = 'GTIFILTERT'
    myFiltInfo(6)%name = 'MASKFILTERT'
    myFiltInfo(7)%name = 'REGIONBLOCKFILTERT'

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      columnName = dssFilterName(filter)
      filterType = dssFilterType(filter)

      typeNotFound = .true.
      do i = 1, 7
        if (myFiltInfo(i)%index .eq. filterType) then
          typeNotFound = .false.
          filterTypeStr = myFiltInfo(i)%name
      exit
        end if
      end do
      if (typeNotFound) then
        write(errStr, '(a, i3.3, a, i3.3, a)') "ssclib:dumpDssComponent: Filter "&
         , filtNum, " in the current DSS component has type integer ", filterType&
         , "; this is not recognized."
call error('unknownDssFilterType')
      end if

      write(*, '(a, i3.3, a)') '  Filter=', filtNum&
        , ', axis='//trim(columnName)//', type='//filterTypeStr

      select case(filterType)
      case(RANGEFILTERT)
        range => dssRanges(filter)
        call dumpRanges(range, 4)

      case(REGIONBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '    Block name = ', trim(blockName)

      case(BITMASKFILTERT)
        call dssFilterMask(filter, onBitMasks, offBitMasks)
        print *, '    On-bit masks:'
        do i = lbound(onBitMasks, 1), ubound(onBitMasks, 1)
          write(bitMaskStr,  '(a5, i2, a3, z8)') 'mask ', i, ' = '&
            , onBitMasks(i)
          print *, '      ', bitMaskStr
        end do
        print *, '    Off-bit masks:'
        do i = lbound(offBitMasks, 1), ubound(offBitMasks, 1)
          write(bitMaskStr,  '(a5, i2, a3, z8)') 'mask ', i, ' = '&
            , offBitMasks(i)
          print *, '      ', bitMaskStr
        end do
!!!        deallocate(onBitMasks, offBitMasks)

      case(RANGEBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '    Block name = ', trim(blockName)

      case(GTIFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '    Block name = ', trim(blockName)

      case(MASKFILTERT, REGIONFILTERT)
        axisNames = dssFilterComps(filter)
        print *, '    Axis names', axisNames
      end select
    end do
    print *, ''

  end subroutine dumpDssComponent

  !---------------------------------------------------------------------
  subroutine getNum1stDssCompWithValue(dssPointer, axisName, axisValue&
  , firstCompNum)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue
    integer, intent(out) :: firstCompNum

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    firstCompNum = -1	! default to invalid value.

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue)) then
        firstCompNum = compNum
    exit
      end if
    end do

    if (firstCompNum < 0) then
      write(errStr, '(a, es9.3, a)') "ssclib:getNum1stDssCompWithValue: &
        &No DSS component passes value ", axisValue, " on column "//trim(axisName)//"."
call error('noDssCompThisValue')
    end if

  end subroutine getNum1stDssCompWithValue

  !---------------------------------------------------------------------
  logical(bool) function dssCompObeys(dssComp, dssConstraints)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),           intent(in) :: dssComp
    type(dssConstraintType), intent(in) :: dssConstraints(:)

    integer :: numConstraints, i, filtNum
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    dssCompObeys = .true. ! default

    numConstraints = size(dssConstraints)
    do i = 1, numConstraints
    if (numFiltersOfAxis(dssComp, dssConstraints(i)%name) .le. 0) cycle
      do filtNum = 0, dssFilters(dssComp) - 1
        filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .ne. dssConstraints(i)%name) cycle

        if (filterOverlaps(filter, dssConstraints(i))) then
          dssCompObeys = .false.
  return
        end if
      end do
    end do

  end function dssCompObeys

  !---------------------------------------------------------------------
  integer function numDssCompsOfAxis(dssPointer, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssCompsOfAxis = 0
    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (anyFilterOfThisAxis(dssComp, axisName))&
        numDssCompsOfAxis = numDssCompsOfAxis + 1
    end do

  end function numDssCompsOfAxis

  !---------------------------------------------------------------------
  integer function numDssCompsWithValue(dssPointer, axisName, axisValue)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssCompsWithValue = 0 ! default

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue))&
        numDssCompsWithValue = numDssCompsWithValue + 1
    end do

  end function numDssCompsWithValue

  !---------------------------------------------------------------------
  logical(bool) function valueIsPassedByDssComp(dssComp, axisName, axisValue)
  !---------------------------------------------------------------------
  ! This returns true ONLY if the specified component contains no filter on the specified axis OR the value is passed by all of the filters of this component that do exist for the given axis.

    implicit none

    type(DScompT),intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterIsOnSpecifiedAxis(:), valueIsPassed(:)

    !-------------------------------------------------------------------
    ! Start

!print *, '      aaa valueIsPassedByDssComp'

    numFilters = dssFilters(dssComp)
!print *, '      bbb'
    allocate(filterIsOnSpecifiedAxis(numFilters), valueIsPassed(numFilters))
!print *, '      ccc'
    filterIsOnSpecifiedAxis = .false.	! default
!print *, '      ddd'
    valueIsPassed           = .false.	! default

!print *, '      eee'
!print *, '      valueIsPassedByDssComp: numFilters=', numFilters
    do filtNum = 0, numFilters - 1
!print *, ''
!print *, '      filtNum:', filtNum
      filter = dssFilter(dssComp, filtNum)
!print *, '      fff'
      if (dssFilterName(filter) .eq. axisName) then
!print *, '      ggg'
        filterIsOnSpecifiedAxis(filtNum) = .true.
!print *, '      hhh'
        if (valueIsPassedByFilter(axisValue, filter)) then
!print *, '      iii'
          valueIsPassed(filtNum) = .true.
        end if
      end if
    end do
!print *, '      jjj'

    if (all(filterIsOnSpecifiedAxis .eqv. valueIsPassed)) then
      valueIsPassedByDssComp = .true.
    else
      valueIsPassedByDssComp = .false.
    end if
!print *, '      zzz valueIsPassedByDssComp'

  end function valueIsPassedByDssComp

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalInt(dssPointer, compNumA, compNumB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in)           :: dssPointer
    integer,      intent(in)           :: compNumA, compNumB
    character(*), intent(in), optional :: axisNamesToInclude(:),&
                                          axisNamesToNeglect(:)

    type(DScompT) :: dssCompA, dssCompB

    !-------------------------------------------------------------------
    ! Start

    dssCompA = dssComponent(dssPointer, compNumA)
    dssCompB = dssComponent(dssPointer, compNumB)

!*** DON'T forget comp numbers start at ZERO

    if (present(axisNamesToInclude)) then
      if (present(axisNamesToNeglect)) then
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude, axisNamesToNeglect)

      else
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude)
      end if

    else
      if (present(axisNamesToNeglect)) then
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToNeglect=axisNamesToNeglect)

      else
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB)
      end if
    end if

  end function dssCompsIdenticalInt

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdentNumVecNameScal(dssPointer, dssCompNums&
  , axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    integer,      intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisName

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdentNumVecNameScal = dssCompsIdentNumVecNameVec(dssPointer&
      , dssCompNums, (/axisName/))

  end function dssCompsIdentNumVecNameScal

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdentNumVecNameVec(dssPointer, dssCompNums&
  , axisNames)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    integer,      intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisNames(:)

    integer :: numDssCompIndices, numAxes, compIndex, compNum, i
    type(DScompT) :: dssCompA, dssCompB

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa dssCompsIdentNumVecNameVec'

    numDssCompIndices = size(dssCompNums)
    numAxes           = size(axisNames)
if (testFlag) print *, 'bbb numDssCompIndices=',numDssCompIndices
if (testFlag) print *, 'ccc numAxes=',numAxes
    if (numDssCompIndices < 2) then
      dssCompsIdentNumVecNameVec = .true.
  return
    end if

    dssCompsIdentNumVecNameVec = .false. ! default

    compIndex = 1
    compNum = dssCompNums(compIndex)
    dssCompA = dssComponent(dssPointer, compNum)
    do compIndex = 2, numDssCompIndices
      compNum = dssCompNums(compIndex)
      dssCompB = dssComponent(dssPointer, compNum)
if (testFlag) print *, '  c0 2nd compIndex=',compIndex
if (testFlag) print *, '  c1 2nd compNum=',compNum

      do i = 1, numAxes
if (testFlag) print *, '    c1a axis num=',i
if (testFlag) print *, '    c1b axis name=',trim(axisNames(i))
        if (.not. dssCompsIdenticalSimple(dssCompA, dssCompB, axisNames(i))) then
if (testFlag) print *, 'zzz components NOT identical.'
  return
        else
if (testFlag) print *, '    c1c components are identical.'
        end if
      end do
    end do

    dssCompsIdentNumVecNameVec = .true. ! passed all its tests.
if (testFlag) print *, 'zzz dssCompsIdentNumVecNameVec'

  end function dssCompsIdentNumVecNameVec

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalComp(dssCompA, dssCompB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------
!*** I don't think this works. Use List if you have no axisNamesToNeglect.


!**** NOTE comparison of reals is a bad idea - replace '==' by 'almost-equal'.

    implicit none

    type(DScompT), intent(in)           :: dssCompA, dssCompB
    character(*),  intent(in), optional :: axisNamesToInclude(:),&
                                           axisNamesToNeglect(:)

    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB
    logical, allocatable :: compBFilterMatched(:)
    character(20) :: filterNameA, filterNameB
    logical :: unableToMatchFilter
    type(DSfilterT) :: filterA, filterB

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdenticalComp = .false. ! default

    numFiltersA = dssFilters(dssCompA)
    numFiltersB = dssFilters(dssCompA)

    allocate(compBFilterMatched(numFiltersB))
    compBFilterMatched = .false. ! initially.
    do filtNumA = 1, numFiltersA
      ! Skip over filterA if it either _isn't_ in the list of axis names
      ! to include, or it _is_ in the list of axis names to neglect.

      filterNameA = dssFilterName(dssFilter(dssCompA, filtNumA-1))
      if (present(axisNamesToInclude)) then
    if (.not. any(filterNameA .eq. axisNamesToInclude)) cycle
      end if
      if (present(axisNamesToNeglect)) then
    if (any(filterNameA .eq. axisNamesToNeglect)) cycle
      end if

      unableToMatchFilter = .true. ! default
      do filtNumB = 1, numFiltersB
      if (compBFilterMatched(filtNumB)) cycle

        filterNameB = dssFilterName(dssFilter(dssCompB, filtNumB-1))
      if (filterNameA .ne. filterNameB) cycle

        ! Tick filterB off if it either _isn't_ in the list of axis names
        ! to include, or it _is_ in the list of axis names to neglect.

        if (present(axisNamesToInclude)) then
          if (.not. any(filterNameB .eq. axisNamesToInclude))&
            compBFilterMatched(filtNumB) = .true.
        end if
        if (present(axisNamesToNeglect)) then
          if (any(filterNameB .eq. axisNamesToNeglect))&
            compBFilterMatched(filtNumB) = .true.
        end if

        filterA = dssFilter(dssCompA, filtNumA - 1)
        filterB = dssFilter(dssCompB, filtNumB - 1)
        if (filtersAreIdentical(filterA, filterB)) then
          unableToMatchFilter = .false.
          compBFilterMatched(filtNumB) = .true.
      exit
        end if
      end do

  if (unableToMatchFilter) return
    end do

  if (any(.not. compBFilterMatched)) return

    dssCompsIdenticalComp = .true. ! made it through all the tests.

  end function dssCompsIdenticalComp

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalList(dssCompA, dssCompB, axisNames)
  !---------------------------------------------------------------------
  ! This returns true only if either (i) both components have no filter of the given axis name, or (ii) both components have the same number, type and values of filters of the given axis name.

    implicit none

    type(DScompT), intent(in) :: dssCompA, dssCompB
    character(*),  intent(in) :: axisNames(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdenticalList = .false. ! default

    do i = 1, size(axisNames)
      if (.not. dssCompsIdenticalSimple(dssCompA, dssCompB, axisNames(i)))&
  return
    end do

    dssCompsIdenticalList = .true. ! passed all its tests.

  end function dssCompsIdenticalList

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalSimple(dssCompA, dssCompB, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if either (i) both components have no filter of the given axis name, or (ii) both components have the same number, type and values of filters of the given axis name.

    implicit none

    type(DScompT), intent(in) :: dssCompA, dssCompB
    character(*),  intent(in) :: axisName

    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB, filtNumA1, filtNumA2, filtNumB1, filtNumB2
    logical, allocatable :: filtersThisAxisA(:), filtersThisAxisB(:)
    type(DSfilterT) :: filterA1, filterA2, filterB1, filterB2, filterA, filterB
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    dssCompsIdenticalSimple: axis name is '//trim(axisName)//'.'

    dssCompsIdenticalSimple = .false. ! default

    if   (.not. anyFilterOfThisAxis(dssCompA, axisName)&
    .and. .not. anyFilterOfThisAxis(dssCompB, axisName)) then
      dssCompsIdenticalSimple = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   no filter this name in either comp.'
  return
    end if

    numFiltersA = dssFilters(dssCompA)
    allocate(filtersThisAxisA(numFiltersA))
    filtersThisAxisA = .false. ! default
    do filtNumA = 1, numFiltersA
      if (dssFilterName(dssFilter(dssCompA, filtNumA - 1)) == axisName) then
!print *, 1
        filtersThisAxisA(filtNumA) = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   compA has a filter of this name, filter number', filtNumA-1
end if
    end do

    ! Knock out duplicated dssCompA filters:

    do filtNumA1 = 1, numFiltersA - 1
    if (.not. filtersThisAxisA(filtNumA1)) cycle
      filterA1 = dssFilter(dssCompA, filtNumA1 - 1)
      do filtNumA2 = filtNumA1 + 1, numFiltersA
      if (.not. filtersThisAxisA(filtNumA2)) cycle
        filterA2 = dssFilter(dssCompA, filtNumA2 - 1)
        if (filtersAreIdentical(filterA1, filterA2))&
          filtersThisAxisA(filtNumA2) = .false.
      end do
    end do

    numFiltersB = dssFilters(dssCompB)
    allocate(filtersThisAxisB(numFiltersB))
    filtersThisAxisB = .false. ! default
    do filtNumB = 1, numFiltersB
      if (dssFilterName(dssFilter(dssCompB, filtNumB - 1)) == axisName) then
!print *, 2
        filtersThisAxisB(filtNumB) = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   compB has a filter of this name, filter number', filtNumB-1
end if
    end do

    ! Knock out duplicated dssCompB filters:

    do filtNumB1 = 1, numFiltersB - 1
    if (.not. filtersThisAxisB(filtNumB1)) cycle
      filterB1 = dssFilter(dssCompB, filtNumB1 - 1)
      do filtNumB2 = filtNumB1 + 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB2)) cycle
        filterB2 = dssFilter(dssCompB, filtNumB2 - 1)
        if (filtersAreIdentical(filterB1, filterB2))&
          filtersThisAxisB(filtNumB2) = .false.
      end do
    end do

    if (count(filtersThisAxisA) .ne. count(filtersThisAxisB)) then
!print *, 3
  return
    end if
!print *, 4

    ! Now see if filters from A and B all match up. Each match-up is marked by setting the appropriate element of filtersThisAxisA (an arbitary choice, could equally well have chosen B for the honour) to false. Thus if any elements of filtersThisAxisA remain true at the end of the looping, this means not every filter was matched.

    ALOOP: do filtNumA = 1, numFiltersA
    if (.not. filtersThisAxisA(filtNumA)) cycle
      filterA = dssFilter(dssCompA, filtNumA - 1)

      BLOOP: do filtNumB = 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB)) cycle
        filterB = dssFilter(dssCompB, filtNumB - 1)

if (testFlag) print *, '    dssCompsIdenticalSimple: test filter numbers '&
, filtNumA-1, ' and ', filtNumB-1 ,' for identity'
        if (filtersAreIdentical(filterA, filterB)) then
!print *, '5 - YES'
          filtersThisAxisA(filtNumA) = .false.
if (testFlag) print *, '    dssCompsIdenticalSimple: for filtNumA=',filtNumA-1&
,' and filtNumB=',filtNumB-1,', filters are identical. Exit Bloop.'
      exit BLOOP
        else
!print *, '5 - NO'
        end if

      end do BLOOP
    end do ALOOP

if (testFlag) print *, '    dssCompsIdenticalSimple: filters unmatched?', filtersThisAxisA

    if (.not. any(filtersThisAxisA)) dssCompsIdenticalSimple = .true.

if (testFlag) print *, '    dssCompsIdenticalSimple: comps are identical=', dssCompsIdenticalSimple

  end function dssCompsIdenticalSimple

  !---------------------------------------------------------------------
  subroutine dssCompNumsWhichPassThisValue(dssPointer, axisName, axisValue&
  , dssCompNums)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate dssCompNums after use.

    implicit none

    type(DssT),   intent(in)  :: dssPointer
    character(*), intent(in)  :: axisName
    real(single), intent(in)  :: axisValue
    integer,      pointer     :: dssCompNums(:)

    integer :: numDssComps, compNum, numMatchingComps
    integer, allocatable :: tempCompNums(:)
    type(DScompT) :: dssComp
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa dssCompNumsWhichPassThisValue'
if (testFlag) print *, '  bbb axisName, axisValue: ', trim(axisName), axisValue

    numDssComps = dssComponents(dssPointer)
if (testFlag) print *, '  ccc numDssComps=', numDssComps
    allocate(tempCompNums(numDssComps))
    numMatchingComps = 0
    do compNum = 0, numDssComps - 1
if (testFlag) print *, '  ddd compNum=', compNum
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue)) then
if (testFlag) print *, '  eee allFiltersPassValue=T'
        numMatchingComps = numMatchingComps + 1
        tempCompNums(numMatchingComps) = compNum
      end if
    end do
if (testFlag) print *, '  fff numMatchingComps=', numMatchingComps
    allocate(dssCompNums(numMatchingComps))
    if (numMatchingComps > 0) dssCompNums = tempCompNums(1:numMatchingComps)

if (testFlag) print *, '  zzz dssCompNumsWhichPassThisValue'
  end subroutine dssCompNumsWhichPassThisValue

  !---------------------------------------------------------------------
  subroutine determineUniqueFilterCombos(dssPointer, dssCompNums, axisAName&
  , axisBName, uniqueABCombos)
  !---------------------------------------------------------------------
  ! Work through the list of components and find out how many unique sets of axisA-axisB filters there are for the present ccd.
  !
!*** test this with more complicated selections.

    implicit none

    type(DssT), intent(in) :: dssPointer
    integer, intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisAName, axisBName
    integer(int16), intent(out) :: uniqueABCombos(size(dssCompNums))

    integer(int16) :: numDssComps, numUniqueABFilters, startCompIndex, compIndex
    logical(bool), allocatable :: compIndexIsFree(:)
    integer :: startCompNum, dssCompNum
    type(DScompT) :: startComp, dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssComps = size(dssCompNums)
  if (numDssComps <= 0) return

    allocate(compIndexIsFree(numDssComps))

    uniqueABCombos = 0		! default
    compIndexIsFree = .true.	! default
    numUniqueABFilters = 0	! Although there should be
    ! at least 1. The routine returns if not, a few lines earlier.

    do while(any(compIndexIsFree))
      numUniqueABFilters = numUniqueABFilters + 1
      startCompIndex = smallestTrueIndex(compIndexIsFree)
      ! smallestTrueIndex is in ssclib/array_utils.

      startCompNum = dssCompNums(startCompIndex)
      startComp = dssComponent(dssPointer, startCompNum)
      uniqueABCombos(startCompIndex) = numUniqueABFilters
      compIndexIsFree(startCompIndex) = .false.

      do compIndex = startCompIndex + 1, numDssComps
      if (.not. compIndexIsFree(compIndex)) cycle

        dssCompNum = dssCompNums(compIndex)
        dssComp = dssComponent(dssPointer, dssCompNum)

        if   (dssCompsIdentical(startComp, dssComp, axisAName)&
        .and. dssCompsIdentical(startComp, dssComp, axisBName)) then
          uniqueABCombos(compIndex) = numUniqueABFilters
          compIndexIsFree(compIndex) = .false.
        end if
      end do
    end do

    if (numUniqueABFilters == 0)&
call error('dss_auxBug0',"ssclib:determineUniqueFilterCombos: please contact the code developer.")
    if (numUniqueABFilters > numDssComps)&
call error('dss_auxBug1',"ssclib:determineUniqueFilterCombos: please contact the code developer.")
    if (any(uniqueABCombos <= 0))&
call error('dss_auxBug2',"ssclib:determineUniqueFilterCombos: please contact the code developer.")

  end subroutine determineUniqueFilterCombos

  !---------------------------------------------------------------------
  function filterTypeToStr(filterType) result(filterTypeStr)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: filterType
    character(20) :: filterTypeStr

    !-------------------------------------------------------------------
    ! Start

    select case(filterType)
    case(RANGEFILTERT)
      filterTypeStr = 'RANGEFILTERT'
    case(REGIONFILTERT)
      filterTypeStr = 'REGIONFILTERT'
    case(BITMASKFILTERT)
      filterTypeStr = 'BITMASKFILTERT'
    case(RANGEBLOCKFILTERT)
      filterTypeStr = 'RANGEBLOCKFILTERT'
    case(MASKFILTERT)
      filterTypeStr = 'MASKFILTERT'
    case(REGIONBLOCKFILTERT)
      filterTypeStr = 'REGIONBLOCKFILTERT'
    case(GTIFILTERT)
      filterTypeStr = 'GTIFILTERT'
    case default
      filterTypeStr = 'UNKNOWN'
    end select

  end function filterTypeToStr

  !---------------------------------------------------------------------
  logical(bool) function filterOverlaps(filter, dssConstraint)
  !---------------------------------------------------------------------

    implicit none

    type(DSfilterT),         intent(in) :: filter
    type(dssConstraintType), intent(in) :: dssConstraint

    type(RangeT), pointer :: ranges(:)
    type(RangeT), pointer :: newRanges(:)

    !-------------------------------------------------------------------
    ! Start

    filterOverlaps = .true. ! default

  if (dssFilterType(filter) .ne. RANGEFILTERT) return
  if (size(dssConstraint%ranges) <= 0) return

    ranges => dssRanges(filter)
    if (size(ranges) <= 0) then
      deallocate(ranges)
  return
    end if

!!    newRanges = andRangesPair(oldToNew(ranges), dssConstraint%ranges)
!    newRanges => andRangesPair(ranges, dssConstraint%ranges)
call andRanges(ranges, dssConstraint%ranges, newRanges)
    if (size(newRanges) <= 0) filterOverlaps = .false.

    deallocate(ranges, newRanges)

  end function filterOverlaps

  !---------------------------------------------------------------------
  logical(bool) function anyFilterOfThisAxis(dssComp, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if the specified component contains at least one filter on the specified axis.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    numFilters = dssFilters(dssComp)
    anyFilterOfThisAxis = .false.	! default

    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        anyFilterOfThisAxis = .true.
  return
      end if
    end do

  end function anyFilterOfThisAxis

  !---------------------------------------------------------------------
  integer function numFiltersOfAxis(dssComp, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) :: dssComp
    character(*),  intent(in) :: axisName
!    real(single),  intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

!print *, '    aaa numFiltersOfAxis'
    numFiltersOfAxis = 0  ! default
!print *, '    bbb'

    numFilters = dssFilters(dssComp)
!print *, '    ccc'
    do filtNum = 0, numFilters - 1
!print *, '    ddd'
      filter = dssFilter(dssComp, filtNum)
!print *, '    eee'
      if (dssFilterName(filter) .eq. axisName)&
        numFiltersOfAxis = numFiltersOfAxis + 1 
    end do
!print *, '    zzz numFiltersOfAxis'

  end function numFiltersOfAxis

  !---------------------------------------------------------------------
  subroutine get1stFilterThisAxis(dssComp, axisName, firstFilter)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),   intent(in)  :: dssComp
    character(*),    intent(in)  :: axisName
    type(DSfilterT), intent(out) :: firstFilter

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical(bool) :: noFilterThisAxis

    !-------------------------------------------------------------------
    ! Start

    noFilterThisAxis = .true. ! default to bad status value.

    numFilters = dssFilters(dssComp)
    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        noFilterThisAxis = .false.
        firstFilter = filter
    exit
      end if
    end do

    if (noFilterThisAxis)&
call error('noDssFilterThisAxis',"ssclib:get1stFilterThisAxis: The current DSS &
&component contains no filter on column "//trim(axisName)//".")

  end subroutine get1stFilterThisAxis

  !---------------------------------------------------------------------
  logical(bool) function allFiltersPassValue(dssComp, axisName, axisValue)
  !---------------------------------------------------------------------
  ! This returns true EITHER if the specified component contains no filter on the specified axis OR all of the filters of this component and on this axis pass the specified value.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterOnThisAxisExists(:), valueIsPassed(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa allFiltersPassValue'
if (testFlag) print *, '    bbb axisName, axisValue: ', trim(axisName), axisValue

    numFilters = dssFilters(dssComp)
if (testFlag) print *, '    ccc numFilters=', numFilters
    allocate(filterOnThisAxisExists(0:numFilters - 1)&
    , valueIsPassed(0:numFilters - 1))
    filterOnThisAxisExists = .false.	! default
    valueIsPassed          = .false.	! default

    do filtNum = 0, numFilters - 1
if (testFlag) print *, '    ddd filtNum=', filtNum
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        filterOnThisAxisExists(filtNum) = .true.
        if (valueIsPassedByFilter(axisValue, filter))&
          valueIsPassed(filtNum) = .true. 
      end if
if (testFlag) print *, '    eee filterOnThisAxisExists=', filterOnThisAxisExists(filtNum)
if (testFlag) print *, '    fff valueIsPassed=', valueIsPassed(filtNum)
    end do

!    if (.not. any(filterOnThisAxisExists) .or. any(filterOnThisAxisExists&
!    .and. .not. valueIsPassed)) then
if (any(filterOnThisAxisExists .and. .not. valueIsPassed)) then
      allFiltersPassValue = .false.
    else
      allFiltersPassValue = .true.
    end if
if (testFlag) print *, '    ggg allFiltersPassValue=', allFiltersPassValue

if (testFlag) print *, '    zzz allFiltersPassValue'

  end function allFiltersPassValue

  !---------------------------------------------------------------------
  logical(bool) function valueIsPassedByFilter(value, filter)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: value
    type(DSfilterT), intent(in) :: filter

    type(RangeT), pointer :: ranges(:)
    character(255)  :: gtiBlockName
    type(TableT)    :: tab
    integer         :: numGtis, i, filterType
    type(RangeT), allocatable :: gti(:)

    !-------------------------------------------------------------------
    ! Start

    valueIsPassedByFilter = .false.

!    select case(dssFilterType(filter))
filterType = dssFilterType(filter)
    select case(filterType)
    case(RANGEFILTERT)
      ranges => dssRanges(filter)
      do i = 0, size(ranges) - 1
        if (valueIsWithinRange(value, ranges(i))) then
          valueIsPassedByFilter = .true.
      exit
        end if
      end do

      deallocate(ranges)

!    case(REGIONFILTERT)
!      selectionExpression = selectExpression(filter)
!*** for a future version of valueIsPassedByFilter() with 2-value
!*** alternative input: set up pseudo-table with 1 entry, filter it,
!*** count num filtered rows.
!    case(BITMASKFILTERT)
!    case(RANGEBLOCKFILTERT)
!    case(MASKFILTERT)
!    case(REGIONBLOCKFILTERT)

    case(GTIFILTERT) ! Treat GTIs as ranges:
      gtiBlockName = dssFilterBlock(filter)
      tab = table(dataSet(gtiBlockName, READ), gtiBlockName)
      numGtis = numberOfRows(tab)
      allocate(gti(numGtis))
      call readColDataToFixed(tab, 'START', gti%lower%value)
      call readColDataToFixed(tab, 'STOP',  gti%upper%value)
      call release(tab)

      gti%lower%type = INCLUSIVE
      gti%upper%type = INCLUSIVE
      do i = 1, size(gti)
        if (valueIsWithinRange(value, gti(i))) then
          valueIsPassedByFilter = .true.
      exit
        end if
      end do

    case default
      write(errStr, '(a, i3.3, a, i3.3, a, i3.3, a)') "ssclib:valueIsPassedByFilter: &
      &The filter of the current DSS component is of type ", filterType&
      , "; however it must be either a range filter (type ", RANGEFILTERT&
      , ") or a GTI filter (type ", GTIFILTERT, ")."
call error('badDssFilterType')
    end select

  end function valueIsPassedByFilter

  !---------------------------------------------------------------------
  logical(bool) function filtersAreIdentical(filterA, filterB)
  !---------------------------------------------------------------------

!**** NOTE comparison of reals is a bad idea - replace '==' by 'almost-equal'.

    implicit none

    type(DSfilterT), intent(in) :: filterA, filterB

    type(RangeT), pointer :: rangesA(:), rangesB(:)
    integer :: i, numRowsA, numRowsB, row, filterType
    character(255) :: blockNameA, blockNameB
    character(20) :: xyNameString, xColName, yColName, shapeA, shapeB
    type(TableT) :: tableA, tableB
    real(double), pointer :: colPtrA(:,:), colPtrB(:,:)
    type(ColumnT) :: colA, colB
    type(ArrayT) :: arrayA, arrayB
!    integer(int32), pointer :: imageA(:,:), imageB(:,:)
logical(bool), pointer :: imageA(:,:), imageB(:,:)
integer(int32), pointer ::&
  onBitMasksA(:), onBitMasksB(:), offBitMasksA(:), offBitMasksB(:)

logical :: testFlag = .false.
character(80) :: strA, strB
integer :: intA, intB

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa filtersAreIdentical'

    filtersAreIdentical = .false. ! default

!!!  if (dssFilterName(filterA)  .ne. dssFilterName(filterB))  return
strA = dssFilterName(filterA)
strB = dssFilterName(filterB)
if (strA .ne. strB)  return

if (testFlag) print *, '  1'

!!!  if (dssFilterType(filterA)  .ne. dssFilterType(filterB))  return
intA = dssFilterType(filterA)
intB = dssFilterType(filterB)
if (intA .ne. intB)  return

if (testFlag) print *, '  2'

!!!  if (dssFilterUnits(filterA) .ne. dssFilterUnits(filterB)) return
strA = dssFilterUnits(filterA)
strB = dssFilterUnits(filterB)
if (testFlag) print *, '*'//strA//'*'
if (testFlag) print *, '$'//strB//'$'
if (strA .ne. strB) return

if (testFlag) print *, '  3'

!    select case(dssFilterType(filterA))
    filterType = dssFilterType(filterA)
    select case(filterType)
    case(RANGEFILTERT)
if (testFlag) print *, '  RANGEFILTERT'
!      rangesA => oldToNewPtr(dssRanges(filterA))
!      rangesB => oldToNewPtr(dssRanges(filterB))
rangesA => dssRanges(filterA)
rangesB => dssRanges(filterB)

  if (size(rangesA) .ne. size(rangesB)) return

      do i = 0, size(rangesA) - 1
        if  (rangesA(i)%lower%type  .ne. rangesB(i)%lower%type&
        .or. rangesA(i)%lower%value .ne. rangesB(i)%lower%value&
        .or. rangesA(i)%upper%type  .ne. rangesB(i)%upper%type&
        .or. rangesA(i)%upper%value .ne. rangesB(i)%upper%value)&
  return
      end do

    case(REGIONBLOCKFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      tableA = table(dataSet(blockNameA, READ), blockNameA)
      tableB = table(dataSet(blockNameB, READ), blockNameB)
      if (.not. columnSetIsIdentical(tableA, tableB)) then ! in dal_aux
        call release(tableA)
        call release(tableB)
  return
      end if

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableA)
        call release(tableB)
  return
      end if

      xyNameString = stringAttribute(tableA, 'MFORM1')
      call decomposeRegionXyNameString(xyNameString, xColName, yColName)

      call readColDataToPtr(tableA, xColName, colPtrA)
      call readColDataToPtr(tableB, xColName, colPtrB)
!***** these readColDatas might be better toFixed, then wouldn't have to keep deallocating the darn pointers!

      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
        deallocate(colPtrA, colPtrB)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      call readColDataToPtr(tableA, yColName, colPtrA)
      call readColDataToPtr(tableB, yColName, colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
        deallocate(colPtrA, colPtrB)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      if(hasColumn(tableA, 'SHAPE')) then
        colA = column(tableA, 'SHAPE')
        colB = column(tableB, 'SHAPE')
        do row = 0, numRowsA - 1
          shapeA = ucase(stringCell(colA, row))
          shapeB = ucase(stringCell(colB, row))
          if (shapeA .ne. shapeB) then
            call release(tableA)
            call release(tableB)
  return
          end if
        end do
      end if

      if(hasColumn(tableA, 'R')) then
        call readColDataToPtr(tableA, 'R', colPtrA)
        call readColDataToPtr(tableB, 'R', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableA)
          call release(tableB)
          deallocate(colPtrA, colPtrB)
  return
        end if
!*** possibly could be fooled by nulls
      end if

      deallocate(colPtrA, colPtrB)

      if(hasColumn(tableA, 'ROTANG')) then
        call readColDataToPtr(tableA, 'ROTANG', colPtrA)
        call readColDataToPtr(tableB, 'ROTANG', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableA)
          call release(tableB)
          deallocate(colPtrA, colPtrB)
  return
        end if
!*** possibly could be fooled by nulls
      end if

      deallocate(colPtrA, colPtrB)

    case(BITMASKFILTERT)
!  if(dssFilterMask(filterA) .ne. dssFilterMask(filterB)) return

!*** for devtrack replace this by the following:
call dssFilterMask(filterA, onBitMasksA, offBitMasksA)
call dssFilterMask(filterB, onBitMasksB, offBitMasksB)
if (size(onBitMasksA) .ne. size(onBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (size(offBitMasksA) .ne. size(offBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (any(onBitMasksA .ne. onBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (any(offBitMasksA .ne. offBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if

    case(GTIFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      tableA = table(dataSet(blockNameA, READ), blockNameA)
      tableB = table(dataSet(blockNameB, READ), blockNameB)

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableA)
        call release(tableB)
  return
      end if

      call readColDataToPtr(tableA, 'TSTART', colPtrA)
      call readColDataToPtr(tableB, 'TSTART', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
        deallocate(colPtrA, colPtrB)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      call readColDataToPtr(tableA, 'TSTOP', colPtrA)
      call readColDataToPtr(tableB, 'TSTOP', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
        deallocate(colPtrA, colPtrB)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

    case(MASKFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      arrayA = array(dataSet(blockNameA, READ), blockNameA, READ)
      arrayB = array(dataSet(blockNameB, READ), blockNameB, READ)
      call readArrayData(arrayA, imageA)
      call readArrayData(arrayB, imageB)

      if  (size(imageA, 1) .ne. size(imageB, 1)&
      .or. size(imageA, 2) .ne. size(imageB, 2)) then
        deallocate(imageA, imageB)
        call release(arrayA)
        call release(arrayB)
  return
      end if

!      if  (any(imageA .ne. imageB)) then
if  (any(imageA .neqv. imageB)) then
        deallocate(imageA, imageB)
        call release(arrayA)
        call release(arrayB)
  return
      end if

!    case(RANGEBLOCKFILTERT)	! doesn't seem to be any firm convention
!				! for it yet.

!    case(REGIONFILTERT)	! not normally used in the DSS.

    case default
      write(errStr, '(a, i3.3, a)') "ssclib:filtersAreIdentical: The input &
        &filters are of type ", filterType, ", which is not recognized."
call error('badDssFilterType')
    end select

    filtersAreIdentical = .true. ! Made it through all the tests!
if (testFlag) print *, 'zzz filtersAreIdentical'

  end function filtersAreIdentical

  !---------------------------------------------------------------------
  function getRangesSpan(dssPointer, axisName) result(dssRangeSpan)
  !---------------------------------------------------------------------
  ! This subroutine does a logical OR of all the ranges of all the
  ! filters on axisName in all the components in the DSS and returns
  ! the extrema of that ORed range.

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    type(RangeT) :: dssRangeSpan

    type(DScompT) :: dssComp
    type(RangeT), pointer  :: oredRanges(:), compRanges(:)
!, tempOredRanges(:)
    integer(int32) :: numComps, compNum, numRanges

    !-------------------------------------------------------------------
    ! Start

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
      dssRangeSpan%lower%type = UNDEFINED
      dssRangeSpan%lower%value = 0.0
      dssRangeSpan%upper%type = UNDEFINED
      dssRangeSpan%upper%value = 0.0
  return
    end if

    dssComp = dssComponent(dssPointer, 0)
    call readRangesComp(dssComp, axisName, oredRanges)

    do compNum = 1, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)
      call readRangesComp(dssComp, axisName, compRanges)
!      tempOredRanges => orRangesPair(compRanges, oredRanges)
!      deallocate(oredRanges) ! clears the memory
!      oredRanges => tempOredRanges
call orRanges(compRanges, oredRanges)
deallocate(compRanges)
    end do

    numRanges = size(oredRanges)
    dssRangeSpan%lower%type  = oredRanges(1)%lower%type
    dssRangeSpan%lower%value = oredRanges(1)%lower%value
    dssRangeSpan%upper%type  = oredRanges(numRanges)%upper%type
    dssRangeSpan%upper%value = oredRanges(numRanges)%upper%value

    deallocate(oredRanges) ! also clears tempOredRanges, since they
    ! are associated and thus point to the same memory.

  end function getRangesSpan

!***** add a variant that takes just a dss component??



  !---------------------------------------------------------------------
  subroutine readRangesDss(dssPointer, axisName, ranges, dssConstraints)
  !---------------------------------------------------------------------
  ! The function of this subroutine is to return a vector of dss type
  ! RangeT containing the ranges of the axis 'axisName'. Note that the sub
  ! will succeed only if that subset of the set of DSS components which obeys the supplied constraints can be expressed as (ranges in axisName) OR (all filters on other axes). This has been a complicated job to implement because a chain of ranges is implicitly an OR of the individual ranges and for this reason can be distributed among several components. The subroutine works therefore as follows. Firstly, blocks of components are identified which each include all components obeying the constraints and having the same range chain in axisName. Secondly, the non-axisName filters of each block of components are checked for identity. If they all prove to be the same, then their respective range chains are ORed together.
  !
  ! The returned pointer ranges should be deallocated after use.

    implicit none

    type(DssT),              intent(in)            :: dssPointer
    character(*),            intent(in)            :: axisName
    type(RangeT),            pointer               :: ranges(:)
    type(dssConstraintType), intent(in), optional  :: dssConstraints(:)

    logical :: thereAreConstraints, noCompObeysConstraints, isNewRangeChain
    integer :: numComps, compNum, numCompSubsets, compSubsetNum
    type(dssCompSubsetType), pointer :: compSubsets(:)
    type(DScompT) :: dssComp
    type(RangeT), pointer :: compRanges(:)
!, tempRanges(:)
    character(20) :: axisNamesToNeglect(1)

    !-------------------------------------------------------------------
    ! Start

    thereAreConstraints = .false.
    if (present(dssConstraints)) then
      if (size(dssConstraints) > 0) thereAreConstraints = .true.
    end if

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
      allocate(ranges(0))
  return
    end if

    noCompObeysConstraints = .true. ! default
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      noCompObeysConstraints = .false.
    exit
    end do

    if (noCompObeysConstraints) then
      allocate(ranges(0))
  return
    end if

    !...................................................................
    ! Start of main routine. First load all the blocks of constraint-
    ! -passing components:

    numCompSubsets = 0
    allocate(compSubsets(0))
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      call readRangesComp(dssComp, axisName, compRanges)

      ! Does it have a range chain the same as any of the previous range chains?
      isNewRangeChain = .true. ! default
      do compSubsetNum = 1, numCompSubsets
        if (rangesPairIsIdentical(compRanges&
          , compSubsets(compSubsetNum)%ranges)) then

          isNewRangeChain = .false.
        end if
      end do

      if (isNewRangeChain) then ! make a new component subset:
        call addNewCompSubset(compSubsets, numCompSubsets, compRanges, compNum)
      else ! it is the same ranges as a previous block
        call addNewCompNumToSubset(compSubsets(numCompSubsets)%compNums&
          , compNum)
      end if

      deallocate(compRanges)
    end do
if (numCompSubsets <= 0) call error('dss_auxBug3',"ssclib:readRangesDss: please contact the code developer.")

    !...................................................................
    ! Now check that, apart from filters on axisName, the subsets are the same:

    axisNamesToNeglect(1) = axisName
    do compSubsetNum = 2, numCompSubsets
      if (.not. compSubsetsIdentical(dssPointer, compSubsets(compSubsetNum)&
        , compSubsets(1), axisNamesToNeglect)) then

        allocate(ranges(0))
  return
      end if
    end do

    !...................................................................
    ! OR together the total range chain:

    allocate(ranges(size(compSubsets(1)%ranges)))
    ranges = compSubsets(1)%ranges
    do compSubsetNum = 2, numCompSubsets
!      tempRanges => orRangesPair(compSubsets(compSubsetNum)%ranges, ranges)
!      deallocate(ranges)
!      ranges => tempRanges
call orRanges(compSubsets(compSubsetNum)%ranges, ranges)
    end do

    !...................................................................
    ! Clear up all the subset pointers:

    do compSubsetNum = 1, numCompSubsets
      deallocate(compSubsets(compSubsetNum)%compNums, compSubsets(compSubsetNum)%ranges)
    end do
    deallocate(compSubsets)

  end subroutine readRangesDss

  !---------------------------------------------------------------------
  subroutine addNewCompSubset(compSubsets, numCompSubsets, ranges, compNum)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), pointer   :: compSubsets(:)
    integer,             intent(inout) :: numCompSubsets
    type(RangeT),        intent(in)    :: ranges(:)
    integer,             intent(in)    :: compNum

    type(dssCompSubsetType), pointer :: tempCompSubsets(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempCompSubsets(numCompSubsets))
    call copyCompSubsets(compSubsets, tempCompSubsets)
    deallocate(compSubsets)
    numCompSubsets = numCompSubsets + 1
    allocate(compSubsets(numCompSubsets))
    call copyCompSubsets(tempCompSubsets&
      , compSubsets(:numCompSubsets-1))
    deallocate(tempCompSubsets)
    call copyRanges(ranges, compSubsets(numCompSubsets)%ranges)
    allocate(compSubsets(numCompSubsets)%compNums(1))
    compSubsets(numCompSubsets)%compNums(1) = compNum

  end subroutine addNewCompSubset

  !---------------------------------------------------------------------
  subroutine addNewCompNumToSubset(compNums, compNum)
  !---------------------------------------------------------------------

    implicit none

    integer, pointer    :: compNums(:)
    integer, intent(in) :: compNum

    integer :: numComps
    integer, allocatable :: tempCompNums(:)

    !-------------------------------------------------------------------
    ! Start

    numComps = size(compNums)
    allocate(tempCompNums(numComps))
    tempCompNums = compNums
    deallocate(compNums)
    numComps = numComps + 1
    allocate(compNums(numComps))
    compNums(:numComps-1) = tempCompNums
    compNums(numComps) = compNum

  end subroutine addNewCompNumToSubset

  !---------------------------------------------------------------------
  subroutine copyCompSubsets(inCompSubsets, outCompSubsets)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(in)  :: inCompSubsets(:)
    type(dssCompSubsetType), intent(out) :: outCompSubsets(size(inCompSubsets))

    integer :: numCompSubsets, compSubsetNum

    !-------------------------------------------------------------------
    ! Start

    numCompSubsets = size(inCompSubsets)
    do compSubsetNum = 1, numCompSubsets
      outCompSubsets(compSubsetNum)&
        = copyCompSubset(inCompSubsets(compSubsetNum))
    end do

  end subroutine copyCompSubsets

  !---------------------------------------------------------------------
  function copyCompSubset(inCompSubset) result(outCompSubset)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(in) :: inCompSubset
    type(dssCompSubsetType)             :: outCompSubset

    !-------------------------------------------------------------------
    ! Start

    call copyRanges(inCompSubset%ranges, outCompSubset%ranges)
    allocate(outCompSubset%compNums(size(inCompSubset%compNums)))
    outCompSubset%compNums = inCompSubset%compNums

  end function copyCompSubset

  !---------------------------------------------------------------------
  logical(bool) function compSubsetsIdentical(dssPointer, compSubsetA&
  , compSubsetB, axisNamesToNeglect)
  !---------------------------------------------------------------------
!**** Not finished. axisNamesToNeglect should be optional (also should include optional axisNamesToInclude), with some (graceful hopefully!) way of passing them to dssCompsIdentical().

    implicit none

    type(DssT),              intent(in) :: dssPointer
    type(dssCompSubsetType), intent(in) :: compSubsetA,&
                                           compSubsetB
    character(*),            intent(in) :: axisNamesToNeglect(:)

    integer :: numComps, compNumA, compNumB
    logical, allocatable :: compSubsetBCompMatched(:)
    logical :: unableToMatchComponent

    !-------------------------------------------------------------------
    ! Start

    compSubsetsIdentical = .false. ! default

    numComps = size(compSubsetA%compNums)
  if (numComps .ne. size(compSubsetB%compNums)) return

    allocate(compSubsetBCompMatched(numComps))
    compSubsetBCompMatched = .false. ! initially.
    do compNumA = 1, numComps
      unableToMatchComponent = .true. ! default
      do compNumB = 1, numComps
      if (compSubsetBCompMatched(compNumB)) cycle
        if (dssCompsIdentical(dssPointer, compSubsetA%compNums(compNumA)&
          , compSubsetB%compNums(compNumB), axisNamesToNeglect)) then
          unableToMatchComponent = .false.
          compSubsetBCompMatched(compNumB) = .true.
      exit
        end if
      end do

  if (unableToMatchComponent) return
    end do

  if (any(.not. compSubsetBCompMatched)) return

    compSubsetsIdentical = .true. ! made it through all the tests.

  end function compSubsetsIdentical

  !---------------------------------------------------------------------
  subroutine readRangesComp(dssComp, axisName, compRanges)
  !---------------------------------------------------------------------
  ! NOTE that the pointer compRanges() should be deallocated after use.

    type(DScompT), intent(in) :: dssComp
    character(*),  intent(in) :: axisName
    type(RangeT), pointer     :: compRanges(:)

    integer :: filtNum
    type(DSfilterT) :: filter
    type(RangeT),  pointer :: ranges(:)
!, tempRangeChain(:)
    character(20) :: dssFilterTypeStr

    !-------------------------------------------------------------------
    ! Start

    allocate(compRanges(1))
    compRanges%lower%type = UNDEFINED
    compRanges%upper%type = UNDEFINED
    compRanges%lower%value = 0.0 ! unnecessary, but I don't
    compRanges%upper%value = 0.0 ! like to leave them hanging.

  if (numFiltersOfAxis(dssComp, axisName) .le. 0) return

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
    if (dssFilterName(filter) .ne. axisName) cycle

      if (dssFilterType(filter) .eq. RANGEFILTERT) then
        ranges => dssRanges(filter)
!        tempRangeChain => andRangesPair(compRanges, ranges)
!        deallocate(compRanges)
!        deallocate(ranges)
!        compRanges => tempRangeChain
call andRanges(ranges, compRanges)
deallocate(ranges)

      else
        allocate(compRanges(0))
        dssFilterTypeStr = filterTypeToStr(dssFilterType(filter))
        call warning('badDssFilterType', "ssclib:readRangesComp: A filter of type "&
          //trim(dssFilterTypeStr)//" on the axis "//trim(axisName)&
          //" was detected. I couldn't read the information from this: &
          &I can only read the information in RANGEFILTERT filters.")
      end if
    end do

  end subroutine readRangesComp

  !---------------------------------------------------------------------
  subroutine readGtisFromDssBlock(set, inBlock, ccdNum, gti)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gti() should be deallocated after use.
!*** obsolete - should replace use of this call by readGtisFromDss.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(BlockT),    intent(in) :: inBlock
    integer,         intent(in) :: ccdNum
    type(IntervalT), pointer    :: gti(:)

    type(DssT)      :: dssPointer
    type(DScompT)   :: dssComp
    integer         :: compNum, filtNum, filterType
    type(DSfilterT) :: filter
    logical         :: firstFilterOnTime
    type(IntervalT), pointer :: localGti(:)

    !-------------------------------------------------------------------
    ! Start

    dssPointer = dataSubspace(inBlock)

    if (numDssCompsWithValue(dssPointer, 'CCDNR', real(ccdNum, single))<=0) then
      allocate(gti(0))
  return
    end if

    ! Find the dss component with the correct CCDNR:
    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', real(ccdNum, single)&
      , compNum)

    dssComp = dssComponent(dssPointer, compNum)

    if (numFiltersOfAxis(dssComp, 'TIME') <= 0) then
      allocate(gti(0))
  return
    end if

    firstFilterOnTime = .true.
    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
    if (dssFilterName(filter) .ne. 'TIME') cycle

!      if (dssFilterType(filter) .eq. GTIFILTERT) then
      filterType = dssFilterType(filter)
      if (filterType .eq. GTIFILTERT) then
        if (firstFilterOnTime) then
          firstFilterOnTime = .false.
          call getGtiFromFilter(set, filter, gti)

        else
          call getGtiFromFilter(set, filter, localGti)
          call andIntervals(localGti, gti)
          deallocate(localGti)
        end if

      else
        write(errStr, '(a, i2.2, a, i3.3, a, i3.3, a)') "ssclib:readGtisFromDssBlock: &
          &This routine attempts to read a DSS filter on the TIME column for CCD "&
          , ccdNum, ". A filter of type ID ", filterType, " was found, however only &
          &filters of type GTIFILTERT (ID ", GTIFILTERT, ") are recognized."
        call warning('dssUnknownTimeFilter')
      end if
    end do

    call dssRelease(dssPointer)

  end subroutine readGtisFromDssBlock

  !---------------------------------------------------------------------
  subroutine readGtisFromDssPointer(set, dssPointer, ccdNum, gtis)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gtis() should be deallocated after use.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(DssT),      intent(in) :: dssPointer
    integer,         intent(in) :: ccdNum
    type(IntervalT), pointer    :: gtis(:)

    type(DScompT)   :: dssComp
    integer         :: compNum, filtNum, filterType
    type(DSfilterT) :: filter
    logical         :: firstFilterOnTime
    type(IntervalT), pointer :: localGtis(:)

    !-------------------------------------------------------------------
    ! Start

    if (numDssCompsWithValue(dssPointer, 'CCDNR', real(ccdNum, single))<=0) then
      allocate(gtis(0))
  return
    end if

    ! Find the dss component with the correct CCDNR:
    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', real(ccdNum, single)&
      , compNum)

    dssComp = dssComponent(dssPointer, compNum)

    if (numFiltersOfAxis(dssComp, 'TIME') <= 0) then
      allocate(gtis(0))
  return
    end if

    firstFilterOnTime = .true.
    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
    if (dssFilterName(filter) .ne. 'TIME') cycle

!      if (dssFilterType(filter) .eq. GTIFILTERT) then
      filterType = dssFilterType(filter)
      if (filterType .eq. GTIFILTERT) then
        if (firstFilterOnTime) then
          firstFilterOnTime = .false.
          call getGtiFromFilter(set, filter, gtis)

        else
          call getGtiFromFilter(set, filter, localGtis)
          call andIntervals(localGtis, gtis)
          deallocate(localGtis)
        end if

      else
        write(errStr, '(a, i2.2, a, i3.3, a, i3.3, a)') "ssclib:readGtisFromDssPointer: &
          &This routine attempts to read a DSS filter on the TIME column for CCD "&
          , ccdNum, ". A filter of type ID ", filterType, " was found, however only &
          &filters of type GTIFILTERT (ID ", GTIFILTERT, ") are recognized."
        call warning('dssUnknownTimeFilter')
      end if
    end do

  end subroutine readGtisFromDssPointer

  !---------------------------------------------------------------------
  subroutine getGtiFromFilter(set, filter, gti)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gti() should be deallocated after use.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(DSfilterT), intent(in) :: filter
    type(IntervalT), pointer    :: gti(:)

    character(255)  :: gtiBlockName
    type(TableT)    :: tab
    integer         :: numGtis

    !-------------------------------------------------------------------
    ! Start

    gtiBlockName = dssFilterBlock(filter)
    tab = table(set, gtiBlockName)
    numGtis = numberOfRows(tab)
    allocate(gti(numGtis))
    call readColDataToFixed(tab, 'START', gti%lower)
    call readColDataToFixed(tab, 'STOP',  gti%upper)
    call release(tab)

  end subroutine getGtiFromFilter

  !---------------------------------------------------------------------
  subroutine decomposeRegionXyNameString(xyNameString, xColName, yColName)
  !---------------------------------------------------------------------
  ! Designed to analyse the output of the call dssFilterComps, which returns a string containing the names of the 'X' and 'Y' columns of a region specification block separated by a comma.

    implicit none

    character(*), intent(in) :: xyNameString
    character(*), intent(out) :: xColName, yColName

    integer :: posOfComma

    !-------------------------------------------------------------------
    ! Start

    xColName = '' ! default
    yColName = '' ! default

    posOfComma = index(xyNameString, ',')
    if (posOfComma == 0)&
call error('commaNotFound', 'ssclib:decomposeRegionXyNameString: The call &
&dssFilterComps is expected to return a string containing the names of the &
&X and Y columns, separated by a comma. However no comma was found in the string.')

    xColName = trim(adjustl(xyNameString(:posOfComma-1)))
    yColName = trim(adjustl(xyNameString(posOfComma+1:)))

  end subroutine decomposeRegionXyNameString

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_aux
