!
! Module: dump
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: dump.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dump

  use dal
!  use errorHandling !*** convert prints to message() calls??

  implicit none 

  private 

  public printMatrix
  public dumpString
  public dumpImageToFits
!  public tableNullsDump

  interface dumpImageToFits
    module procedure dumpImageToFitsSingle, dumpImageToFitsDouble
    module procedure dumpImageToFitsInt8, dumpImageToFitsInt16
    module procedure dumpImageToFitsInt32, dumpImageToFitsBool_temp
  end interface dumpImageToFits

contains 

  !---------------------------------------------------------------------
  subroutine printMatrix(message, matrix, form)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: message
    real(double), intent(in) :: matrix(3,3)
    character(*), intent(in), optional :: form

    integer :: yi
    character(5) :: yiStr(3)
    character(8) :: localForm

    !-------------------------------------------------------------------
    ! Start

    if (present(form)) then
      localForm = form
    else
      localForm = 'floating'
    end if

    yiStr = (/'     ', 'yi:  ', '     '/)

    print *, ''
    print *, message
    print *, '                             xi:'
    print *, '               1              2              3'
    if (localForm .eq. 'floating') then
      do yi = 1, 3
        write(*, '(a5, i1, a2, 3f14.9)')&
        yiStr(yi), yi, '  ', matrix(1, yi), matrix(2, yi), matrix(3, yi)
      end do
    else
      do yi = 1, 3
        write(*, '(a5, i1, a2, 3es14.5)')&
        yiStr(yi), yi, '  ', matrix(1, yi), matrix(2, yi), matrix(3, yi)
      end do
    end if
    print *, ''

  end subroutine printMatrix

  !---------------------------------------------------------------------
  subroutine dumpString(str)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str

    character(1) :: chr
    integer :: length, i, index

    !-------------------------------------------------------------------
    ! Start:

    length = len(str)

    if (length .le. 0) then
      print *, 'There are 0 characters in the string.'
  return

    else if (length .eq. 1) then
      print *, 'There is 1 character in the string.'

    else
      print *, 'There are', length, 'characters in the string.'
    end if
    print *, ''

    do i = 1, length
      chr = str(i:i)
      index = ichar(chr)
      if (index .le. 32 .or. index .ge. 127) then
        chr = ' '
      end if
      write(*, '(a3, z2)') chr//'  ', index
    end do

  end subroutine dumpString

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsSingle(image, setName)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping real(single) image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, '', '', 0)
    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = image
    call release(outSet)

  end subroutine dumpImageToFitsSingle

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsDouble(image, setName)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping real(double) image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)
    outArray = addArray(outSet, 'PRIMARY', REAL64, dimVector, '', '', 0)
    arrayDataReal64 => real64Array2Data(outArray)
    arrayDataReal64 = image
    call release(outSet)

  end subroutine dumpImageToFitsDouble

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsInt8(image, setName)
  !---------------------------------------------------------------------

    implicit none

    integer(int8), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping int8 image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)
    outArray = addArray(outSet, 'PRIMARY', INTEGER8, dimVector, '', '', 0)
    arrayDataInt8 => int8Array2Data(outArray)
    arrayDataInt8 = image
    call release(outSet)

  end subroutine dumpImageToFitsInt8

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsInt16(image, setName)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping int16 image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)
    outArray = addArray(outSet, 'PRIMARY', INTEGER16, dimVector, '', '', 0)
    arrayDataInt16 => int16Array2Data(outArray)
    arrayDataInt16 = image
    call release(outSet)

  end subroutine dumpImageToFitsInt16

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsInt32(image, setName)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping int32 image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)
    outArray = addArray(outSet, 'PRIMARY', INTEGER32, dimVector, '', '', 0)
    arrayDataInt32 => int32Array2Data(outArray)
    arrayDataInt32 = image
    call release(outSet)

  end subroutine dumpImageToFitsInt32

  !---------------------------------------------------------------------
  subroutine dumpImageToFitsBool_temp(image, setName)
  !---------------------------------------------------------------------

    implicit none

    logical(bool), intent(in) :: image(:,:)
    character(*), intent(in) :: setName

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

print *, 'Dumping boolean image to fits file '//trim(setName)//'.'

    outSet = dataSet(setName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(image, 1)
    dimVector(1) = size(image, 2)

    outArray = addArray(outSet, 'PRIMARY', INTEGER8, dimVector, '', '', 0)
    arrayDataInt8 => int8Array2Data(outArray)
    where(image)
      arrayDataInt8=1
    elsewhere
      arrayDataInt8=0
    end where

    call release(outSet)

  end subroutine dumpImageToFitsBool_temp

!  !---------------------------------------------------------------------
!  subroutine dumpImageToFitsBoolean(image, setName)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    logical, intent(in) :: image(:,:)
!    character(*), intent(in) :: setName
!
!    type(DataSetT) :: outSet
!    type(ArrayT) :: outArray
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean
!    integer(int32), dimension(:), pointer ::&
!      dimVector
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    outSet = dataSet(setName, CREATE)
!    allocate(dimVector(0:1))
!    dimVector(0) = size(image, 1)
!    dimVector(1) = size(image, 2)
!    outArray = addArray(outSet, 'PRIMARY', BOOLEAN, dimVector, '', '', 0)
!    arrayDataBoolean => boolArray2Data(outArray)
!    arrayDataBoolean = image
!    call release(outSet)
!
!  end subroutine dumpImageToFitsBoolean

!  !---------------------------------------------------------------------
!  subroutine tableNullsDump(tab)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(TableT), intent(in) :: tab
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    numCols = numberOfColumns(tab)
!    numRows = numberOfRows(tab)
!    allocate(charArray(numRows, numCols))
!
!    do coli = 0, numCols - 1
!      call minNonNullValue(column(tab, coli), minValue, allRowsNull)
!      call maxNonNullValue(column(tab, coli), minValue, allRowsNull)
!    end do
!
!  end subroutine tableNullsDump

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dump
