!+
! Module: histogram_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module histogram_utils

  use types

  implicit none 

  private 

  public makeHistogram

  interface makeHistogram
!    module procedure makeHistogramInt8, makeHistogramInt16, makeHistogramInt32
    module procedure makeHistogramEdgesSingle, makeHistogramBoundsSingle
    module procedure makeHistogramEdgesDouble, makeHistogramBoundsDouble
  end interface

contains 

!  !---------------------------------------------------------------------
!  subroutine makeHistogramInt32(inVector, binContents&
!  , underflows, overflows, status)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    integer(int32), intent(in)            :: inVector(:)
!    integer,        pointer               :: binContents(:)
!    integer,        intent(out), optional :: underflows, overflows, status
!
!!    integer(int32) :: lEdge, rEdge
!!    integer :: hsize, i, localUnderflows, localOverflows, localStatus
!!    real(single) :: frac
!!    real(single), allocatable :: binEdges(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    localStatus = 0 ! default.
!
!    if (associated(binContents))
!
!!    hsize = size(binContents)
!!    allocate(binEdges(hsize + 1))
!!
!!    do i = 1, hsize + 1
!!      frac = (i - 1) / real(hsize)
!!      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
!!    end do
!!
!!    call makeHistogramKernel(dble(inVector), binContents, dble(binEdges)&
!!    , localUnderflows, localOverflows, localStatus)
!!
!!    if (present(underflows)) underflows = localUnderflows
!!    if (present(overflows)) overflows = localOverflows
!!    if (present(status)) status = localStatus
!
!  end subroutine makeHistogramInt32

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsSingle(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: inVector(:)
    integer,      intent(out)           :: binContents(:)
    real(single), intent(in)            :: lEdge, rEdge
    integer,      intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real(single) :: frac
    real(single), allocatable :: binEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    allocate(binEdges(hsize + 1))

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(dble(inVector), binContents, dble(binEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsSingle

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesSingle(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: inVector(:)
    integer,      intent(out)           :: binContents(:)
    real(single), intent(in),  optional :: binEdges(:)
    integer,      intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real(single) :: frac, minx, maxx
    real(single), allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(dble(inVector), binContents, dble(myBinEdges)&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesSingle

  !---------------------------------------------------------------------
  subroutine makeHistogramBoundsDouble(inVector, binContents, lEdge, rEdge&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: inVector(:)
    integer,      intent(out)           :: binContents(:)
    real(double), intent(in)            :: lEdge, rEdge
    integer,      intent(out), optional :: underflows, overflows, status

    integer :: hsize, i, localUnderflows, localOverflows, localStatus
    real(double) :: frac
    real(double), allocatable :: binEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    allocate(binEdges(hsize + 1))

    do i = 1, hsize + 1
      frac = (i - 1) / real(hsize)
      binEdges(i) = lEdge * (1.0 - frac) + rEdge * frac
    end do

    call makeHistogramKernel(inVector, binContents, binEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramBoundsDouble

  !---------------------------------------------------------------------
  subroutine makeHistogramEdgesDouble(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: inVector(:)
    integer,      intent(out)           :: binContents(:)
    real(double), intent(in),  optional :: binEdges(:)
    integer,      intent(out), optional :: underflows, overflows, status

    integer :: hsize, binEdge, localUnderflows, localOverflows, localStatus
    real(double) :: frac, minx, maxx
    real(double), allocatable :: myBinEdges(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    hsize = size(binContents)
    minx = minval(inVector)
    maxx = maxval(inVector)

    allocate(myBinEdges(hsize + 1))
    if (present(binEdges)) then
      myBinEdges = binEdges(1:hsize + 1)

    else
      do binEdge = 1, hsize + 1
        frac = (binEdge - 1) / real(hsize)
        myBinEdges(binEdge) = minx * (1.0 - frac) + maxx * frac
      end do
    end if

    call makeHistogramKernel(inVector, binContents, myBinEdges&
    , localUnderflows, localOverflows, localStatus)

    if (present(underflows)) underflows = localUnderflows
    if (present(overflows)) overflows = localOverflows
    if (present(status)) status = localStatus

  end subroutine makeHistogramEdgesDouble

  !---------------------------------------------------------------------
  subroutine makeHistogramKernel(inVector, binContents, binEdges&
  , underflows, overflows, status)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: inVector(:)
    integer,      intent(out) :: binContents(:)
    real(double), intent(in)  :: binEdges(:)
    integer,      intent(out) :: underflows, overflows, status

    integer :: vsize, hsize, i, bin
    real(double) :: minx, maxx, x

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    vsize = size(inVector)
    hsize = size(binContents)

    if (vsize .le. 1) then
      status = -1 ! only 1 value.
  return
    end if

    if (hsize .le. 1) then
      status = -2 ! only 1 bin.
  return
    end if

    if (size(binEdges) .lt. hsize + 1) then
      status = -3
  return
    else if (size(binEdges) .gt. hsize + 1) then
      status = status + 1
    end if

    do i = 1, hsize
      if (binEdges(i) .ge. binEdges(i + 1)) then
        status = -4
  return
      end if
    end do

    minx = binEdges(1)
    maxx = binEdges(hsize + 1)

    binContents = 0
    underflows = 0
    overflows = 0
    do i = 1, vsize
      x = inVector(i)
      if (x .lt. minx) then
        if (underflows .ge. huge(underflows)) then
          status = -5
  return
        end if
        underflows = underflows + 1

      else if (x .gt. maxx) then
        if (overflows .ge. huge(overflows)) then
          status = -6
  return
        end if
        overflows = overflows + 1

      else
        do bin = hsize, 1, -1
          if (x .ge. binEdges(bin)) then
            if (binContents(bin) .ge. huge(binContents)) then
              status = -7
  return
            end if
            binContents(bin) = binContents(bin) + 1
        exit
          end if
        end do
      end if
    end do

  end subroutine makeHistogramKernel

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module histogram_utils
