!
! Module: rebinners
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to rebin vectors and arrays.
! 
! Comments: TODO:
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module rebinners

  use types

  use ssc_misc		! for fracVal
  use array_utils	! for smallestTrueIndex, largestTrueIndex

  implicit none 

  private 

  public rebinImage
  public rebinVector
  public remapArray

  interface rebinImage
    module procedure rebinImageWithEdgesSingle, rebinCubeWithEdgesSingle
    module procedure rebinImageWithEdgesBool
  end interface

  interface rebinVector
    module procedure rebinVectorSingle, rebinVectorDouble, rebinVectorBool
  end interface

  integer, parameter, public :: DOAND = 1, DOOR = 2, DOXOR = 3

contains 

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesSingle(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the
  ! procedure rebinVector.

!*** this procedure has been taken from simutils/rebin.f90 and modified to use the 1d rebinner that I had already here (in lccorr) separately developed. That is in line with the probability that the lccorr version will be the one that is used in ssclib.

    implicit none

    real(single), intent(in)            :: oldImage(:,:),&
                                           oldXedges(size(oldImage,1)+1),&
                                           oldYedges(size(oldImage, 2) + 1),&
                                           newXedges(:),&
                                           newYedges(:)
    real(single), intent(out)           :: newImage(size(newXedges)-1&
                                                   ,size(newYedges)-1)
    integer,      intent(out), optional :: status

    real(single), allocatable ::&
      tempImage(:,:),&
      oldXuncerts(:), newXuncerts(:),& 
      oldYuncerts(:), newYuncerts(:)
    integer ::&
      localStatus, foreignStatus, xi, yi,&
      oldXSize, oldYSize, newXSize, newYSize
    logical(bool), allocatable :: yNulls(:), xNulls(:)

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    oldXSize = size(oldImage, 1)
    oldYSize = size(oldImage, 2)
    newXSize = size(newImage, 1)
    newYSize = size(newImage, 2)
    allocate(tempImage(oldXSize, newYSize))
    allocate(oldXuncerts(oldXSize), newXuncerts(newXSize))
    allocate(oldYuncerts(oldYSize), newYuncerts(newYSize))
    allocate(yNulls(newYSize), xNulls(newXSize))
    oldXuncerts = 0.0
    oldYuncerts = 0.0

    tempImage = 0.0 ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebinVectorSingle(oldYedges, newYedges, 'counts'&
        , oldImage(xi, :), oldYuncerts, tempImage(xi, :), newYuncerts, yNulls&
        , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = 0.0 ! array op, default
    do yi = 1, size(newImage, 2)
      call rebinVectorSingle(oldXEdges, newXEdges, 'counts'&
       , tempImage(:, yi), oldXuncerts, newImage(:, yi), newXuncerts, xNulls&
       , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesSingle

  !---------------------------------------------------------------------
  subroutine rebinCubeWithEdgesSingle(oldCube, oldXedges, oldYedges&
  , newCube, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the
  ! procedure rebinVector.

!*** this procedure has been taken from simutils/rebin.f90 and modified to use the 1d rebinner that I had already here (in lccorr) separately developed. That is in line with the probability that the lccorr version will be the one that is used in ssclib.

    implicit none

    real(single), intent(in)            :: oldCube(:,:,:),&
                                           oldXedges(size(oldCube,1)+1),&
                                           oldYedges(size(oldCube,2)+1),&
                                           newXedges(:),&
                                           newYedges(:)
    real(single), intent(out)           :: newCube(size(newXedges)-1&
                                                  ,size(newYedges)-1&
                                                  ,size(oldCube,3))
    integer,      intent(out), optional :: status

    real(single), allocatable ::&
      tempCube(:,:,:),&
      oldXuncerts(:), newXuncerts(:),& 
      oldYuncerts(:), newYuncerts(:)
    integer ::&
      localStatus, foreignStatus, xi, yi,&
      oldXSize, oldYSize, newXSize, newYSize, numPlanes
    logical(bool), allocatable :: yNulls(:), xNulls(:)

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    oldXSize  = size(oldCube, 1)
    oldYSize  = size(oldCube, 2)
    numPlanes = size(oldCube, 3)
    newXSize = size(newCube, 1)
    newYSize = size(newCube, 2)
    allocate(tempCube(oldXSize, newYSize, numPlanes))
    allocate(oldXuncerts(oldXSize), newXuncerts(newXSize))
    allocate(oldYuncerts(oldYSize), newYuncerts(newYSize))
    allocate(yNulls(newYSize), xNulls(newXSize))
    oldXuncerts = 0.0
    oldYuncerts = 0.0

    tempCube = 0.0 ! array op, default
    do xi = 1, size(oldCube, 1)
      call rebinSliceSingle(oldYedges, newYedges, 'counts'&
        , oldCube(xi,:,:), oldYuncerts, tempCube(xi,:,:), newYuncerts, yNulls&
        , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newCube = 0.0 ! array op, default
    do yi = 1, size(newCube, 2)
      call rebinSliceSingle(oldXEdges, newXEdges, 'counts'&
       , tempCube(:,yi,:), oldXuncerts, newCube(:,yi,:), newXuncerts, xNulls&
       , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinCubeWithEdgesSingle

  !---------------------------------------------------------------------
  subroutine rebinImageWithEdgesBool(oldImage, oldXedges, oldYedges&
  , newImage, newXedges, newYedges, status)
  !---------------------------------------------------------------------
  ! The rebinning is done in two passes, x first then y, using the
  ! procedure rebinVector.

!*** this procedure has been taken from simutils/rebin.f90 and modified to use the 1d rebinner that I had already here (in lccorr) separately developed. That is in line with the probability that the lccorr version will be the one that is used in ssclib.

    implicit none

    logical(bool), intent(in)            :: oldImage(:,:)
    real(single),  intent(in)            :: oldXedges(size(oldImage,1)+1),&
                                            oldYedges(size(oldImage, 2) + 1),&
                                            newXedges(:),&
                                            newYedges(:)
    logical(bool), intent(out)           :: newImage(size(newXedges)-1&
                                                    ,size(newYedges)-1)
    integer,       intent(out), optional :: status

    logical(bool), allocatable :: tempImage(:,:), yNulls(:), xNulls(:)
    integer :: localStatus, foreignStatus, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    allocate(tempImage(size(oldImage, 1), size(newImage, 2)))
    allocate(yNulls(size(newImage, 2)), xNulls(size(newImage, 1)))

    tempImage = .false. ! array op, default
    do xi = 1, size(oldImage, 1)
      call rebinVectorBool(oldYedges, newYedges&
        , oldImage(xi, :), tempImage(xi, :), yNulls&
        , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 0)
!***
      end if
    end do

    newImage = .false. ! array op, default
    do yi = 1, size(newImage, 2)
      call rebinVectorBool(oldXEdges, newXEdges&
       , tempImage(:, yi), newImage(:, yi), xNulls&
       , foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus
  return

      else if (foreignStatus > 0) then
        localStatus = ibset(localStatus, 1)
!***
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine rebinImageWithEdgesBool

  !---------------------------------------------------------------------
  subroutine rebinVectorSingle(oldBinEdges, newBinEdges, style&
  , oldY, oldYuncerts, newY, newYuncerts, nulls, status)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in)  :: oldBinEdges(:),&
                                 newBinEdges(:)
    character(*), intent(in)  :: style
    real(single), intent(in)  :: oldY(size(oldBinEdges) - 1),&
                                 oldYuncerts(size(oldBinEdges) - 1)
    real(single), intent(out) :: newY(size(newBinEdges) - 1),&
                                 newYuncerts(size(newBinEdges) - 1)
    logical(bool),intent(out) :: nulls(size(newBinEdges) - 1)
    integer,      intent(out) :: status

    real(double), allocatable ::&
      oldBinSizes(:), newBinSizes(:),&
      tempOldY(:), tempOldYuncerts(:),&
      tempNewY(:), tempNewYuncerts(:)
    integer :: numOldEdges, numNewEdges
    real(double), allocatable :: newYDble(:), newYuncertsDble(:)

    !-------------------------------------------------------------------
    ! Start:

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)

    allocate(oldBinSizes(numOldEdges-1), newBinSizes(numNewEdges-1))
    allocate(tempOldY(numOldEdges-1), tempOldYuncerts(numOldEdges-1))
    allocate(tempNewY(numNewEdges-1), tempNewYuncerts(numNewEdges-1))
    allocate(newYDble(numNewEdges-1), newYuncertsDble(numNewEdges-1))

    oldBinSizes = oldBinEdges(2:) - oldBinEdges(:numOldEdges-1)
    newBinSizes = newBinEdges(2:) - newBinEdges(:numNewEdges-1)

    if (style .eq. 'counts') then
      call rebinVectorKernel(dble(oldBinEdges), oldBinSizes, dble(newBinEdges)&
        , newBinSizes, dble(oldY), dble(oldYuncerts), newYDble&
        , newYuncertsDble, nulls, status)

      newY        = newYDble
      newYuncerts = newYuncertsDble

    else ! style .eq. 'rates'
      tempOldY        = oldY        * oldBinSizes(:numOldEdges-1)
      tempOldYuncerts = oldYuncerts * oldBinSizes(:numOldEdges-1)

      call rebinVectorKernel(dble(oldBinEdges), oldBinSizes, dble(newBinEdges)&
        , newBinSizes, tempOldY, tempOldYuncerts, tempNewY&
        , tempNewYuncerts, nulls, status)

      newY        = tempNewY        / newBinSizes(:numNewEdges-1)
      newYuncerts = tempNewYuncerts / newBinSizes(:numNewEdges-1)
    end if

  end subroutine rebinVectorSingle

  !---------------------------------------------------------------------
  subroutine rebinSliceSingle(oldBinEdges, newBinEdges, style&
  , oldY, oldYuncerts, newY, newYuncerts, nulls, status)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in)  :: oldY(:,:),&
                                 oldYuncerts(size(oldY,1),size(oldY,2))
    real(single), intent(in)  :: oldBinEdges(size(oldY,1)+1),&
                                 newBinEdges(:)
    real(single), intent(out) :: newY(size(newBinEdges) - 1,size(oldY,2)),&
                                 newYuncerts(size(newBinEdges) - 1,size(oldY,2))
    character(*), intent(in)  :: style
    logical(bool),intent(out) :: nulls(size(newBinEdges) - 1)
    integer,      intent(out) :: status

    real(double), allocatable ::&
      oldBinSizes(:), newBinSizes(:),&
      tempOldY(:,:), tempOldYuncerts(:,:),&
      tempNewY(:,:), tempNewYuncerts(:,:)
    integer :: numOldEdges, numNewEdges, numPlanes
    real(double), allocatable :: newYDble(:,:), newYuncertsDble(:,:)

    !-------------------------------------------------------------------
    ! Start:

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)
    numPlanes = size(oldY,2)

    allocate(oldBinSizes(numOldEdges-1), newBinSizes(numNewEdges-1))
    allocate(tempOldY(numOldEdges-1,numPlanes), tempOldYuncerts(numOldEdges-1,numPlanes))
    allocate(tempNewY(numNewEdges-1,numPlanes), tempNewYuncerts(numNewEdges-1,numPlanes))
    allocate(newYDble(numNewEdges-1,numPlanes), newYuncertsDble(numNewEdges-1,numPlanes))

    oldBinSizes = oldBinEdges(2:) - oldBinEdges(:numOldEdges-1)
    newBinSizes = newBinEdges(2:) - newBinEdges(:numNewEdges-1)

    if (style .eq. 'counts') then
      call rebinSliceKernel(dble(oldBinEdges), oldBinSizes, dble(newBinEdges)&
        , newBinSizes, dble(oldY), dble(oldYuncerts), newYDble&
        , newYuncertsDble, nulls, status)

      newY        = newYDble
      newYuncerts = newYuncertsDble

    else ! style .eq. 'rates'
      tempOldY        = oldY        * spread(oldBinSizes(:numOldEdges-1),2,numPlanes)
      tempOldYuncerts = oldYuncerts * spread(oldBinSizes(:numOldEdges-1),2,numPlanes)

      call rebinSliceKernel(dble(oldBinEdges), oldBinSizes, dble(newBinEdges)&
        , newBinSizes, tempOldY, tempOldYuncerts, tempNewY&
        , tempNewYuncerts, nulls, status)

      newY        = tempNewY        / spread(newBinSizes(:numNewEdges-1),2,numPlanes)
      newYuncerts = tempNewYuncerts / spread(newBinSizes(:numNewEdges-1),2,numPlanes)
    end if

  end subroutine rebinSliceSingle

  !---------------------------------------------------------------------
  subroutine rebinVectorDouble(oldBinEdges, newBinEdges, style&
  , oldY, oldYuncerts, newY, newYuncerts, nulls, status)
  !---------------------------------------------------------------------

    implicit none 

    real(double), intent(in)  :: oldBinEdges(:),&
                                 newBinEdges(:)
    character(*), intent(in)  :: style
    real(double), intent(in)  :: oldY(size(oldBinEdges) - 1),&
                                 oldYuncerts(size(oldBinEdges) - 1)
    real(double), intent(out) :: newY(size(newBinEdges) - 1),&
                                 newYuncerts(size(newBinEdges) - 1)
    logical(bool),intent(out) :: nulls(size(newBinEdges) - 1)
    integer,      intent(out) :: status

    real(double), allocatable ::&
      oldBinSizes(:), newBinSizes(:),&
      tempOldY(:), tempOldYuncerts(:),&
      tempNewY(:), tempNewYuncerts(:)
    integer :: numOldEdges, numNewEdges

    !-------------------------------------------------------------------
    ! Start:

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)

    allocate(oldBinSizes(numOldEdges-1), newBinSizes(numNewEdges-1))
    allocate(tempOldY(numOldEdges-1), tempOldYuncerts(numOldEdges-1))
    allocate(tempNewY(numNewEdges-1), tempNewYuncerts(numNewEdges-1))

    oldBinSizes = oldBinEdges(2:) - oldBinEdges(:numOldEdges-1)
    newBinSizes = newBinEdges(2:) - newBinEdges(:numNewEdges-1)

    if (style .eq. 'counts') then
      call rebinVectorKernel(oldBinEdges, oldBinSizes, newBinEdges, newBinSizes&
        , oldY, oldYuncerts, newY, newYuncerts, nulls, status)

    else ! style .eq. 'rates'
      tempOldY        = oldY        * oldBinSizes(:numOldEdges-1)
      tempOldYuncerts = oldYuncerts * oldBinSizes(:numOldEdges-1)

      call rebinVectorKernel(oldBinEdges, oldBinSizes, newBinEdges, newBinSizes&
        , tempOldY, tempOldYuncerts, tempNewY, tempNewYuncerts, nulls, status)

      newY        = tempNewY        / newBinSizes(:numNewEdges-1)
      newYuncerts = tempNewYuncerts / newBinSizes(:numNewEdges-1)
    end if

  end subroutine rebinVectorDouble

  !---------------------------------------------------------------------
  subroutine rebinVectorBool(oldBinEdges, newBinEdges&
  , oldY, newY, nulls, status)
  !---------------------------------------------------------------------

    implicit none 

    real(single),  intent(in)  :: oldBinEdges(:),&
                                  newBinEdges(:)
    logical(bool), intent(in)  :: oldY(size(oldBinEdges) - 1)
    logical(bool), intent(out) :: newY(size(newBinEdges) - 1),&
                                  nulls(size(newBinEdges) - 1)
    integer,       intent(out) :: status

    real(double), allocatable :: oldBinSizes(:), newBinSizes(:)
    integer :: numOldEdges, numNewEdges

    !-------------------------------------------------------------------
    ! Start:

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)

    allocate(oldBinSizes(numOldEdges-1), newBinSizes(numNewEdges-1))

    oldBinSizes = oldBinEdges(2:) - oldBinEdges(:numOldEdges-1)
    newBinSizes = newBinEdges(2:) - newBinEdges(:numNewEdges-1)

    call rebinVectorBoolKernel(dble(oldBinEdges), oldBinSizes&
      , dble(newBinEdges), newBinSizes, oldY, DOAND, newY, nulls, status)

  end subroutine rebinVectorBool

  !---------------------------------------------------------------------
  subroutine rebinVectorKernel(oldBinEdges, oldBinSizes, newBinEdges&
  , newBinSizes, oldCounts, oldUncerts, newCounts, newUncerts, nulls, status)
  !---------------------------------------------------------------------
  ! This subroutine is designed to handle data which is 'binned'. For
  ! binned data, the x axis is divided into N bins, each of which has a
  ! y value (and an uncertainty) associated with it. The bins here are
  ! defined by the vector oldBinEdges, of size N+1, and the y and
  ! uncertainty values are contained respectively in vectors oldCounts
  ! and oldUncerts, each of size N. Given a new set of bin boundaries
  ! defined in newBinEdges, the subroutine returns newCounts and newUncerts
  ! with the appropriate rebinned y and uncertainty values.
  !
  ! Rebinning is complicated by the fact that there are two ways of
  ! considering the y values, which are here designated 'count' style
  ! and 'rate' style. In the 'count' style, a smaller 'new' bin would
  ! be expected to have proportionately smaller y value; in 'rate' style,
  ! the new y value should be independent of bin size. 'Rate' style is
  ! perhaps a more natural way to think of rebinning, but 'count' style
  ! has been chosen as the fundamental style in the present subroutine
  ! because the calculation of the rebinned uncertainties is simpler
  ! and clearer in this style. 
  !
  ! To illustrate the 'count' style, consider the following diagram
  ! showing an example of old vs new bins:
  !
  ! X:	0                   5                                        15
  !	|...................|........................................|
  ! Y:	          10                            20
  ! unc:        3.1623 (root(10))             4.4721 (root(20))
  !
  ! X:	                4               8
  !	                |...............|
  !
  ! Clearly the output y value is
  !
  !	              (5 - 4)           (8 - 5)
  !	y_new = 10 * --------- + 20 * ----------
  !	              (5 - 0)          (15 - 5)
  !
  !	      = 8;
  !
  ! The transformation of uncertainties is best understood by considering
  ! rather the variance = square of the uncertainty: this transforms in
  ! an identical fashion to the y values; hence
  !
  !	              (             (5 - 4)                 (8 - 5)  )
  !	unc_new = sqrt( 3.1623^2 * --------- + 4.4721^2 * ---------- )
  !	              (             (5 - 0)                (15 - 5)  )
  !
  !	        = 2.8284 (root(8)).
  !
  !
  ! 'Rate' style is illustrated as follows:
  !
  ! X:	0                   5                                        15
  !	|...................|........................................|
  ! Y:	          2.0                           2.0
  ! unc:        0.6325                        0.4472
  !
  ! X:	                4               8
  !	                |...............|
  !
  ! Converting these values to counts and counts-uncertainties by
  ! multiplying by the bin width yields y_new = 2.0 +- 0.7071. The value
  ! has remained the same but the uncertainty is larger, reflecting the
  ! smaller bin size.
  !---------------------------------------------------------------------

    implicit none 

    real(double),  intent(in)  :: oldBinEdges(:),&
                                  newBinEdges(:)
    real(double),  intent(in)  :: oldBinSizes(&
                                    size(oldBinEdges)-1),&
                                  newBinSizes(&
                                    size(newBinEdges)-1),&
                                  oldCounts(&
                                    size(oldBinEdges) - 1),&
                                  oldUncerts(&
                                    size(oldBinEdges) - 1)
    real(double),  intent(out) :: newCounts(&
                                    size(newBinEdges) - 1),&
                                  newUncerts(&
                                    size(newBinEdges) - 1)
    logical(bool), intent(out) :: nulls(size(newBinEdges) - 1)
    integer,       intent(out) :: status


    real(double), allocatable :: oldVariance(:), newVariance(:)
    integer ::&
      numOldEdges, numNewEdges, locVec(1), startNewBin, finisNewBin,&
      startOldBin, oldBin, newBin
    real(double) :: weight

    !-------------------------------------------------------------------
    ! Start:

    ! defaults:
    status = 0
    newCounts   = 0.0
    nulls = .true.

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)

    allocate(oldVariance(numOldEdges-1), newVariance(numNewEdges-1))
    newVariance = 0.0

    if (any(oldBinSizes(:numOldEdges - 1) <= 0.0)) then
      status = -1
  return
    end if

    if (any(newBinSizes(:numNewEdges - 1) <= 0.0)) then
      status = -2
  return
    end if

    if (any(oldUncerts < 0.0)) then
      status = -3
  return
    end if

    if (any(newBinEdges < oldBinEdges(1))) then
      locVec = maxloc(newBinEdges, (newBinEdges < oldBinEdges(1)))
      startNewBin = max(1, locVec(1) + 1)
    else
      startNewBin = 1
    end if

    if (any(newBinEdges > oldBinEdges(numOldEdges))) then
      locVec = minloc(newBinEdges&
        , (newBinEdges > oldBinEdges(numOldEdges)))
      finisNewBin = min(numNewEdges - 1, locVec(1) - 2)
    else
      finisNewBin = numNewEdges - 1
    end if

    if (startNewBin > finisNewBin) then
      status = ibset(status, 0)
  return

    else if (startNewBin > 1 .or. finisNewBin < numNewEdges - 1) then
      status = ibset(status, 1)
    end if

    nulls(startNewBin:finisNewBin) = .false.

    locVec = minloc(oldBinEdges&
      , (oldBinEdges > newBinEdges(startNewBin)))
    startOldBin = locVec(1) - 1
    if (startOldBin < 1) then
      status = -4
  return
    end if

    oldVariance = oldUncerts * oldUncerts

    oldBin = startOldBin
    NEW: do newBin = startNewBin, finisNewBin
      OLD: do
        if (oldBinEdges(oldBin) <= newBinEdges(newBin)) then
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) .eq. newBinEdges(newBin+1)) then
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
            oldBin = oldBin + 1
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
            weight = (newBinEdges(newBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
    cycle NEW

          end if
        else ! oldBinEdges(oldBin) > newBinEdges(newBin)
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
            weight = 1.0
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) .eq. newBinEdges(newBin+1)) then
            weight = 1.0
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
            oldBin = oldBin + 1
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
            weight = (newBinEdges(newBin+1) - oldBinEdges(oldBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin) = newCounts(newBin) + weight * oldCounts(oldBin)
            newVariance(newBin) = newVariance(newBin) + weight&
              * oldVariance(oldBin)
    cycle NEW

          end if
        end if
      end do OLD
    end do NEW

    newUncerts = sqrt(newVariance)

  end subroutine rebinVectorKernel

  !---------------------------------------------------------------------
  subroutine rebinSliceKernel(oldBinEdges, oldBinSizes, newBinEdges&
  , newBinSizes, oldCounts, oldUncerts, newCounts, newUncerts, nulls, status)

!**** this should take over from rebinVectorKernel since it is the more general routine.

  !---------------------------------------------------------------------
  ! This subroutine is designed to handle data which is 'binned'. For
  ! binned data, the x axis is divided into N bins, each of which has a
  ! y value (and an uncertainty) associated with it. The bins here are
  ! defined by the vector oldBinEdges, of size N+1, and the y and
  ! uncertainty values are contained respectively in vectors oldCounts
  ! and oldUncerts, each of size N. Given a new set of bin boundaries
  ! defined in newBinEdges, the subroutine returns newCounts and newUncerts
  ! with the appropriate rebinned y and uncertainty values.
  !
  ! Rebinning is complicated by the fact that there are two ways of
  ! considering the y values, which are here designated 'count' style
  ! and 'rate' style. In the 'count' style, a smaller 'new' bin would
  ! be expected to have proportionately smaller y value; in 'rate' style,
  ! the new y value should be independent of bin size. 'Rate' style is
  ! perhaps a more natural way to think of rebinning, but 'count' style
  ! has been chosen as the fundamental style in the present subroutine
  ! because the calculation of the rebinned uncertainties is simpler
  ! and clearer in this style. 
  !
  ! To illustrate the 'count' style, consider the following diagram
  ! showing an example of old vs new bins:
  !
  ! X:	0                   5                                        15
  !	|...................|........................................|
  ! Y:	          10                            20
  ! unc:        3.1623 (root(10))             4.4721 (root(20))
  !
  ! X:	                4               8
  !	                |...............|
  !
  ! Clearly the output y value is
  !
  !	              (5 - 4)           (8 - 5)
  !	y_new = 10 * --------- + 20 * ----------
  !	              (5 - 0)          (15 - 5)
  !
  !	      = 8;
  !
  ! The transformation of uncertainties is best understood by considering
  ! rather the variance = square of the uncertainty: this transforms in
  ! an identical fashion to the y values; hence
  !
  !	              (             (5 - 4)                 (8 - 5)  )
  !	unc_new = sqrt( 3.1623^2 * --------- + 4.4721^2 * ---------- )
  !	              (             (5 - 0)                (15 - 5)  )
  !
  !	        = 2.8284 (root(8)).
  !
  !
  ! 'Rate' style is illustrated as follows:
  !
  ! X:	0                   5                                        15
  !	|...................|........................................|
  ! Y:	          2.0                           2.0
  ! unc:        0.6325                        0.4472
  !
  ! X:	                4               8
  !	                |...............|
  !
  ! Converting these values to counts and counts-uncertainties by
  ! multiplying by the bin width yields y_new = 2.0 +- 0.7071. The value
  ! has remained the same but the uncertainty is larger, reflecting the
  ! smaller bin size.
  !---------------------------------------------------------------------

    implicit none 

    real(double),  intent(in)  :: oldCounts(:,:),&
                                  oldUncerts(&
                                    size(oldCounts,1),size(oldCounts,2))
    real(double),  intent(in)  :: oldBinEdges(size(oldCounts,1)+1),&
                                  newBinEdges(:)
    real(double),  intent(in)  :: oldBinSizes(&
                                    size(oldCounts,1)),&
                                  newBinSizes(&
                                    size(newBinEdges)-1)
    real(double),  intent(out) :: newCounts(&
                                    size(newBinEdges) - 1,size(oldCounts,2)),&
                                  newUncerts(&
                                    size(newBinEdges) - 1,size(oldCounts,2))
    logical(bool), intent(out) :: nulls(size(newBinEdges) - 1)
    integer,       intent(out) :: status


    real(double), allocatable :: oldVariance(:,:), newVariance(:,:)
    integer ::&
      numOldEdges, numNewEdges, locVec(1), startNewBin, finisNewBin,&
      startOldBin, oldBin, newBin, numPlanes
    real(double) :: weight

    !-------------------------------------------------------------------
    ! Start:

    ! defaults:
    status = 0
    newCounts   = 0.0
    nulls = .true.

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)
    numPlanes = size(oldCounts,2)

    allocate(oldVariance(numOldEdges-1,numPlanes), newVariance(numNewEdges-1,numPlanes))
    newVariance = 0.0

    if (any(oldBinSizes(:numOldEdges - 1) <= 0.0)) then
      status = -1
  return
    end if

    if (any(newBinSizes(:numNewEdges - 1) <= 0.0)) then
      status = -2
  return
    end if

    if (any(oldUncerts < 0.0)) then
      status = -3
  return
    end if

    if (any(newBinEdges < oldBinEdges(1))) then
      locVec = maxloc(newBinEdges, (newBinEdges < oldBinEdges(1)))
      startNewBin = max(1, locVec(1) + 1)
    else
      startNewBin = 1
    end if

    if (any(newBinEdges > oldBinEdges(numOldEdges))) then
      locVec = minloc(newBinEdges&
        , (newBinEdges > oldBinEdges(numOldEdges)))
      finisNewBin = min(numNewEdges - 1, locVec(1) - 2)
    else
      finisNewBin = numNewEdges - 1
    end if

    if (startNewBin > finisNewBin) then
      status = ibset(status, 0)
  return

    else if (startNewBin > 1 .or. finisNewBin < numNewEdges - 1) then
      status = ibset(status, 1)
    end if

    nulls(startNewBin:finisNewBin) = .false.

    locVec = minloc(oldBinEdges&
      , (oldBinEdges > newBinEdges(startNewBin)))
    startOldBin = locVec(1) - 1
    if (startOldBin < 1) then
      status = -4
  return
    end if

    oldVariance = oldUncerts * oldUncerts

    oldBin = startOldBin
    NEW: do newBin = startNewBin, finisNewBin
      OLD: do
        if (oldBinEdges(oldBin) <= newBinEdges(newBin)) then
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) .eq. newBinEdges(newBin+1)) then
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
            oldBin = oldBin + 1
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
            weight = (newBinEdges(newBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
    cycle NEW

          end if
        else ! oldBinEdges(oldBin) > newBinEdges(newBin)
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
            weight = 1.0
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) .eq. newBinEdges(newBin+1)) then
            weight = 1.0
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
            oldBin = oldBin + 1
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
            weight = (newBinEdges(newBin+1) - oldBinEdges(oldBin))&
              / oldBinSizes(oldBin)
            newCounts(newBin,:) = newCounts(newBin,:) + weight * oldCounts(oldBin,:)
            newVariance(newBin,:) = newVariance(newBin,:) + weight&
              * oldVariance(oldBin,:)
    cycle NEW

          end if
        end if
      end do OLD
    end do NEW

    newUncerts = sqrt(newVariance)

  end subroutine rebinSliceKernel

  !---------------------------------------------------------------------
  subroutine rebinVectorBoolKernel(oldBinEdges, oldBinSizes, newBinEdges&
  , newBinSizes, oldMask, logicalOperation, newMask, nulls, status)
  !---------------------------------------------------------------------
  ! This subroutine is designed to handle boolean data which is 'binned'. For
  ! binned data, the x axis is divided into N bins, each of which has a
  ! boolean y value associated with it. The bins here are defined by the vector oldBinEdges, of size N+1, and the y values are contained in the vector oldMask of size N. Given a new set of bin boundaries defined in newBinEdges, the subroutine returns newMask which contains the appropriate rebinned boolean y.
  !
  ! Rebinning is complicated by the fact that there are several ways of
  ! combining boolean values. There may be more than 1 old bin overlapping each new one. The method for combining these overlapping old booleans into the new is dictated by the integer logicalOperation, which may take values DOAND, DOOR or DOXOR. (Actually the XOR is not a true XOR, rather it sets true those new bins for which some but not all of the overlapping old bins were true.)
  !---------------------------------------------------------------------

    implicit none 

    real(double),  intent(in)  :: oldBinEdges(:), newBinEdges(:)
    real(double),  intent(in)  :: oldBinSizes(size(oldBinEdges)),&
                                  newBinSizes(size(newBinEdges))
    logical(bool), intent(in)  :: oldMask(size(oldBinEdges) - 1)
    integer,       intent(in)  :: logicalOperation
    logical(bool), intent(out) :: newMask(size(newBinEdges) - 1)
    logical(bool), intent(out) :: nulls(size(newBinEdges) - 1)
    integer,       intent(out) :: status

    integer ::&
      numOldEdges, numNewEdges, startNewBin, finisNewBin,&
      startOldBin, oldBin, newBin, numOldTotal, numOldTrue
!, locVec(1)
    real(single) :: weight

    !-------------------------------------------------------------------
    ! Start:

    ! defaults:
    status = 0
    newMask   = .false.
!    newVariance = 0.0
    nulls = .true.

    numOldEdges = size(oldBinEdges)
    numNewEdges = size(newBinEdges)

    if (any(oldBinSizes(:numOldEdges - 1) <= 0.0)) then
      status = -1
  return
    end if

    if (any(newBinSizes(:numNewEdges - 1) <= 0.0)) then
      status = -2
  return
    end if

!    if (any(oldUncerts < 0.0)) then
!      status = -3
!  return
!    end if

    if (    logicalOperation .ne. DOAND&
      .and. logicalOperation .ne. DOOR&
      .and. logicalOperation .ne. DOXOR) then

      status = -3
  return
    end if

    if (any(newBinEdges < oldBinEdges(1))) then
!      locVec = maxloc(newBinEdges, (newBinEdges < oldBinEdges(1)))
!      startNewBin = max(1, locVec(1) + 1)
startNewBin = smallestTrueIndex(logical(newBinEdges < oldBinEdges(1), bool))
startNewBin = max(1, startNewBin + 1)
    else
      startNewBin = 1
    end if

    if (any(newBinEdges > oldBinEdges(numOldEdges))) then
!      locVec = minloc(newBinEdges&
!        , (newBinEdges > oldBinEdges(numOldEdges)))
!      finisNewBin = min(numNewEdges - 1, locVec(1) - 2)
finisNewBin = largestTrueIndex(logical(newBinEdges > oldBinEdges(numOldEdges), bool))
finisNewBin = min(numNewEdges - 1, finisNewBin - 2)
    else
      finisNewBin = numNewEdges - 1
    end if

    if (startNewBin > finisNewBin) then
      status = ibset(status, 0)
  return

    else if (startNewBin > 1 .or. finisNewBin < numNewEdges - 1) then
      status = ibset(status, 1)
    end if

    nulls(startNewBin:finisNewBin) = .false.

!    locVec = minloc(oldBinEdges&
!      , (oldBinEdges > newBinEdges(startNewBin)))
!    startOldBin = locVec(1) - 1
startOldBin = smallestTrueIndex(logical(oldBinEdges > newBinEdges(startNewBin), bool)) - 1
    if (startOldBin < 1) then
      status = -4
  return
    end if

!    oldVariance = oldUncerts * oldUncerts

    oldBin = startOldBin
    NEW: do newBin = startNewBin, finisNewBin
      numOldTotal = 0
      numOldTrue  = 0
      OLD: do
        if (oldBinEdges(oldBin) <= newBinEdges(newBin)) then
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
! O:  |------|       or      |-----|
! N:    |------|             |-------|
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) == newBinEdges(newBin+1)) then
! O:  |--------|     or      |-------|
! N:    |------|             |-------|
            weight = (oldBinEdges(oldBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            oldBin = oldBin + 1
            if (numOldTotal > 0) then
              select case(logicalOperation)
              case(DOAND)
                if (numOldTrue == numOldTotal) newMask(newBin) = .true.
              case(DOOR)
                if (numOldTrue > 0) newMask(newBin) = .true.
              case(DOXOR)
                if (numOldTrue > 0 .and. numOldTrue < numOldTotal)&
                  newMask(newBin) = .true.
              end select
            end if
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
! O:  |----------|     or    |---------|
! N:    |------|             |-------|
            weight = (newBinEdges(newBin+1) - newBinEdges(newBin))&
              / oldBinSizes(oldBin)
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            if (numOldTotal > 0) then
              select case(logicalOperation)
              case(DOAND)
                if (numOldTrue == numOldTotal) newMask(newBin) = .true.
              case(DOOR)
                if (numOldTrue > 0) newMask(newBin) = .true.
              case(DOXOR)
                if (numOldTrue > 0 .and. numOldTrue < numOldTotal)&
                  newMask(newBin) = .true.
              end select
            end if
    cycle NEW
          end if

        else ! oldBinEdges(oldBin) > newBinEdges(newBin)
          if (oldBinEdges(oldBin+1) < newBinEdges(newBin+1)) then
! O:    |-----|
! N:  |---------|
            weight = 1.0
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            oldBin = oldBin + 1

          else if (oldBinEdges(oldBin+1) == newBinEdges(newBin+1)) then
! O:    |-------|
! N:  |---------|
            weight = 1.0
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            oldBin = oldBin + 1
            if (numOldTotal > 0) then
              select case(logicalOperation)
              case(DOAND)
                if (numOldTrue == numOldTotal) newMask(newBin) = .true.
              case(DOOR)
                if (numOldTrue > 0) newMask(newBin) = .true.
              case(DOXOR)
                if (numOldTrue > 0 .and. numOldTrue < numOldTotal)&
                  newMask(newBin) = .true.
              end select
            end if
    cycle NEW

          else ! oldBinEdges(oldBin+1) > newBinEdges(newBin+1)
! O:    |---------|
! N:  |---------|
            weight = (newBinEdges(newBin+1) - oldBinEdges(oldBin))&
              / oldBinSizes(oldBin)
            if (weight > 0.0) then
              numOldTotal = numOldTotal + 1
              if (oldMask(oldBin)) numOldTrue = numOldTrue + 1
            end if
            if (numOldTotal > 0) then
              select case(logicalOperation)
              case(DOAND)
                if (numOldTrue == numOldTotal) newMask(newBin) = .true.
              case(DOOR)
                if (numOldTrue > 0) newMask(newBin) = .true.
              case(DOXOR)
                if (numOldTrue > 0 .and. numOldTrue < numOldTotal)&
                  newMask(newBin) = .true.
              end select
            end if
    cycle NEW

          end if
        end if
      end do OLD
    end do NEW

!    newUncerts = sqrt(newVariance)

  end subroutine rebinVectorBoolKernel

  !---------------------------------------------------------------------
  subroutine remapArray(oldArray, minNewXi, maxNewXi, minNewYi, maxNewYi&
  , oldCentreX, oldCentreY, oldPixelSizeX, oldPixelSizeY&
  , newCentreX, newCentreY, newPixelSizeX, newPixelSizeY, newArray)
  !---------------------------------------------------------------------
  ! This takes a rectangular array (both of whose indices are supposed to start at 1) and remaps it to another parallel, rectilinear pixellization (defined by different reference pixel and pixel sizes). Only that part of the new array that overlaps the bounds given in minNewXi etc is returned. If there is no overlap at all, this is signified by setting the size in both dimensions of newArray() to zero.
  !
  ! Note that the pointer array newArray() should be deallocated after use.

    implicit none

    real(single), intent(in) ::&
      oldArray(:,:),&
      oldCentreX, oldCentreY, newCentreX, newCentreY,&
      oldPixelSizeX, oldPixelSizeY, newPixelSizeX, newPixelSizeY
    integer, intent(in) ::&
      minNewXi, maxNewXi, minNewYi, maxNewYi
    real(single), pointer ::&
      newArray(:,:)

    integer :: oldXsize, newStartXi, newFinisXi, newXsize, oldYsize, newStartYi, newFinisYi, newYsize
    real(single) :: oldLoX, oldHiX, newLoX, newHiX, oldLoY, oldHiY, newLoY, newHiY
    real(single), allocatable ::&
      oldXedges(:), oldYedges(:), newXedges(:), newYedges(:)
    integer :: status, i
!    integer, pointer :: indices(:)

    !-------------------------------------------------------------------
    ! Start

!*** note that similar algorithms are found in subroutine resizePatch() in simutils/temp_simulations.f90 and also in interpolatePsfCubeToCcd() in lccorr/lccorr_corrector.f90.

    oldXsize = size(oldArray, 1)
    oldLoX = 0.5
    oldHiX = 0.5 + real(oldXsize)
    newLoX = newCentreX + (oldLoX - oldCentreX) * newPixelSizeX / oldPixelSizeX
    newHiX = newCentreX + (oldHiX - oldCentreX) * newPixelSizeX / oldPixelSizeX
    newStartXi = max(minNewXi, floor(newLoX))
    newFinisXi = min(maxNewXi, floor(newHiX))
    newXsize = 1 + newFinisXi - newStartXi

    oldYsize = size(oldArray, 2)
    oldLoY = 0.5
    oldHiY = 0.5 + real(oldYsize)
    newLoY = newCentreY + (oldLoY - oldCentreY) * newPixelSizeY / oldPixelSizeY
    newHiY = newCentreY + (oldHiY - oldCentreY) * newPixelSizeY / oldPixelSizeY
    newStartYi = max(minNewYi, floor(newLoY))
    newFinisYi = min(maxNewYi, floor(newHiY))
    newYsize = 1 + newFinisYi - newStartYi

    if (newStartXi > newFinisXi .or. newStartYi > newFinisYi) then
      allocate(newArray(0,0))
  return
    end if

    allocate(newArray(newStartXi:newFinisXi,newStartYi:newFinisYi))
    allocate(oldXedges(oldXsize+1), oldYedges(oldYsize+1))
    allocate(newXedges(newXsize+1), newYedges(newYsize+1))

!    indices => indicize(0, oldXsize) ! in ssclib/ssc_misc
!    oldXedges = indices / real(oldXsize)
oldXedges = (/(i/real(oldXsize), i=0,oldXsize)/)
    oldXedges = fracVal(oldLoX, oldHiX, oldXedges)
!    deallocate(indices)

!    indices => indicize(0, oldYsize) ! in ssclib/ssc_misc
!    oldYedges = indices / real(oldYsize)
oldYedges = (/(i/real(oldYsize), i=0,oldYsize)/)
    oldYedges = fracVal(oldLoY, oldHiY, oldYedges)
!    deallocate(indices)

!    indices => indicize(newStartXi, newFinisXi + 1) ! in ssclib/ssc_misc
!    newXedges = real(indices) - 0.5
newXedges = (/(real(i)-0.5, i=newStartXi, newFinisXi + 1)/)
!*** note that interpolatePsfCubeToCcd doesn't have the 0.5!
!    deallocate(indices)

!    indices => indicize(newStartYi, newFinisYi + 1) ! in ssclib/ssc_misc
!    newYedges = real(indices) - 0.5
newYedges = (/(real(i)-0.5, i=newStartYi, newFinisYi + 1)/)
!*** note that interpolatePsfCubeToCcd doesn't have the 0.5!
!    deallocate(indices)

    call rebinImage(oldArray, oldXedges, oldYedges&
      , newArray, newXedges, newYedges, status)

    if (status < 0) then
print *, status
call error('rebinStatus','')
    end if

  end subroutine remapArray

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module rebinners
