!
! Module: save_image
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines for saving old and new images.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: save_image.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module save_image

  use dal
  use dss

  use dss_aux
  use wcs_aux
!use dal_aux

  implicit none 

  private 

  public replaceImage
  public saveImage

  interface replaceImage
    module procedure replaceImageSingle, replaceImageDouble, replaceImageBool
    module procedure replaceImageInt8,   replaceImageInt16,  replaceImageInt32
  end interface

  interface saveImage
    module procedure saveImageSingle, saveImageDouble, saveImageBool
    module procedure saveImageInt8,   saveImageInt16,  saveImageInt32
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine replaceImageSingle(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

    implicit none

    real(single),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    real(single), pointer ::&
      arrayDataReal32(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL32&
      , set, newImageArray)

    arrayDataReal32 => real32Array2Data(newImageArray)
    arrayDataReal32 = newImage
    call release(set)

  end subroutine replaceImageSingle

  !---------------------------------------------------------------------
  subroutine replaceImageDouble(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    real(double),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    real(double), pointer ::&
      arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL64&
      , set, newImageArray)

    arrayDataReal64 => real64Array2Data(newImageArray)
    arrayDataReal64 = newImage
    call release(set)

  end subroutine replaceImageDouble

  !---------------------------------------------------------------------
  subroutine replaceImageInt8(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int8),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , set, newImageArray)

    arrayDataInt8 => int8Array2Data(newImageArray)
    arrayDataInt8 = newImage
    call release(set)

  end subroutine replaceImageInt8

  !---------------------------------------------------------------------
  subroutine replaceImageInt16(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int16), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int16), pointer ::&
      arrayDataInt16(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER16&
      , set, newImageArray)

    arrayDataInt16 => int16Array2Data(newImageArray)
    arrayDataInt16 = newImage
    call release(set)

  end subroutine replaceImageInt16

  !---------------------------------------------------------------------
  subroutine replaceImageInt32(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.

!***** add optional newPixelSize arguments.


    implicit none

    integer(int32), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int32), pointer ::&
      arrayDataInt32(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER32&
      , set, newImageArray)

    arrayDataInt32 => int32Array2Data(newImageArray)
    arrayDataInt32 = newImage
    call release(set)

  end subroutine replaceImageInt32

  !---------------------------------------------------------------------
  subroutine replaceImageBool(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg)
  !---------------------------------------------------------------------
  ! This subroutine clones oldImageSet to newImageSetName, replaces the
  ! 0th array with an array containing the new image, then adjusts wcs
  ! keywords if present to conform to the start pixels xStartArg and
  ! yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

!***** add optional newPixelSize arguments.



    implicit none

    logical(bool),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      set
    type(ArrayT) ::&
      newImageArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call replaceImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), BOOLEAN&
      , set, newImageArray)

    arrayDataInt8 => int8Array2Data(newImageArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(set)

  end subroutine replaceImageBool

  !---------------------------------------------------------------------
  subroutine replaceImageCommon(oldImageSetName, newImageSetName&
  , tempImageSetName, xStart, yStart, xSize, ySize, newDataType&
  , newSet, newArray)
  !---------------------------------------------------------------------
!*** The previous version seemed to be writing the new array to the primary HDU and tasks that use the dal could not open it. Let's try another way:


!***** add optional newPixelSize arguments.

!*** allow for input of name or block number for the block to replace??


    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName,&
                                   tempImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer,        intent(in)  :: xSize,&
                                   ySize
    integer(int32), intent(in)  :: newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
!    integer :: foreignStatus
    type(WcsT) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default
    integer(int32), pointer ::&
      newDimVector(:)

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      newSet = dataSet(oldImageSetName, MODIFY)
      refSet = clone(oldImageSetName, tempImageSetName, TEMP)

    else
      refSet = dataSet(oldImageSetName, READ)
      newSet = clone(oldImageSetName, newImageSetName, MODIFY)
    end if

    call deleteBlock(newSet, 0)
!*** allow for input of name or block number??

call release(newSet)
newSet = dataSet(newImageSetName, MODIFY)

    allocate(newDimVector(0:1))
    newDimVector = (/xSize, ySize/)

    newArray=addArray(newSet, "PRIMARY", newDataType, newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))
    call release(refSet)
    deallocate(newDimVector)

    ! Regularize and adjust the wcs:

!    call getWcsInfo(newSet, setWcs, foreignStatus)
!    if (foreignStatus == 0) then ! set contains the wcs keywords. Check
!    ! to see if the array block also has a set:
!
!      call getWcsInfo(newArray, wcs, foreignStatus)
!      if (foreignStatus == 0) then ! array also contains the wcs keywords.
!        call warning('setAlsoHasWcs','')
!
!      else ! array doesn't contains any wcs keywords. Copy over those from
!      ! the set:
!        call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
!        call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
!        call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
!        call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
!        call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
!        call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
!        call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
!        call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)
!
!        call warning('setAlsoHasWcs','')
!      end if
!
!      wcsPresent = .true.
!
!    else ! set header doesn't contain any wcs keywords. Check the array:
!      call getWcsInfo(newArray, wcs, foreignStatus)
!      if (foreignStatus .ne. 0) wcsPresent = .true.
!    end if

if (hasWcs(newSet)) then ! set contains the wcs keywords.

  call readWcs(newSet, setWcs)
  ! Now check to see if the array block also has a set:

  if (hasWcs(newArray)) then ! array also contains the wcs keywords.
    call warning('setAlsoHasWcs','')

  else ! array doesn't contains any wcs keywords. Copy over those from
  ! the set:
!    call setAttribute(newArray, 'CRPIX1', setWcs%X%refPixels)
!    call setAttribute(newArray, 'CRPIX2', setWcs%Y%refPixels)
!    call setAttribute(newArray, 'CDELT1', setWcs%X%pixelSizeWorld)
!    call setAttribute(newArray, 'CDELT2', setWcs%Y%pixelSizeWorld)
!    call setAttribute(newArray, 'CRVAL1', setWcs%X%refWorld)
!    call setAttribute(newArray, 'CRVAL2', setWcs%Y%refWorld)
!    call setAttribute(newArray, 'CTYPE1', setWcs%X%ctype)
!    call setAttribute(newArray, 'CTYPE2', setWcs%Y%ctype)

call writeWcs(newArray, setWcs) ! ssclib/wcs_aux

    call warning('onlySetHadWcs','')
  end if

  wcsPresent = .true.

else ! set header doesn't contain any wcs keywords. Check the array:
  if (hasWcs(newArray)) wcsPresent = .true.
end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine replaceImageCommon

  !---------------------------------------------------------------------
  subroutine saveImageSingle(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    real(single),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(single), pointer ::&
      arrayDataReal32(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL32&
      , newSet, newArray, wcs)

    arrayDataReal32 => real32Array2Data(newArray)
    arrayDataReal32 = newImage

    call release(newSet)

  end subroutine saveImageSingle

  !---------------------------------------------------------------------
  subroutine saveImageDouble(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    real(double),   intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    real(double), pointer ::&
      arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), REAL64&
      , newSet, newArray, wcs)

    arrayDataReal64 => real64Array2Data(newArray)
    arrayDataReal64 = newImage

    call release(newSet)

  end subroutine saveImageDouble

  !---------------------------------------------------------------------
  subroutine saveImageBool(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.
  !
  ! The output array is actually int8 rather than boolean.

    implicit none

    logical(bool),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)
    ! Can't read or write an array of type BOOLEAN yet.

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = 0
    where(newImage) arrayDataInt8 = 1

    call release(newSet)

  end subroutine saveImageBool

  !---------------------------------------------------------------------
  subroutine saveImageInt8(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int8),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int8), pointer ::&
      arrayDataInt8(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER8&
      , newSet, newArray, wcs)

    arrayDataInt8 => int8Array2Data(newArray)
    arrayDataInt8 = newImage

    call release(newSet)

  end subroutine saveImageInt8

  !---------------------------------------------------------------------
  subroutine saveImageInt16(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int16),  intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int16), pointer ::&
      arrayDataInt16(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER16&
      , newSet, newArray, wcs)

    arrayDataInt16 => int16Array2Data(newArray)
    arrayDataInt16 = newImage

    call release(newSet)

  end subroutine saveImageInt16

  !---------------------------------------------------------------------
  subroutine saveImageInt32(newImage, oldImageSetName, newImageSetName&
  , tempImageSetName, xStartArg, yStartArg, wcs)
  !---------------------------------------------------------------------
  ! This subroutine creates a new set and array, copies all keywords from
  ! the old set and old 0th array, then adjusts wcs keywords if present
  ! to conform to the start pixels xStartArg and yStartArg of the new image.

    implicit none

    integer(int32), intent(in)           :: newImage(:,:)
    character(*),   intent(in)           :: oldImageSetName,&
                                            newImageSetName,&
                                            tempImageSetName
    integer(int16), intent(in), optional :: xStartArg,&
                                            yStartArg
    type(WcsT),     intent(in), optional :: wcs

    integer(int16) ::&
      xStart, yStart
    type(DataSetT) ::&
      newSet
    type(ArrayT) ::&
      newArray
    integer(int32), pointer ::&
      arrayDataInt32(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    call saveImageCommon(oldImageSetName, newImageSetName, tempImageSetName&
      , xStart, yStart, size(newImage, 1), size(newImage, 2), INTEGER32&
      , newSet, newArray, wcs)

    arrayDataInt32 => int32Array2Data(newArray)
    arrayDataInt32 = newImage

    call release(newSet)

  end subroutine saveImageInt32

  !---------------------------------------------------------------------
  subroutine saveImageCommon(oldImageSetName, newImageSetName&
  , tempImageSetName, xStart, yStart, xSize, ySize, newDataType&
  , newSet, newArray, wcsArg)
  !---------------------------------------------------------------------

!***** add optional newPixelSize arguments.

!*** allow for input of name or block number of the ref block??


    implicit none

    character(*),   intent(in)  :: oldImageSetName,&
                                   newImageSetName,&
                                   tempImageSetName
    integer(int16), intent(in)  :: xStart,&
                                   yStart
    integer,        intent(in)  :: xSize,&
                                   ySize
    integer(int32), intent(in)  :: newDataType
    type(DataSetT), intent(out) :: newSet ! just so it can be released
    type(ArrayT),   intent(out) :: newArray
    type(WcsT),     intent(in), optional :: wcsArg

    type(DataSetT) :: refSet
    type(ArrayT) :: refArray
!    integer :: foreignStatus
    type(WcsT) :: setWcs, wcs
    logical :: wcsPresent = .false. ! default
    integer(int32), pointer ::&
      newDimVector(:)

    !-------------------------------------------------------------------
    ! Start

    if (oldImageSetName .eq. newImageSetName) then
      refSet = clone(oldImageSetName, tempImageSetName, TEMP)

    else
      refSet = dataSet(oldImageSetName, READ)
    end if
    newSet = dataSet(newImageSetName, CREATE)
    call setAttributes(attributable(newSet), attributable(refSet))

    allocate(newDimVector(0:1))
    newDimVector = (/xSize, ySize/)
    newArray = addArray(newSet, "PRIMARY", newDataType, newDimVector)
    deallocate(newDimVector)
    refArray = array(refSet, 0, READ)
    call setAttributes(attributable(newArray), attributable(refArray))

    if (hasDss(newArray)) then
      call dssDelete(block(newArray)) ! Gets rid of DSS attributes
      ! carried over from the ref set.
    end if

    call release(refSet)

    ! Regularize and adjust the wcs:

    if (present(wcsArg)) then
      call writeWcs(newArray, wcsArg) ! in ssclib/wcs_aux
      wcsPresent = .true.
    else
      if (hasWcs(newSet)) then ! set contains the wcs keywords.

        call readWcs(newSet, setWcs)
        ! Now check to see if the array block also has a set:

        if (hasWcs(newArray)) then ! array also contains the wcs keywords.
          call warning('setAlsoHasWcs','')

        else ! array doesn't contains any wcs keywords. Copy over those from
        ! the set:
          call deleteWcs(newSet) ! in ssclib/wcs_aux
          call writeWcs(newArray, setWcs) ! in ssclib/wcs_aux

          call warning('onlySetHadWcs','')
        end if

        wcsPresent = .true.

      else ! set header doesn't contain any wcs keywords. Check the array:
        if (hasWcs(newArray)) wcsPresent = .true.
      end if
    end if

    if (wcsPresent) then
      if (xStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX1', wcs%X%refPixels - (xStart - 1))
      end if
      if (yStart .ne. 1) then
        call setAttribute(newArray, 'CRPIX2', wcs%Y%refPixels - (yStart - 1))
      end if
!*** all this assumes that the pixel sizes stay the same!
    end if

  end subroutine saveImageCommon

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module save_image
