!
! Module: coordinate
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains several routines to do with coordinate transformations.
! 
! Comments: TODO:
!
! $Id: coordinate.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module coordinate

  use types
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use oal	! contains 'use oaltypes'
  use errorHandling

  use constants
  use geometric_types	! for Point2dT and Point2dDbleT.
  use wcs_aux		! for WcsT in getThetaPhiMaps().
  use array_utils	! for vectorCross and indicize().

  implicit none 

  private 

  public ims_getFocalLength
  public calcMmToDetConversionFactor
  private calcRefMatrixInSkyBasis
  private ims_getBoresightMatrix

  public dumpAttitudeMatrix
  public getInstrumentAttitude
  public getScAttitude
!public getInstAttitude_attcalcTest

  public raDecToInst
  public raDecToDet
  public raDecToTel
  public detToRaDec
  public telToRaDec
  public tanToInst
  public tanToDet
  public tanToTel
  public detToTan
  public telToTan
  public rawToInst
  public rawToDet
  public rawToTel
  public detToRaw
  public telToRaw

  public raDecToTan
  public tanToRaDec
  public polarsToRaDec

  public telToDet
  public detToTel
  public psfToDet
!  public psfToTan
!  public tanToPsf

private cc1ToRawMaster
private cc1ToRawMasterCcd
private cc1ToRawSlave
  public calcTransformAndOffset
  public linearTransform

  public getPsfImagePixelCorners
  public getPsfPixelCnrsThetaPhi
  public loadXYindices

  public getThetaPhiMaps
  public skyToCartesian
  public cartesianToSky
  public calcAitoff


  interface raDecToInst
    module procedure raDecToInstScalar
    module procedure raDecToInstVector
    module procedure raDecToInstArray
  end interface

  interface raDecToDet
    module procedure raDecToDetScalar
    module procedure raDecToDetVector
    module procedure raDecToDetArray
  end interface

  interface raDecToTel
    module procedure raDecToTelScalar
    module procedure raDecToTelVector
    module procedure raDecToTelArray
  end interface

  interface detToRaDec
    module procedure detToRaDecScalar
    module procedure detToRaDecVector
    module procedure detToRaDecArray
  end interface

  interface telToRaDec
    module procedure telToRaDecScalar
    module procedure telToRaDecVector
    module procedure telToRaDecArray
  end interface

  interface tanToInst
    module procedure tanToInstScalar, tanToInstScalarWcs
    module procedure tanToInstVector, tanToInstVectorWcs
    module procedure tanToInstArray,  tanToInstArrayWcs
  end interface

  interface tanToDet
    module procedure tanToDetScalar, tanToDetScalarWcs
    module procedure tanToDetVector, tanToDetVectorWcs
    module procedure tanToDetArray,  tanToDetArrayWcs
  end interface

  interface tanToTel
    module procedure tanToTelScalar, tanToTelScalarWcs
    module procedure tanToTelVector, tanToTelVectorWcs
    module procedure tanToTelArray,  tanToTelArrayWcs
  end interface

  interface detToTan
    module procedure detToTanScalar, detToTanScalarWcs
    module procedure detToTanVector, detToTanVectorWcs
    module procedure detToTanArray,  detToTanArrayWcs
  end interface

  interface telToTan
    module procedure telToTanScalar, telToTanScalarWcs
    module procedure telToTanVector, telToTanVectorWcs
    module procedure telToTanArray,  telToTanArrayWcs
  end interface

  interface rawToInst
    module procedure rawToInstScalar
    module procedure rawToInstVector
    module procedure rawToInstArray
  end interface

  interface rawToDet
    module procedure rawToDetScalar
    module procedure rawToDetVector
    module procedure rawToDetArray
  end interface

  interface rawToTel
    module procedure rawToTelScalar
    module procedure rawToTelVector
    module procedure rawToTelArray
  end interface

  interface detToRaw
    module procedure detToRawScalar
    module procedure detToRawVector
    module procedure detToRawArray
  end interface

  interface telToRaw
    module procedure telToRawScalar
    module procedure telToRawVector
    module procedure telToRawArray
  end interface


  interface raDecToTan
    module procedure raDecToTanScalar, raDecToTanScalarWcs
    module procedure raDecToTanVector, raDecToTanVectorWcs
    module procedure raDecToTanArray,  raDecToTanArrayWcs
  end interface

  interface tanToRaDec
    module procedure tanToRaDecScalar, tanToRaDecScalarWcs
    module procedure tanToRaDecVector, tanToRaDecVectorWcs
    module procedure tanToRaDecArray,  tanToRaDecArrayWcs
  end interface

  interface polarsToRaDec
    module procedure polarsToRaDecScalar
    module procedure polarsToRaDecVector
    module procedure polarsToRaDecArray
  end interface


  interface telToDet
    module procedure telToDetScalar
    module procedure telToDetVector
    module procedure telToDetArray
  end interface

  interface detToTel
    module procedure detToTelScalar
    module procedure detToTelVector
    module procedure detToTelArray
  end interface


  interface getInstrumentAttitude
    module procedure getInstrumentAttitudeAtt, getInstrumentAttitudeMat
  end interface

  interface getScAttitude
    module procedure getScAttitudeAtt, getScAttitudeMat
  end interface


  interface psfToDet
    module procedure psfToDetScalar, psfToDetVector, psfToDetArray
  end interface

!  interface psfToTan
!    module procedure psfToTanArray
!  end interface

!  interface tanToPsf
!    module procedure tanToPsfArray
!  end interface

  real(double), public, parameter :: detUnitArcsec = 0.05d0
  ! Is this the best place for this thing?

  logical, public :: applyAttcalcFudge = .false.

contains 





!******** NOTE!!! All the calls with scAttitude as an argument should be changed to instrumAttitude after the PNT keywords are changed to contain the right values.





  !---------------------------------------------------------------------
  function ims_getFocalLength() result(focalLengthMm)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !
  ! See SOC SPR 2518 for an explanation of why this routine is presently necessary.

    implicit none

    real(double) :: focalLengthMm

    integer(int32) :: instrumentId

    !-------------------------------------------------------------------
    ! Start

    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
    call CAL_getState(instrument=instrumentId)

    select case(instrumentId)
    case(EPN)
      focalLengthMm = focalLengthMm + 0.525930866355
    case(EMOS1)
      focalLengthMm = focalLengthMm + 0.09819238106
    case(EMOS2)
      focalLengthMm = focalLengthMm + 0.69819238106
    case default
call error('badInstrument','ssclib: Only EPIC instruments permitted.')
    end select

  end function ims_getFocalLength

  !---------------------------------------------------------------------
  function calcMmToDetConversionFactor() result(mmToDetUnits)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double) :: mmToDetUnits
    real(double) :: focalLengthMm

    !-------------------------------------------------------------------
    ! Start

!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
focalLengthMm = ims_getFocalLength()
    mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec
!**** need to get a different FL correction for each inst until SOC SPR 2518 is solved.

  end function calcMmToDetConversionFactor

  !---------------------------------------------------------------------
  function calcRefMatrixInSkyBasis(refRa, refDec) result(refMatrix_sky)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: refRa, refDec
    real(double)             :: refMatrix_sky(3,3)

    !-------------------------------------------------------------------
    ! Start

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
      , refMatrix_sky(2, :)) ! in ssclib/array_utils.

  end function calcRefMatrixInSkyBasis

  !---------------------------------------------------------------------
  function ims_getBoresightMatrix(timeStamp) result(boresightMatrix)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.
  !.....................................................................
  !
  ! The boresight matrix INS_sc is defined as follows:
  !
  !	INS_sky = INS_sc * SC_sky
  !
  ! where the ith row of the generic matrix INS_<basis> is the vector of
  ! direction cosines (in the given basis) of the ith axis of the
  ! instrument frame. The instrument frame defines the TELCOORD system:
  ! the essential point is that its x axis is aligned with the optical
  ! axis of the instrument, roughly in the opposite direction to the
  ! x axis of the spacecraft or SACCOORD frame.
  !
  ! The boresight matrix INS_sc can therefore be seen to be just the
  ! instrument attitude matrix, but expressed in a basis defined by
  ! the spacecraft frame.
  !.....................................................................
  !
  ! Relation between euler angles and boresight matrix:
  !
  ! If we do
  !
  !	boresight = CAL_getBoresight(timeStamp)
  !	euler = CAL_toEulerAngles(boresight), then
  !
  !	aa(1,:) = (/            1.0d0,           0.0d0,           0.0d0/)
  !	aa(2,:) = (/            0.0d0,  cos(euler%psi),  sin(euler%psi)/)
  !	aa(3,:) = (/            0.0d0, -sin(euler%psi),  cos(euler%psi)/)
  !
  !	bb(1,:) = (/ cos(euler%theta),           0.0d0, -sin(euler%theta)/)
  !	bb(2,:) = (/            0.0d0,           1.0d0,             0.0d0/)
  !	bb(3,:) = (/ sin(euler%theta),           0.0d0,  cos(euler%theta)/)
  !
  !	cc(1,:) = (/   cos(euler%phi),  sin(euler%phi),             0.0d0/)
  !	cc(2,:) = (/  -sin(euler%phi),  cos(euler%phi),             0.0d0/)
  !	cc(3,:) = (/            0.0d0,           0.0d0,             1.0d0/)
  !
  ! we find then that
  !
  !	boresightMatrix = matmul(aa, matmul(bb, cc)).
  !
  ! Note that the convention in use is that the first index of a fortran
  ! matrix represents the row number of the corresponding mathematical matrix.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: timeStamp
    real(double)             :: boresightMatrix(3,3)

    type(BoresightT) :: boresight
    type(EulerAngles321) :: euler321
    real(double), dimension(:,:), pointer ::&
      boresightMatrixPtr

    !-------------------------------------------------------------------
    ! Start

    boresight = CAL_getBoresight(timeStamp)
    euler321 = CAL_toEulerAngles(boresight)

if (applyAttcalcFudge) euler321%psi = -euler321%psi ! reverse sign of psi
! to mimic error in attcalc-4.32.

    boresightMatrixPtr => CAL_toDirectionCosineMatrix(euler321)
    boresightMatrix = boresightMatrixPtr
    call CAL_releaseMemory(boresightMatrixPtr)

  end function ims_getBoresightMatrix


  !---------------------------------------------------------------------
  subroutine dumpAttitudeMatrix(matrix, messageArg, formatArg)
  !---------------------------------------------------------------------
  ! This prints out the contents of the input 3x3 matrix. NOTE that the
  ! first index of the matrix is assumed to denote the row and the second
  ! the column.

    implicit none

    real(double), intent(in)           :: matrix(3,3)
    character(*), intent(in), optional :: messageArg
    character(*), intent(in), optional :: formatArg

    integer :: row
    character(9) :: yiStr(3)
    character(8) :: localForm

    !-------------------------------------------------------------------
    ! Start

    if (present(formatArg)) then
      localForm = formatArg
    else
      localForm = 'floating'
    end if

    yiStr = (/'m(1,:):  ', 'm(2,:):  ', 'm(3,:):  '/)

    if (present(messageArg)) then
      print *, ''
      print *, messageArg
    end if
      print *, '              m(:,1)         m(:,2)         m(:,3)'
    if (localForm .eq. 'floating') then
      do row = 1, 3
        write(*, '(a9, 3f14.9)')&
        yiStr(row), matrix(row, 1), matrix(row, 2), matrix(row, 3)
      end do
    else
      do row = 1, 3
        write(*, '(a9, 3es14.5)')&
        yiStr(row), matrix(row, 1), matrix(row, 2), matrix(row, 3)
      end do
    end if
    print *, ''

  end subroutine dumpAttitudeMatrix

  !---------------------------------------------------------------------
  function getInstrumentAttitudeAtt(scAttitude, timeStamp)&
  result(instrumAttitude)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.

    implicit none

    real(double), intent(in) ::&
      timeStamp
    type(SpacecraftAttitudeType), intent(in) ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90
    type(SpacecraftAttitudeType) ::&
      instrumAttitude ! typedef from caloalutils/src/caloaldefs.f90

!    type(BoresightT) :: boresight
!    real(double), dimension(:,:), pointer ::&
real(double), dimension(3,3) ::&
      boresightMatrix
    real(double), dimension(3,3) ::&
      scAttitudeMatrix, instrumAttitudeMatrix
    type(EulerAnglesType) ::&
      instrumentEulerAngles ! typedef from oal/f90/oaltypes.f90
!real(double) :: x, y, z

    !-------------------------------------------------------------------
    ! Start

!***** Has NOT yet been checked.

    call OAL_toAttitudeMatrix(scAttitude, scAttitudeMatrix)
    !
    ! The array returned by this call represents the spacecraft attitude
    ! matrix ASC_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the spacecraft
    ! reference frame (the same frame as defines the SACCOORD system).
    ! For example, the x or 1st axis of this frame (which points
    ! approximately in the same direction to the star tracker and
    ! approximately opposite to the optical axes of the instruments)
    ! appears in the first row of
    ! ASC_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the spacecraft or SACCOORD basis, as follows:
    !
    !		v_sc = ASC_sky * v_sky.		(1)
    !
    ! NOTE! The second index of the array scAttitudeMatrix corresponds
    ! to the row number of ASC_sky. scAttitudeMatrix should therefore
    ! be transposed in order to conform to the normal convention in which
    ! the first index corresponds to row number.
    !
    ! So:

    scAttitudeMatrix = transpose(scAttitudeMatrix)

!    boresight = CAL_getBoresight(timeStamp)
!    boresightMatrix => CAL_toDirectionCosineMatrix(boresight)
boresightMatrix = ims_getBoresightMatrix(timeStamp)
    !
    ! The boresight matrix AIN_sc is defined as follows:
    !
    !		AIN_sky = AIN_sc * ASC_sky		(2)
    !
    ! where the ith row of the generic matrix AIN_basis is the vector of
    ! direction cosines (in the given basis) of the ith axis of the
    ! instrument frame. The boresight matrix AIN_sc can therefore be
    ! seen to be just the instrument attitude matrix, but expressed in
    ! a basis defined by the spacecraft frame.
    !
    ! The spacecraft frame is contained in the matrix ASC_sky which has,
    ! in its ith row, a vector of direction cosines in the sky basis of
    ! the ith axis of the spacecraft reference frame (the same frame as
    ! defines the SACCOORD system). For example, the x or 1st axis of
    ! this frame (which points approximately in the opposite direction to
    ! the star tracker and the optical axes of the instruments) appears
    ! in the first row of ASC_sky.
    !
    ! NOTE!! The call CAL_toDirectionCosineMatrix(boresight), in contrast
    ! to OAL_toAttitudeMatrix, returns an array in the CORRECT convention:
    ! that is, the first index of the array boresightMatrix as returned
    ! above represents the row index of the associated matrix AIN_sc. Thus
    ! it should NOT be transposed before use in equation (2) above.

    instrumAttitudeMatrix = matmul(boresightMatrix, scAttitudeMatrix)

!x = instrumAttitudeMatrix(1,1)
!y = instrumAttitudeMatrix(1,2)
!z = instrumAttitudeMatrix(1,3)
!if (x .ne. 0.0 .and. y .ne. 0.0) then
!  instrumAttitude%ra   =  atan2(y, x) 
!else
!  instrumAttitude%ra   =  0.0 
!end if
!instrumAttitude%dec  = asin(z)

    call OAL_toEulerAngles(transpose(instrumAttitudeMatrix)&
      , instrumentEulerAngles)
    instrumAttitude%ra   =  instrumentEulerAngles%phi 
    instrumAttitude%dec  = -instrumentEulerAngles%theta
    instrumAttitude%apos = -instrumentEulerAngles%psi
    !
    ! Don't ask me why this works! But it does. If you're worried about
    ! it then use the commented-out xyz stuff immediately above! That
    ! works too and it is easier to follow why.

    do while(instrumAttitude%ra < 0.0)
      instrumAttitude%ra = instrumAttitude%ra + TwoPi
    end do
    do while(instrumAttitude%apos < 0.0)
      instrumAttitude%apos = instrumAttitude%apos + TwoPi
    end do
!!!instrumAttitude%ra   = modulo(instrumAttitude%ra,   TwoPi) ??
!!!instrumAttitude%apos = modulo(instrumAttitude%apos, TwoPi) ??

!    call CAL_releaseMemory(boresightMatrix)

  end function getInstrumentAttitudeAtt

  !---------------------------------------------------------------------
  function getInstrumentAttitudeMat(scAttitudeMat_sky, timeStamp)&
  result(instrumAttitudeMat_sky)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.
  !
  !.....................................................................
  ! The boresight matrix INS_sc is defined as follows:
  !
  !	INS_sky = INS_sc * SC_sky
  !
  ! where the ith row of the generic matrix INS_<basis> is the vector of
  ! direction cosines (in the given basis) of the ith axis of the
  ! instrument frame. The instrument frame defines the TELCOORD system:
  ! the essential point is that its x axis is aligned with the optical
  ! axis of the instrument, roughly in the opposite direction to the
  ! x axis of the spacecraft or SACCOORD frame.
  !
  ! The boresight matrix INS_sc can therefore be seen to be just the
  ! instrument attitude matrix, but expressed in a basis defined by
  ! the spacecraft frame.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: scAttitudeMat_sky(3,3), timeStamp
    real(double)             :: instrumAttitudeMat_sky(3,3)

!    type(BoresightT) :: boresight
!    real(double), dimension(:,:), pointer ::&
real(double), dimension(3,3) ::&
      boresightMatrix

    !-------------------------------------------------------------------
    ! Start

!    boresight = CAL_getBoresight(timeStamp)
!    boresightMatrix => CAL_toDirectionCosineMatrix(boresight)
boresightMatrix = ims_getBoresightMatrix(timeStamp)

    instrumAttitudeMat_sky = matmul(boresightMatrix, scAttitudeMat_sky)
    !
    ! This array is in the correct convention, ie the first index of
    ! the array represents the row index of the associated matrix INS_sky.

!    call CAL_releaseMemory(boresightMatrix)

  end function getInstrumentAttitudeMat

  !---------------------------------------------------------------------
  function getInstAttitude_attcalcTest(scAttitudeMat_sky, timeStamp)&
  result(instrumAttitudeMat_sky)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.
  !.....................................................................
  !
  ! Relation between euler angles and boresight matrix:
  !
  ! If we do
  !
  !	boresight = CAL_getBoresight(timeStamp)
  !	euler = CAL_toEulerAngles(boresight), then
  !
  !	aa(1,:) = (/            1.0d0,           0.0d0,           0.0d0/)
  !	aa(2,:) = (/            0.0d0,  cos(euler%psi),  sin(euler%psi)/)
  !	aa(3,:) = (/            0.0d0, -sin(euler%psi),  cos(euler%psi)/)
  !
  !	bb(1,:) = (/ cos(euler%theta),           0.0d0, -sin(euler%theta)/)
  !	bb(2,:) = (/            0.0d0,           1.0d0,             0.0d0/)
  !	bb(3,:) = (/ sin(euler%theta),           0.0d0,  cos(euler%theta)/)
  !
  !	cc(1,:) = (/   cos(euler%phi),  sin(euler%phi),             0.0d0/)
  !	cc(2,:) = (/  -sin(euler%phi),  cos(euler%phi),             0.0d0/)
  !	cc(3,:) = (/            0.0d0,           0.0d0,             1.0d0/)
  !
  ! we find then that
  !
  !	boresightMatrix = matmul(aa, matmul(bb, cc)).
  !
  ! Note that the convention in use is that the first index of a fortran
  ! matrix represents the row number of the corresponding mathematical matrix.
  !.....................................................................
  !
  ! The boresight matrix INS_sc is defined as follows:
  !
  !	INS_sky = INS_sc * SC_sky
  !
  ! where the ith row of the generic matrix INS_basis is the vector of
  ! direction cosines (in the given basis) of the ith axis of the
  ! instrument frame. The instrument frame defines the TELCOORD system:
  ! the essential point is that its x axis is aligned with the optical
  ! axis of the instrument, roughly in the opposite direction to the
  ! x axis of the spacecraft or SACCOORD frame.
  !
  ! The boresight matrix INS_sc can therefore be seen to be just the
  ! instrument attitude matrix, but expressed in a basis defined by
  ! the spacecraft frame.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: scAttitudeMat_sky(3,3), timeStamp
    real(double)             :: instrumAttitudeMat_sky(3,3)

    type(BoresightT) :: boresight
    type(EulerAngles321) :: euler321
    real(double), dimension(:,:), pointer ::&
      boresightMatrix

    !-------------------------------------------------------------------
    ! Start

    boresight = CAL_getBoresight(timeStamp)
    euler321 = CAL_toEulerAngles(boresight)
    euler321%psi = -euler321%psi ! reverse sign of psi to mimic error
    ! in attcalc-4.32.

    boresightMatrix => CAL_toDirectionCosineMatrix(euler321)
    instrumAttitudeMat_sky = matmul(boresightMatrix, scAttitudeMat_sky)
    call CAL_releaseMemory(boresightMatrix)

  end function getInstAttitude_attcalcTest

  !---------------------------------------------------------------------
  function getScAttitudeAtt(instrumAttitude, timeStamp)&
  result(scAttitude)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.

    implicit none

    real(double), intent(in) ::&
      timeStamp
    type(SpacecraftAttitudeType), intent(in) ::&
      instrumAttitude ! typedef from caloalutils/src/caloaldefs.f90
    type(SpacecraftAttitudeType) ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90

!    type(BoresightT) :: boresight
!    real(double), dimension(:,:), pointer ::&
real(double), dimension(3,3) ::&
      boresightMatrix
    real(double), dimension(3,3) ::&
      scAttitudeMatrix, instrumAttitudeMatrix
    type(EulerAnglesType) ::&
      scEulerAngles ! typedef from oal/f90/oaltypes.f90
!real(double) :: x, y, z

    !-------------------------------------------------------------------
    ! Start

!***** Has NOT yet been checked.

    call OAL_toAttitudeMatrix(instrumAttitude, instrumAttitudeMatrix)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame (which points
    ! approximately in the same direction to the star tracker and
    ! the optical axes of the instruments) appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_ins = Ain_sky * v_sky.		(1)
    !
    ! NOTE! The second index of the array instrumAttitudeMatrix corresponds
    ! to the row number of AIN_sky. instrumAttitudeMatrix should therefore
    ! be transposed in order to conform to the normal convention in which
    ! the first index corresponds to row number.
    !
    ! So:

    instrumAttitudeMatrix = transpose(instrumAttitudeMatrix)

!    boresight = CAL_getBoresight(timeStamp)
!    boresightMatrix => CAL_toDirectionCosineMatrix(boresight)
boresightMatrix = ims_getBoresightMatrix(timeStamp)
    !
    ! The boresight matrix AIN_sc is defined as follows:
    !
    !		AIN_sky = AIN_sc * ASC_sky		(2)
    !
    ! where the ith row of the generic matrix AIN_basis is the vector of
    ! direction cosines (in the given basis) of the ith axis of the
    ! instrument frame. The boresight matrix AIN_sc can therefore be
    ! seen to be just the instrument attitude matrix, but expressed in
    ! a basis defined by the spacecraft frame.
    !
    ! The spacecraft frame is contained in the matrix ASC_sky which has,
    ! in its ith row, a vector of direction cosines in the sky basis of
    ! the ith axis of the spacecraft reference frame (the same frame as
    ! defines the SACCOORD system). For example, the x or 1st axis of
    ! this frame (which points approximately in the opposite direction to
    ! the star tracker and the optical axes of the instruments) appears
    ! in the first row of ASC_sky.
    !
    ! NOTE!! The call CAL_toDirectionCosineMatrix(boresight), in contrast
    ! to OAL_toAttitudeMatrix, returns an array in the CORRECT convention:
    ! that is, the first index of the array boresightMatrix as returned
    ! above represents the row index of the associated matrix AIN_sc. Thus
    ! it should NOT be transposed before use in equation (2) above. However,
    ! don't be confused! In the present subroutine I'm about to transpose
    ! it anyway. What gives? Well, to extract ASC_sky we need to
    ! premultiply equation (2) by the inverse of AIN_sc. But for attitude
    ! (indeed, any orthogonal) matrices, the inverse is just equal to the
    ! transpose. Hence
    !
    !		ASC_sky = AIN_sc^T * AIN_sky		(3)

    boresightMatrix = transpose(boresightMatrix)
    scAttitudeMatrix = matmul(boresightMatrix, instrumAttitudeMatrix)
    !
    ! The resulting array scAttitudeMatrix is still the correct
    ! convention, ie the first index of the array represents the row
    ! index of the associated matrix AIN_sky.

!x = scAttitudeMatrix(1,1)
!y = scAttitudeMatrix(1,2)
!z = scAttitudeMatrix(1,3)
!if (x .ne. 0.0 .and. y .ne. 0.0) then
!  scAttitude%ra   =  atan2(y, x) 
!else
!  scAttitude%ra   =  0.0 
!end if
!scAttitude%dec  = asin(z)

    call OAL_toEulerAngles(transpose(scAttitudeMatrix)&
    , scEulerAngles)
    scAttitude%ra   =  scEulerAngles%phi 
    scAttitude%dec  = -scEulerAngles%theta
    scAttitude%apos = -scEulerAngles%psi
    !
    ! Don't ask me why this works! But it does. If you're worried about
    ! it then use the commented-out xyz stuff immediately above! That
    ! works too and it is easier to follow why.

    do while(scAttitude%ra < 0.0)
      scAttitude%ra = scAttitude%ra + TwoPi
    end do
    do while(scAttitude%apos < 0.0)
      scAttitude%apos = scAttitude%apos + TwoPi
    end do
!!!scAttitude%ra   = modulo(scAttitude%ra,   TwoPi) ??
!!!scAttitude%apos = modulo(scAttitude%apos, TwoPi) ??

!    call CAL_releaseMemory(boresightMatrix)

  end function getScAttitudeAtt

  !---------------------------------------------------------------------
  function getScAttitudeMat(instrumAttitudeMat_sky, timeStamp)&
  result(scAttitudeMat_sky)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the
  ! calling routine.
  !
  !.....................................................................
  ! The boresight matrix INS_sc is defined as follows:
  !
  !	INS_sky = INS_sc * SC_sky
  !
  ! where the ith row of the generic matrix INS_<basis> is the vector of
  ! direction cosines (in the given basis) of the ith axis of the
  ! instrument frame. The instrument frame defines the TELCOORD system:
  ! the essential point is that its x axis is aligned with the optical
  ! axis of the instrument, roughly in the opposite direction to the
  ! x axis of the spacecraft or SACCOORD frame.
  !
  ! The boresight matrix INS_sc can therefore be seen to be just the
  ! instrument attitude matrix, but expressed in a basis defined by
  ! the spacecraft frame.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: instrumAttitudeMat_sky(3,3), timeStamp
    real(double)             :: scAttitudeMat_sky(3,3)

!    type(BoresightT) :: boresight
!    real(double), dimension(:,:), pointer ::&
real(double), dimension(3,3) ::&
      boresightMatrix

    !-------------------------------------------------------------------
    ! Start

!    boresight = CAL_getBoresight(timeStamp)
!    boresightMatrix => CAL_toDirectionCosineMatrix(boresight)
boresightMatrix = ims_getBoresightMatrix(timeStamp)

    boresightMatrix = transpose(boresightMatrix)
    scAttitudeMat_sky = matmul(boresightMatrix, instrumAttitudeMat_sky)
    !
    ! This array is in the correct convention, ie the first index of
    ! the array represents the row index of the associated matrix INS_sky.

!    call CAL_releaseMemory(boresightMatrix)

  end function getScAttitudeMat


  !---------------------------------------------------------------------
  subroutine raDecToInstScalar(ra, dec, scAttitude, timeStamp, xDet, yDet&
  , tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra,&
                                                 dec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet,&
                                                 tanThetaArcsec,&
                                                 phi

    real(double) :: xDetVec(1), yDetVec(1), tanThetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call raDecToInstVector((/ra/), (/dec/), scAttitude, timeStamp&
      , xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)
    tanThetaArcsec = tanThetaArcsecVec(1)
    phi            = phiVec(1)

  end subroutine raDecToInstScalar

  !---------------------------------------------------------------------
  subroutine raDecToInstVector(ra, dec, scAttitude, timeStamp, xDet, yDet&
  , tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs into the XMM CAMCOORD2 (detx/y)
  ! coordinate system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: ra(:),&
                                                 dec(size(ra))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(ra)),&
                                                 yDet(size(ra)),&
                                                 tanThetaArcsec(size(ra)),&
                                                 phi(size(ra))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z
    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: ra/dec to detx/y.')

    mmToDetUnits = calcMmToDetConversionFactor()
    call raDecToTelVector(ra, dec, scAttitude, timeStamp, tanThetaArcsec, phi)
    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine raDecToInstVector

  !---------------------------------------------------------------------
  subroutine raDecToInstArray(ra, dec, scAttitude, timeStamp, xDet, yDet&
  , tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra(:,:),&
                                                 dec(size(ra,1),size(ra,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(ra,1),size(ra,2)),&
                                                 yDet(size(ra,1),size(ra,2)),&
                                                 tanThetaArcsec(&
                                                   size(ra,1),&
                                                   size(ra,2)),&
                                                 phi(size(ra,1),size(ra,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable ::&
      xDetVec(:), yDetVec(:), tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(ra, 1)
    ySize = size(ra, 2)
    vSize = xSize * ySize

    allocate(raVec(vSize), decVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    raVec  = reshape(ra,  (/vSize/))
    decVec = reshape(dec, (/vSize/))

    call raDecToInstVector(raVec, decVec, scAttitude, timeStamp&
      , xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))
    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi            = reshape(phiVec,            (/xSize, ySize/))

  end subroutine raDecToInstArray


  !---------------------------------------------------------------------
  subroutine raDecToDetScalar(ra, dec, scAttitude, timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra,&
                                                 dec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet

    real(double) :: xDetVec(1), yDetVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call raDecToDetVector((/ra/), (/dec/), scAttitude, timeStamp&
      , xDetVec, yDetVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)

  end subroutine raDecToDetScalar

  !---------------------------------------------------------------------
  subroutine raDecToDetVector(ra, dec, scAttitude, timeStamp, xDet, yDet)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs into the XMM CAMCOORD2 (detx/y)
  ! coordinate system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: ra(:),&
                                                 dec(size(ra))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(ra)),&
                                                 yDet(size(ra))

!    real(double), dimension(:), pointer ::&
    real(double), dimension(:), allocatable ::&
      tanThetaArcsec, phi
!, cc2X, cc2Y, cc2Z
!    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: ra/dec to detx/y.')

!    mmToDetUnits = calcMmToDetConversionFactor()
    allocate(tanThetaArcsec(size(ra)), phi(size(ra)))
    call raDecToTelVector(ra, dec, scAttitude, timeStamp, tanThetaArcsec, phi)

!    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)
!    xDet = cc2X * mmToDetUnits
!    yDet = cc2Y * mmToDetUnits
!    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)
call telToDetVector(tanThetaArcsec, phi, xDet, yDet)

!    deallocate(tanThetaArcsec, phi)

  end subroutine raDecToDetVector

  !---------------------------------------------------------------------
  subroutine raDecToDetArray(ra, dec, scAttitude, timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra(:,:),&
                                                 dec(size(ra,1),size(ra,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(ra,1),size(ra,2)),&
                                                 yDet(size(ra,1),size(ra,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable :: xDetVec(:), yDetVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(ra, 1)
    ySize = size(ra, 2)
    vSize = xSize * ySize

    allocate(raVec(vSize), decVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    raVec  = reshape(ra,  (/vSize/))
    decVec = reshape(dec, (/vSize/))

    call raDecToDetVector(raVec, decVec, scAttitude, timeStamp&
      , xDetVec, yDetVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))

  end subroutine raDecToDetArray


  !---------------------------------------------------------------------
  subroutine raDecToTelScalar(ra, dec, scAttitude, timeStamp, thetaArcsec&
  , phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra,&
                                                 dec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: thetaArcsec,&
                                                 phi

    real(double) :: thetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call raDecToTelVector((/ra/), (/dec/), scAttitude, timeStamp&
      , thetaArcsecVec, phiVec)

    thetaArcsec = thetaArcsecVec(1)
    phi = phiVec(1)

  end subroutine raDecToTelScalar

  !---------------------------------------------------------------------
  subroutine raDecToTelVector(ra, dec, scAttitude, timeStamp&
  , tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs into the XMM TELCOORD polar
  ! coordinate system.
  !
  ! The relevant coordinate systems are related as follows. We start
  ! with the spacecraft (sc) frame, which is defined in the usual way
  ! by a right-hand trio of orthogonal basis vectors. These vectors are
  ! anchored in the spacecraft structure. The second coordinate system
  ! is the telescope or instrument (ins) one: broadly speaking, the X
  ! and Y axes of the ins system point in the opposite direction to
  ! respectively the X and Y axes of the sc system, whereas +Z_ins and
  ! +Z_sc are approximately aligned. (These alignments are not exact,
  ! and are also slightly different from telescope to telescope: the
  ! exact angles between them for a given instrument comprise the
  ! boresight for that telescope.) The origin of the ins system for a
  ! given telescope is in the centre of optical symmetry of the
  ! respective mirror module. The positive X axis of the ins system
  ! points in the opposite direction to the propagation of light in
  ! the telescope, and is defined by the optical axis of that
  ! telescope; the Y and Z axes are tied to the detector array. The
  ! third and final system is not actually a 3D spatial coordinate
  ! system, rather it is a 2D tangential projection coordinate system;
  ! this is the detector (det) system. The projection plane is normal
  ! to the optic axis, the det X and Y axes are aligned with the
  ! CAMCOORD2 axes, and the origin of the det system is defined by the
  ! ray which passes through the CAMCOORD2 origin.
  !
  ! Whereas one can translate between the sc and ins systems by means
  ! of the boresight, one translates between the ins and det systems
  ! via the TELCOORD angles. Consider a ray which passes through the
  ! origin of the ins system and intercepts the detector. The angle
  ! between this ray and the optic axis is defined as TELCOORD theta
  ! (actually at present theta is the _tangent_ of this angle). How
  ! one specifies TELCOORD phi depends on whether one follows the ray
  ! from the ins origin towards the detector, in which case phi =
  ! atan(Y_ins/-Z_ins), or from the ins origin back towards the source
  ! in the sky, in which case phi = atan(-Y_ins/Z_ins). In the present
  ! case the input information (ra and dec) describes a source located
  ! on the celestial sphere, ie it specifies a vector pointing _away_
  ! from the spacecraft; hence the latter expression is used to turn
  ! the source vector in the ins coordinate system into appropriate
  ! TELCOORD angles. The relevant CAL routine can then be used to
  ! translate those angles into the det system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: ra(:),&
                                                 dec(size(ra))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(size(ra)),&
                                                 phi(size(ra))

    integer(int32) :: numPositions, i
    real(double) ::&
      srcVector_ins(3), srcVector_sky(3), semiRadius,&
      skyMatrix_sc(3,3), instrumAttitudeMat_sky(3,3)

    !-------------------------------------------------------------------
    ! Start

    tanThetaArcsec = 0.0d0
    phi = 0.0d0

    numPositions = size(ra)

    call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
    instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
      , timeStamp)

    do i = 1, numPositions
      srcVector_sky = skyToCartesian(ra(i), dec(i))
      srcVector_ins = matmul(instrumAttitudeMat_sky, srcVector_sky)
      if (srcVector_ins(1) <= 0.0)&
call error('thetaGreaterThan90',"ssclib:raDecToTelVector: The given RA/dec is &
&further than 90 deg from the instrument pointing direction.")

      semiRadius = sqrt(srcVector_ins(2) * srcVector_ins(2)&
        + srcVector_ins(3) * srcVector_ins(3))

      tanThetaArcsec(i) = semiRadius / srcVector_ins(1) / ArcsecToRad
!      phi(i) = atan2(srcVector_ins(3), -srcVector_ins(2))
      phi(i) = atan2(-srcVector_ins(2), srcVector_ins(3))
    end do

    do while(any(phi > TwoPi))
      where(phi > TwoPi) phi = phi - TwoPi
    end do
    do while(any(phi < 0.0d0))
      where(phi < 0.0d0) phi = phi + TwoPi
    end do

  end subroutine raDecToTelVector

  !---------------------------------------------------------------------
  subroutine raDecToTelVector_20041109(ra, dec, scAttitude, timeStamp&
  , tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs into the XMM TELCOORD polar
  ! coordinate system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: ra(:),&
                                                 dec(size(ra))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(size(ra)),&
                                                 phi(size(ra))

    type(BoresightT) :: boresight
    type(EquatorialDirectionType), allocatable :: srcDirection(:)
    real(double), dimension(:), pointer ::&
      localTanThetaArcsec, localPhi
    integer(int32) ::&
      numPositions, i

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90
    !              BoresightT typedef from cal/f90/caltypes.f90
    ! EquatorialDirectionType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: ra/dec to TELCOORD.')

    tanThetaArcsec = 0.0d0
    phi = 0.0d0

    numPositions = size(ra)

    boresight = CAL_getBoresight(timeStamp) ! needs instrument

    allocate(srcDirection(numPositions))

    if (any(dec < -PiOverTwo .or. dec > PiOverTwo)) then
call error('badDecValue','ssclib: The value of declination is outside the &
&range -90 to 90 degrees.')
    end if

    srcDirection%ra  = ra
    srcDirection%dec = dec

    do i = 1, numPositions
      do while(srcDirection(i)%ra >= TwoPi)
        srcDirection(i)%ra = srcDirection(i)%ra - TwoPi
      end do
      do while(srcDirection(i)%ra < 0.0)
        srcDirection(i)%ra = srcDirection(i)%ra + TwoPi
      end do
!!!srcDirection(i)%ra = modulo(srcDirection(i)%ra, TwoPi)  ???
    end do

    call CAL_withSacCoordOrientation(boresight, scAttitude) ! needs instrument
    call CAL_projectOntoTelCoord(boresight, srcDirection&
      , localTanThetaArcsec, localPhi)
    ! NOTE! Units of localPhi are radians.

!*** NOTE the kludge I have put in to reverse the boresight psi will NOT fix the present routine. Bugger.

    tanThetaArcsec = localTanThetaArcsec
    phi            = localPhi

    call CAL_releaseMemory(localTanThetaArcsec)
    call CAL_releaseMemory(localPhi)

  end subroutine raDecToTelVector_20041109

  !---------------------------------------------------------------------
  subroutine raDecToTelArray(ra, dec, scAttitude, timeStamp, tanThetaArcsec&
  , phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: ra(:,:),&
                                                 dec(size(ra,1),size(ra,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(&
                                                   size(ra,1),size(ra,2)),&
                                                 phi(size(ra,1),size(ra,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(ra, 1)
    ySize = size(ra, 2)
    vSize = xSize * ySize

    allocate(raVec(vSize), decVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    raVec  = reshape(ra,  (/vSize/))
    decVec = reshape(dec, (/vSize/))

    call raDecToTelVector(raVec, decVec, scAttitude, timeStamp&
      , tanThetaArcsecVec, phiVec)

    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi            = reshape(phiVec, (/xSize, ySize/))

  end subroutine raDecToTelArray


  !---------------------------------------------------------------------
  subroutine detToRaDecScalar(xDet, yDet, scAttitude, timeStamp, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet,&
                                                 yDet
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra,&
                                                 dec

    real(double) :: raVec(1), decVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call detToRaDecVector((/xDet/), (/yDet/), scAttitude, timeStamp&
      , raVec, decVec)

    ra  = raVec(1)
    dec = decVec(1)

  end subroutine detToRaDecScalar

  !---------------------------------------------------------------------
  subroutine detToRaDecVector(xDet, yDet, scAttitude, timeStamp, ra, dec)
  !---------------------------------------------------------------------
  ! This subroutine calculates RAs and decs which correspond to positions
  ! in the XMM CAMCOORD2 (detx/y) coordinate system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xDet(:),&
                                                 yDet(size(xDet))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra(size(xDet)),&
                                                 dec(size(xDet))

!    real(double), dimension(:), pointer :: tanThetaArcsec, phi
    real(double), dimension(:), allocatable :: tanThetaArcsec, phi
!    real(double), dimension(:), allocatable :: cc2X, cc2Y, cc2Z
    integer(int32) :: numPositions
!    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: detx/y to ra/dec.')

    ra  = 0.0d0 ! default
    dec = 0.0d0 ! default

    numPositions = size(xDet)

!    allocate(cc2X(numPositions), cc2Y(numPositions), cc2Z(numPositions))
!    mmToDetUnits = calcMmToDetConversionFactor()
!    cc2X = xDet / mmToDetUnits
!    cc2Y = yDet / mmToDetUnits
!    cc2Z = 0.0d0
!    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, tanThetaArcsec, phi)
!!    deallocate(cc2X, cc2Y, cc2Z)
allocate(tanThetaArcsec(numPositions), phi(numPositions))
call detToTelVector(xDet, yDet, tanThetaArcsec, phi)

    call telToRaDecVector(tanThetaArcsec, phi, scAttitude, timeStamp&
      , ra, dec)

!    call CAL_releaseMemory(localThetaArcsec)
!    call CAL_releaseMemory(localPhi)

  end subroutine detToRaDecVector

  !---------------------------------------------------------------------
  subroutine detToRaDecArray(xDet, yDet, scAttitude, timeStamp, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet(:,:),&
                                                 yDet(size(xDet,1),size(xDet,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra(size(xDet,1),size(xDet,2)),&
                                                 dec(size(xDet,1),size(xDet,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xDetVec(:), yDetVec(:)
    real(double), allocatable :: raVec(:), decVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xDet, 1)
    ySize = size(xDet, 2)
    vSize = xSize * ySize

    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(raVec(vSize), decVec(vSize))
    xDetVec = reshape(xDet, (/vSize/))
    yDetVec = reshape(yDet, (/vSize/))

    call detToRaDecVector(xDetVec, yDetVec, scAttitude, timeStamp&
      , raVec, decVec)

    ra  = reshape(raVec,  (/xSize, ySize/))
    dec = reshape(decVec, (/xSize, ySize/))

  end subroutine detToRaDecArray


  !---------------------------------------------------------------------
  subroutine telToRaDecScalar(tanThetaArcsec, phi, scAttitude, timeStamp&
  , ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec,&
                                                 phi
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra,&
                                                 dec

    real(double) :: raVec(1), decVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call telToRaDecVector((/tanThetaArcsec/), (/phi/), scAttitude, timeStamp&
      , raVec, decVec)

    ra  = raVec(1)
    dec = decVec(1)

  end subroutine telToRaDecScalar

  !---------------------------------------------------------------------
  subroutine telToRaDecVector(tanThetaArcsec, phi, scAttitude, timeStamp&
  , ra, dec)
  !---------------------------------------------------------------------
  ! This subroutine calculates RAs and decs which correspond to positions
  ! in the XMM TELCOORD polar coordinate system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:),&
                                                 phi(size(tanThetaArcsec))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra(size(tanThetaArcsec)),&
                                                 dec(size(tanThetaArcsec))

    real(double) ::&
      tanTheta,&
      srcVector_ins(3),&
      skyMatrix_sc(3,3), srcVector_sky(3), instrumAttitudeMat_sky(3,3)
    integer(int32) ::&
      numPositions, i

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: TELCOORD to ra/dec.')

    ra  = 0.0d0 ! default
    dec = 0.0d0 ! default

    numPositions = size(tanThetaArcsec)

    call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
    instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
      , timeStamp)

    do i = 1, numPositions
!      sinTheta = sin(thetaArcsec(i - 1) * ArcsecToRad)
!      tanTheta = tan(thetaArcsec(i - 1) * ArcsecToRad)
      tanTheta = tanThetaArcsec(i) * ArcsecToRad
      srcVector_ins(1) = 1.0d0
      srcVector_ins(2) = -tanTheta * sin(phi(i))
      srcVector_ins(3) =  tanTheta * cos(phi(i))
      ! NOTE! srcVector_ins is not normalized to length 1. cartesianToSky
      ! should be able to handle this though.

      srcVector_sky = matmul(transpose(instrumAttitudeMat_sky), srcVector_ins)
      call cartesianToSky(srcVector_sky, ra(i), dec(i))
    end do

  end subroutine telToRaDecVector

  !---------------------------------------------------------------------
  subroutine telToRaDecArray(tanThetaArcsec, phi, scAttitude, timeStamp&
  , ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:,:),&
                                                 phi(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: ra(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2)),&
                                                 dec(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanThetaArcsec, 1)
    ySize = size(tanThetaArcsec, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    allocate(raVec(vSize), decVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi,         (/vSize/))

    call telToRaDecVector(tanThetaArcsecVec, phiVec, scAttitude, timeStamp&
      , raVec, decVec)

    ra  = reshape(raVec,  (/xSize, ySize/))
    dec = reshape(decVec, (/xSize, ySize/))

  end subroutine telToRaDecArray


  !---------------------------------------------------------------------
  subroutine tanToInstScalar(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan,&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet,&
                                                 tanThetaArcsec,&
                                                 phi

    real(double) :: xDetVec(1), yDetVec(1), tanThetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToInstVector((/xTan/), (/yTan/), refRa, refDec, scAttitude&
      , timeStamp, xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)
    tanThetaArcsec = tanThetaArcsecVec(1)
    phi            = phiVec(1)

  end subroutine tanToInstScalar

  !---------------------------------------------------------------------
  subroutine tanToInstVector(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into CAMCOORD2 (detx/y) coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(xTan)),&
                                                 yDet(size(xTan)),&
                                                 tanThetaArcsec(size(xTan)),&
                                                 phi(size(xTan))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z
    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: image tangent coordinates &
!      &to detx/y.')

    call tanToTelVector(xTan, yTan, refRa, refDec, scAttitude&
      , timeStamp, tanThetaArcsec, phi)

    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)

    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits

    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine tanToInstVector

  !---------------------------------------------------------------------
  subroutine tanToInstArray(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 yDet(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 tanThetaArcsec(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 phi(&
                                                   size(xTan,1),&
                                                   size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable ::&
      xDetVec(:), yDetVec(:), tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xTanVec  = reshape(xTan,  (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToInstVector(xTanVec, yTanVec, refRa, refDec, scAttitude&
      , timeStamp, xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))
    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi            = reshape(phiVec,            (/xSize, ySize/))

  end subroutine tanToInstArray

  !---------------------------------------------------------------------
  subroutine tanToInstScalarWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan
!,&
!                                                 refRa,&
!                                                 refDec
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet,&
                                                 tanThetaArcsec,&
                                                 phi

    real(double) :: xDetVec(1), yDetVec(1), tanThetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToInstVector((/xTan/), (/yTan/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)
    tanThetaArcsec = tanThetaArcsecVec(1)
    phi            = phiVec(1)

  end subroutine tanToInstScalarWcs

  !---------------------------------------------------------------------
  subroutine tanToInstVectorWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into CAMCOORD2 (detx/y) coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan))
!,&
!                                                 refRa,&
!                                                 refDec
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(xTan)),&
                                                 yDet(size(xTan)),&
                                                 tanThetaArcsec(size(xTan)),&
                                                 phi(size(xTan))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z
    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: image tangent coordinates &
!      &to detx/y.')

    call tanToTelVector(xTan, yTan&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, tanThetaArcsec, phi)

    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)

    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits

    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine tanToInstVectorWcs

  !---------------------------------------------------------------------
  subroutine tanToInstArrayWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2))
!,&
!                                                 refRa,&
!                                                 refDec
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 yDet(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 tanThetaArcsec(&
                                                   size(xTan,1),&
                                                   size(xTan,2)),&
                                                 phi(&
                                                   size(xTan,1),&
                                                   size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable ::&
      xDetVec(:), yDetVec(:), tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xTanVec  = reshape(xTan,  (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToInstVector(xTanVec, yTanVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xDetVec, yDetVec, tanThetaArcsecVec, phiVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))
    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi            = reshape(phiVec,            (/xSize, ySize/))

  end subroutine tanToInstArrayWcs


  !---------------------------------------------------------------------
  subroutine tanToDetScalar(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan,&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet

    real(double) :: xDetVec(1), yDetVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToDetVector((/xTan/), (/yTan/), refRa, refDec, scAttitude&
      , timeStamp, xDetVec, yDetVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)

  end subroutine tanToDetScalar

  !---------------------------------------------------------------------
  subroutine tanToDetVector(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into CAMCOORD2 (detx/y) coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(xTan)),&
                                                 yDet(size(xTan))

!    real(double), dimension(:), pointer ::&
    real(double), dimension(:), allocatable ::&
      tanThetaArcsec, phi
!, cc2X, cc2Y, cc2Z
!    real(double) :: mmToDetUnits

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: image tangent coordinates &
!      &to detx/y.')

    allocate(tanThetaArcsec(size(xTan)), phi(size(xTan)))

    call tanToTelVector(xTan, yTan, refRa, refDec, scAttitude&
      , timeStamp, tanThetaArcsec, phi)

    call telToDetVector(tanThetaArcsec, phi, xDet, yDet)

  end subroutine tanToDetVector

  !---------------------------------------------------------------------
  subroutine tanToDetArray(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 yDet(&
                                                   size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: xDetVec(:), yDetVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    xTanVec  = reshape(xTan,  (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToDetVector(xTanVec, yTanVec, refRa, refDec, scAttitude&
      , timeStamp, xDetVec, yDetVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))

  end subroutine tanToDetArray

  !---------------------------------------------------------------------
  subroutine tanToDetScalarWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet,&
                                                 yDet

    real(double) :: xDetVec(1), yDetVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToDetVector((/xTan/), (/yTan/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xDetVec, yDetVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)

  end subroutine tanToDetScalarWcs

  !---------------------------------------------------------------------
  subroutine tanToDetVectorWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into CAMCOORD2 (detx/y) coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(size(xTan)),&
                                                 yDet(size(xTan))

    real(double), dimension(:), allocatable ::&
      tanThetaArcsec, phi

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: image tangent coordinates &
!      &to detx/y.')

    allocate(tanThetaArcsec(size(xTan)), phi(size(xTan)))

    call tanToTelVector(xTan, yTan&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, tanThetaArcsec, phi)

    call telToDetVector(tanThetaArcsec, phi, xDet, yDet)

  end subroutine tanToDetVectorWcs

  !---------------------------------------------------------------------
  subroutine tanToDetArrayWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xDet(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 yDet(&
                                                   size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: xDetVec(:), yDetVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    xTanVec  = reshape(xTan,  (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToDetVector(xTanVec, yTanVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xDetVec, yDetVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))

  end subroutine tanToDetArrayWcs


  !---------------------------------------------------------------------
  subroutine tanToTelScalar(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan,&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec,&
                                                 phi

    real(double) :: tanThetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToTelVector((/xTan/), (/yTan/), refRa, refDec, scAttitude&
      , timeStamp, tanThetaArcsecVec, phiVec)

    tanThetaArcsec = tanThetaArcsecVec(1)
    phi         = phiVec(1)

  end subroutine tanToTelScalar

  !---------------------------------------------------------------------
  subroutine tanToTelVector(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into TELCOORD polar coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(size(xTan)),&
                                                 phi(size(xTan))

    real(double) ::&
      srcVector_ins(3), srcVector_ref(3), semiRadius,&
      skyMatrix_sc(3,3), instrumAttitudeMat_sky(3,3), refMatrix_sky(3,3)
!    real(double), save :: refMatrix_ins(3,3)
real(double) :: refMatrix_ins(3,3)
    integer(int32) :: numPositions, i
!    logical(bool) :: localIsNewAtt, isFirstCall = .true.

    !-------------------------------------------------------------------
    ! Start

    tanThetaArcsec = 0.0d0
    phi = 0.0d0

    numPositions = size(xTan)

    call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
    instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
      , timeStamp)

    refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)
    refMatrix_ins = matmul(refMatrix_sky, transpose(instrumAttitudeMat_sky))

    do i = 1, numPositions
      srcVector_ref(1) = 1.0
      srcVector_ref(2) = -xTan(i)
      srcVector_ref(3) =  yTan(i)
      ! NOTE that srcVector in this formulation is not of length 1.

      srcVector_ins = matmul(transpose(refMatrix_ins), srcVector_ref)

      if (srcVector_ins(1) <= 0.0)&
call error('thetaGreaterThan90',"ssclib:tanToTelVector: The given tangent-plane &
&position is further than 90 deg from the instrument pointing direction.")

      semiRadius = sqrt(srcVector_ins(2) * srcVector_ins(2)&
        + srcVector_ins(3) * srcVector_ins(3))

      tanThetaArcsec(i) = semiRadius / srcVector_ins(1) / ArcsecToRad

!      phi(i) = atan2(srcVector_ins(3), -srcVector_ins(2))
      phi(i) = atan2(-srcVector_ins(2), srcVector_ins(3))
    end do

    do while(any(phi > TwoPi))
      where(phi > TwoPi) phi = phi - TwoPi
    end do
    do while(any(phi < 0.0d0))
      where(phi < 0.0d0) phi = phi + TwoPi
    end do

  end subroutine tanToTelVector

  !---------------------------------------------------------------------
  subroutine tanToTelVector_20041109(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into TELCOORD polar coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(size(xTan)),&
                                                 phi(size(xTan))

    type(BoresightT) :: boresight
    type(EquatorialDirectionType), allocatable :: srcDirection(:)
    real(double), dimension(:), pointer ::&
      localTanThetaArcsec, localPhi
    integer(int32) :: numPositions, i
    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90
    !              BoresightT typedef from cal/f90/caltypes.f90
    ! EquatorialDirectionType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: image tangent coordinates &
!      &to TELCOORD.')

    tanThetaArcsec = 0.0d0
    phi = 0.0d0

    numPositions = size(xTan)

    boresight = CAL_getBoresight(timeStamp) ! needs instrument

    allocate(srcDirection(numPositions))

    refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)

    do i = 1, numPositions
      srcVector_ref(1) = 1.0d0
      srcVector_ref(2) = -xTan(i)
      srcVector_ref(3) =  yTan(i)

      ! NOTE! srcVector_ref is not normalized to length 1. cartesianToSky
      ! should be able to handle this though.

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky,srcDirection(i)%ra,srcDirection(i)%dec)
    end do

    call CAL_withSacCoordOrientation(boresight, scAttitude)
    call CAL_projectOntoTelCoord(boresight, srcDirection, localTanThetaArcsec&
      , localPhi) 
    ! NOTE! Units of localPhi are radians.

!*** NOTE the kludge I have put in to reverse the boresight psi will NOT fix the present routine. Bugger.

    tanThetaArcsec = localTanThetaArcsec
    phi = localPhi

    call CAL_releaseMemory(localTanThetaArcsec)
    call CAL_releaseMemory(localPhi)

  end subroutine tanToTelVector_20041109

  !---------------------------------------------------------------------
  subroutine tanToTelArray(xTan, yTan, refRa, refDec, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 refRa,&
                                                 refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 phi(&
                                                   size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xTanVec = reshape(xTan, (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToTelVector(xTanVec, yTanVec, refRa, refDec, scAttitude&
      , timeStamp, tanThetaArcsecVec, phiVec)

    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi = reshape(phiVec, (/xSize, ySize/))

  end subroutine tanToTelArray

  !---------------------------------------------------------------------
  subroutine tanToTelScalarWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan,&
                                                 yTan
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec,&
                                                 phi

    real(double) :: tanThetaArcsecVec(1), phiVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call tanToTelVector((/xTan/), (/yTan/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, tanThetaArcsecVec, phiVec)

    tanThetaArcsec = tanThetaArcsecVec(1)
    phi         = phiVec(1)

  end subroutine tanToTelScalarWcs

  !---------------------------------------------------------------------
  subroutine tanToTelVectorWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image or tangent plane
  ! coordinate system into TELCOORD polar coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xTan(:),&
                                                 yTan(size(xTan))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(size(xTan)),&
                                                 phi(size(xTan))

    real(double) ::&
      srcVector_ins(3), srcVector_ref(3), semiRadius,&
      skyMatrix_sc(3,3), instrumAttitudeMat_sky(3,3), refMatrix_sky(3,3)
!    real(double), save :: refMatrix_ins(3,3)
real(double) :: refMatrix_ins(3,3)
    integer(int32) :: numPositions, i
!    logical(bool) :: localIsNewAtt, isFirstCall = .true.

    !-------------------------------------------------------------------
    ! Start

    tanThetaArcsec = 0.0d0
    phi = 0.0d0

    numPositions = size(xTan)

    call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
    instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
      , timeStamp)

    refMatrix_sky = calcRefMatrixInSkyBasis(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180)
    refMatrix_ins = matmul(refMatrix_sky, transpose(instrumAttitudeMat_sky))

    do i = 1, numPositions
      srcVector_ref(1) = 1.0
      srcVector_ref(2) = -xTan(i)
      srcVector_ref(3) =  yTan(i)
      ! NOTE that srcVector in this formulation is not of length 1.

      srcVector_ins = matmul(transpose(refMatrix_ins), srcVector_ref)

      if (srcVector_ins(1) <= 0.0)&
call error('thetaGreaterThan90',"ssclib:tanToTelVectorWcs: The given tangent-plane &
&position is further than 90 deg from the instrument pointing direction.")

      semiRadius = sqrt(srcVector_ins(2) * srcVector_ins(2)&
        + srcVector_ins(3) * srcVector_ins(3))

      tanThetaArcsec(i) = semiRadius / srcVector_ins(1) / ArcsecToRad

!      phi(i) = atan2(srcVector_ins(3), -srcVector_ins(2))
      phi(i) = atan2(-srcVector_ins(2), srcVector_ins(3))
    end do

    do while(any(phi > TwoPi))
      where(phi > TwoPi) phi = phi - TwoPi
    end do
    do while(any(phi < 0.0d0))
      where(phi < 0.0d0) phi = phi + TwoPi
    end do

  end subroutine tanToTelVectorWcs

  !---------------------------------------------------------------------
  subroutine tanToTelArrayWcs(xTan, yTan, wcs, scAttitude&
  , timeStamp, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xTan(:,:),&
                                                 yTan(&
                                                   size(xTan,1),size(xTan,2))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: tanThetaArcsec(&
                                                   size(xTan,1),size(xTan,2)),&
                                                 phi(&
                                                   size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xTanVec = reshape(xTan, (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToTelVector(xTanVec, yTanVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, tanThetaArcsecVec, phiVec)

    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi = reshape(phiVec, (/xSize, ySize/))

  end subroutine tanToTelArrayWcs


  !---------------------------------------------------------------------
  subroutine detToTanScalar(xDet, yDet, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet,&
                                                 yDet
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan,&
                                                 yTan
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) :: xTanVec(1), yTanVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call detToTanVector((/xDet/), (/yDet/), refRa, refDec, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine detToTanScalar

  !---------------------------------------------------------------------
  subroutine detToTanVector(xDet, yDet, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------
  ! This subroutine reprojects positions from the CAMCOORD2 (detx/y)
  ! coordinate system into image or tangent plane coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xDet(:),&
                                                 yDet(size(xDet))
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(size(xDet)),&
                                                 yTan(size(xDet))
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double), dimension(:), allocatable :: tanThetaArcsec, phi
    integer(int32) :: numPositions

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: detx/y to image tangent &
!      &coordinates.')

    xTan = 0.0d0 ! default
    yTan = 0.0d0 ! default

    numPositions = size(xDet)
    allocate(tanThetaArcsec(numPositions), phi(numPositions))
    call detToTelVector(xDet, yDet, tanThetaArcsec, phi)

    call telToTanVector(tanThetaArcsec, phi, refRa, refDec&
      , scAttitude, timeStamp, xTan, yTan, isNewAtt)

  end subroutine detToTanVector

  !---------------------------------------------------------------------
  subroutine detToTanArray(xDet, yDet, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet(:,:),&
                                                 yDet(size(xDet,1),size(xDet,2))
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(&
                                                   size(xDet,1),size(xDet,2)),&
                                                 yTan(size(xDet,1),size(xDet,2))
    logical(bool), optional,      intent(in)  :: isNewAtt

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xDetVec(:), yDetVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xDet, 1)
    ySize = size(xDet, 2)
    vSize = xSize * ySize

    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    xDetVec = reshape(xDet, (/vSize/))
    yDetVec = reshape(yDet, (/vSize/))

    call detToTanVector(xDetVec, yDetVec, refRa, refDec, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine detToTanArray

  !---------------------------------------------------------------------
  subroutine detToTanScalarWcs(xDet, yDet, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet,&
                                                 yDet
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan,&
                                                 yTan
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) :: xTanVec(1), yTanVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call detToTanVector((/xDet/), (/yDet/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine detToTanScalarWcs

  !---------------------------------------------------------------------
  subroutine detToTanVectorWcs(xDet, yDet, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------
  ! This subroutine reprojects positions from the CAMCOORD2 (detx/y)
  ! coordinate system into image or tangent plane coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: xDet(:),&
                                                 yDet(size(xDet))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(size(xDet)),&
                                                 yTan(size(xDet))
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double), dimension(:), allocatable :: tanThetaArcsec, phi
    integer(int32) :: numPositions

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: detx/y to image tangent &
!      &coordinates.')

    xTan = 0.0d0 ! default
    yTan = 0.0d0 ! default

    numPositions = size(xDet)
    allocate(tanThetaArcsec(numPositions), phi(numPositions))
    call detToTelVector(xDet, yDet, tanThetaArcsec, phi)

    call telToTanVector(tanThetaArcsec, phi&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
      , scAttitude, timeStamp, xTan, yTan, isNewAtt)

  end subroutine detToTanVectorWcs

  !---------------------------------------------------------------------
  subroutine detToTanArrayWcs(xDet, yDet, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: xDet(:,:),&
                                                 yDet(size(xDet,1),size(xDet,2))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(&
                                                   size(xDet,1),size(xDet,2)),&
                                                 yTan(size(xDet,1),size(xDet,2))
    logical(bool), optional,      intent(in)  :: isNewAtt

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xDetVec(:), yDetVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xDet, 1)
    ySize = size(xDet, 2)
    vSize = xSize * ySize

    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    xDetVec = reshape(xDet, (/vSize/))
    yDetVec = reshape(yDet, (/vSize/))

    call detToTanVector(xDetVec, yDetVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine detToTanArrayWcs


  !---------------------------------------------------------------------
  subroutine telToTanScalar(tanThetaArcsec, phi, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec,&
                                                 phi
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan,&
                                                 yTan
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) :: xTanVec(1), yTanVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call telToTanVector((/tanThetaArcsec/), (/phi/), refRa, refDec, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine telToTanScalar

  !---------------------------------------------------------------------
  subroutine telToTanVector(tanThetaArcsec, phi, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------
  ! This subroutine reprojects positions from the TELCOORD polar
  ! coordinate system into image or tangent plane coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:),&
                                                 phi(size(tanThetaArcsec))
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(size(tanThetaArcsec)),&
                                                 yTan(size(tanThetaArcsec))
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) ::&
      tanTheta,&
      srcVector_ins(3), srcVector_ref(3),&
      skyMatrix_sc(3,3), instrumAttitudeMat_sky(3,3), refMatrix_sky(3,3)
    real(double), save :: refMatrix_ins(3,3)
    integer(int32) ::&
      numPositions, i
    logical(bool) :: localIsNewAtt, isFirstCall = .true.

!logical :: testFlag = .false.

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: TELCOORD to image tangent &
!      &coordinates.')

    if (present(isNewAtt)) then
      localIsNewAtt = isNewAtt
    else
      localIsNewAtt = .true.
    end if

    xTan = 0.0d0 ! default
    yTan = 0.0d0 ! default

    numPositions = size(tanThetaArcsec)

    if (localIsNewAtt) then
      call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
      instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
        , timeStamp)

      refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)
      refMatrix_ins = matmul(refMatrix_sky, transpose(instrumAttitudeMat_sky))

    else if (isFirstCall) then
call error('uninitializedRefMatrix',"ssclib:telToTanVector: the routine &
&is being called wrongly.")
    end if

    do i = 1, numPositions
!      tanTheta = tan(thetaArcsec(i - 1) * ArcsecToRad)
      tanTheta = tanThetaArcsec(i) * ArcsecToRad
      srcVector_ins(1) = 1.0d0
      srcVector_ins(2) = -tanTheta * sin(phi(i))
      srcVector_ins(3) =  tanTheta * cos(phi(i))
      ! NOTE that srcVector in this formulation is not generally of length 1.

      srcVector_ref = matmul(refMatrix_ins, srcVector_ins)

      if (srcVector_ref(1) <= 0.0)&
call error('angleGreaterThan90',"ssclib:telToTanVector: The input &
&position is further than 90 deg from the tangent (reference) direction.")

      xTan(i) = -srcVector_ref(2) / srcVector_ref(1)
      yTan(i) =  srcVector_ref(3) / srcVector_ref(1)
    end do

    isFirstCall = .false.

  end subroutine telToTanVector

  !---------------------------------------------------------------------
  subroutine telToTanArray(tanThetaArcsec, phi, refRa, refDec, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:,:),&
                                                 phi(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))
    real(double),                 intent(in)  :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2)),&
                                                 yTan(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))
    logical(bool), optional,      intent(in)  :: isNewAtt

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanThetaArcsec, 1)
    ySize = size(tanThetaArcsec, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi,         (/vSize/))

    call telToTanVector(tanThetaArcsecVec, phiVec, refRa, refDec, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine telToTanArray

  !---------------------------------------------------------------------
  subroutine telToTanScalarWcs(tanThetaArcsec, phi, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec,&
                                                 phi
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan,&
                                                 yTan
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) :: xTanVec(1), yTanVec(1)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    call telToTanVector((/tanThetaArcsec/), (/phi/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine telToTanScalarWcs

  !---------------------------------------------------------------------
  subroutine telToTanVectorWcs(tanThetaArcsec, phi, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------
  ! This subroutine reprojects positions from the TELCOORD polar
  ! coordinate system into image or tangent plane coordinates.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:),&
                                                 phi(size(tanThetaArcsec))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(size(tanThetaArcsec)),&
                                                 yTan(size(tanThetaArcsec))
    logical(bool), optional,      intent(in)  :: isNewAtt

    real(double) ::&
      tanTheta,&
      srcVector_ins(3), srcVector_ref(3),&
      skyMatrix_sc(3,3), instrumAttitudeMat_sky(3,3), refMatrix_sky(3,3)
    real(double), save :: refMatrix_ins(3,3)
    integer(int32) ::&
      numPositions, i
    logical(bool) :: localIsNewAtt, isFirstCall = .true.

!logical :: testFlag = .false.

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: TELCOORD to image tangent &
!      &coordinates.')

    if (present(isNewAtt)) then
      localIsNewAtt = isNewAtt
    else
      localIsNewAtt = .true.
    end if

    xTan = 0.0d0 ! default
    yTan = 0.0d0 ! default

    numPositions = size(tanThetaArcsec)

    if (localIsNewAtt) then
      call OAL_toAttitudeMatrix(scAttitude, skyMatrix_sc)
      instrumAttitudeMat_sky = getInstrumentAttitudeMat(transpose(skyMatrix_sc)&
        , timeStamp)

      refMatrix_sky = calcRefMatrixInSkyBasis(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180)
      refMatrix_ins = matmul(refMatrix_sky, transpose(instrumAttitudeMat_sky))

    else if (isFirstCall) then
call error('uninitializedRefMatrix',"ssclib:telToTanVector: the routine &
&is being called wrongly.")
    end if

    do i = 1, numPositions
!      tanTheta = tan(thetaArcsec(i - 1) * ArcsecToRad)
      tanTheta = tanThetaArcsec(i) * ArcsecToRad
      srcVector_ins(1) = 1.0d0
      srcVector_ins(2) = -tanTheta * sin(phi(i))
      srcVector_ins(3) =  tanTheta * cos(phi(i))
      ! NOTE that srcVector in this formulation is not generally of length 1.

      srcVector_ref = matmul(refMatrix_ins, srcVector_ins)

      if (srcVector_ref(1) <= 0.0)&
call error('angleGreaterThan90',"ssclib:telToTanVectorWcs: The input &
&position is further than 90 deg from the tangent (reference) direction.")

      xTan(i) = -srcVector_ref(2) / srcVector_ref(1)
      yTan(i) =  srcVector_ref(3) / srcVector_ref(1)
    end do

    isFirstCall = .false.

  end subroutine telToTanVectorWcs

  !---------------------------------------------------------------------
  subroutine telToTanArrayWcs(tanThetaArcsec, phi, wcs, scAttitude&
  , timeStamp, xTan, yTan, isNewAtt)
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanThetaArcsec(:,:),&
                                                 phi(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))
    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: xTan(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2)),&
                                                 yTan(&
                                                   size(tanThetaArcsec,1),&
                                                   size(tanThetaArcsec,2))
    logical(bool), optional,      intent(in)  :: isNewAtt

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !  SpacecraftAttitudeType typedef from caloalutils/src/caloaldefs.f90

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanThetaArcsec, 1)
    ySize = size(tanThetaArcsec, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi,         (/vSize/))

    call telToTanVector(tanThetaArcsecVec, phiVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
      , timeStamp, xTanVec, yTanVec, isNewAtt)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine telToTanArrayWcs


  !---------------------------------------------------------------------
  subroutine rawToInstScalar(xRaw, yRaw, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw,&
                                   yRaw
    real(double),   intent(out) :: xDet,&
                                   yDet,&
                                   tanThetaArcsec,&
                                   phi

    real(double) :: xDetVec(1), yDetVec(1), tanThetaArcsecVec(1), phiVec(1)

    !-------------------------------------------------------------------
    ! Start

    call rawToInstVector((/xRaw/), (/yRaw/), xDetVec, yDetVec&
      , tanThetaArcsecVec, phiVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)
    tanThetaArcsec = tanThetaArcsecVec(1)
    phi            = phiVec(1)

  end subroutine rawToInstScalar

  !---------------------------------------------------------------------
  subroutine rawToInstVector(xRaw, yRaw, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions in the PIXCOORD1 (also known as
  ! RAWX/Y) system to the TELCOORD polar system.
  !
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

!*** mention no alteration of randomization state.

    implicit none

    integer(int16), intent(in)  :: xRaw(:),&
                                   yRaw(size(xRaw))
    real(double),   intent(out) :: xDet(size(xRaw)),&
                                   yDet(size(xRaw)),&
                                   tanThetaArcsec(size(xRaw)),&
                                   phi(size(xRaw))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z, localTanThetaArcsec, localPhi
    real(double) :: mmToDetUnits

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: rawx/y to TELCOORD.')

    call CAL_rawXY2mm(xRaw, yRaw, cc2X, cc2Y, cc2Z, coordSysType=CAMCOORD2)
    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits

    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, localTanThetaArcsec&
      , localPhi)
    
    tanThetaArcsec = localTanThetaArcsec
    phi            = localPhi

    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)
    call CAL_releaseMemory(localTanThetaArcsec)
    call CAL_releaseMemory(localPhi)

  end subroutine rawToInstVector

  !---------------------------------------------------------------------
  subroutine rawToInstArray(xRaw, yRaw, xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw(:,:),&
                                   yRaw(size(xRaw,1),size(xRaw,2))
    real(double),   intent(out) :: xDet(size(xRaw,1),size(xRaw,2)),&
                                   yDet(size(xRaw,1),size(xRaw,2)),&
                                   tanThetaArcsec(size(xRaw,1),size(xRaw,2)),&
                                   phi(size(xRaw,1),size(xRaw,2))

    integer(int32) :: xSize, ySize, vSize
    integer(int16), allocatable :: xRawVec(:), yRawVec(:)
    real(double), allocatable ::&
      xDetVec(:), yDetVec(:), tanThetaArcsecVec(:), phiVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xRaw, 1)
    ySize = size(xRaw, 2)
    vSize = xSize * ySize

    allocate(xRawVec(vSize), yRawVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xRawVec = reshape(xRaw, (/vSize/))
    yRawVec = reshape(yRaw, (/vSize/))

    call rawToInstVector(xRawVec, yRawVec, xDetVec, yDetVec&
      , tanThetaArcsecVec, phiVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))
    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi            = reshape(phiVec,            (/xSize, ySize/))

  end subroutine rawToInstArray


  !---------------------------------------------------------------------
  subroutine rawToDetScalar(xRaw, yRaw, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw,&
                                   yRaw
    real(double),   intent(out) :: xDet,&
                                   yDet

    real(double) :: xDetVec(1), yDetVec(1)

    !-------------------------------------------------------------------
    ! Start

    call rawToDetVector((/xRaw/), (/yRaw/), xDetVec, yDetVec)

    xDet = xDetVec(1)
    yDet = yDetVec(1)

  end subroutine rawToDetScalar

  !---------------------------------------------------------------------
  subroutine rawToDetVector(xRaw, yRaw, xDet, yDet)
  !---------------------------------------------------------------------
  ! This subroutine converts positions in the PIXCOORD1 (also known as
  ! RAWX/Y) system to the CAMCOORD2 (detx/y) system.
  !
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

!*** mention no alteration of randomization state.

    implicit none

    integer(int16), intent(in)  :: xRaw(:),&
                                   yRaw(size(xRaw))
    real(double),   intent(out) :: xDet(size(xRaw)),&
                                   yDet(size(xRaw))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z
    real(double) :: mmToDetUnits

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: rawx/y to detx/y.')

    mmToDetUnits = calcMmToDetConversionFactor()
    call CAL_rawXY2mm(xRaw, yRaw, cc2X, cc2Y, cc2Z, coordSysType=CAMCOORD2)
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine rawToDetVector

  !---------------------------------------------------------------------
  subroutine rawToDetArray(xRaw, yRaw, xDet, yDet)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw(:,:),&
                                   yRaw(size(xRaw,1),size(xRaw,2))
    real(double),   intent(out) :: xDet(size(xRaw,1),size(xRaw,2)),&
                                   yDet(size(xRaw,1),size(xRaw,2))

    integer(int32) :: xSize, ySize, vSize
    integer(int16), allocatable :: xRawVec(:), yRawVec(:)
    real(double), allocatable :: xDetVec(:), yDetVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xRaw, 1)
    ySize = size(xRaw, 2)
    vSize = xSize * ySize

    allocate(xRawVec(vSize), yRawVec(vSize))
    allocate(xDetVec(vSize), yDetVec(vSize))
    xRawVec = reshape(xRaw, (/vSize/))
    yRawVec = reshape(yRaw, (/vSize/))

    call rawToDetVector(xRawVec, yRawVec, xDetVec, yDetVec)

    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))

  end subroutine rawToDetArray


  !---------------------------------------------------------------------
  subroutine rawToTelScalar(xRaw, yRaw, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw,&
                                   yRaw
    real(double),   intent(out) :: tanThetaArcsec,&
                                   phi

    real(double) :: tanThetaArcsecVec(1), phiVec(1)

    !-------------------------------------------------------------------
    ! Start

    call rawToTelVector((/xRaw/), (/yRaw/), tanThetaArcsecVec, phiVec)

    tanThetaArcsec = tanThetaArcsecVec(1)
    phi            = phiVec(1)

  end subroutine rawToTelScalar

  !---------------------------------------------------------------------
  subroutine rawToTelVector(xRaw, yRaw, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions in the PIXCOORD1 (also known as
  ! RAWX/Y) system to the TELCOORD polar system.
  !
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

!*** mention no alteration of randomization state.

    implicit none

    integer(int16), intent(in)  :: xRaw(:),&
                                   yRaw(size(xRaw))
    real(double),   intent(out) :: tanThetaArcsec(size(xRaw)),&
                                   phi(size(xRaw))

    real(double), dimension(:), pointer ::&
      cc2X, cc2Y, cc2Z, localTanThetaArcsec, localPhi

    !-------------------------------------------------------------------
    ! Start

!    call message(AppMsg, Noisy, 'ssclib: rawx/y to TELCOORD.')

    call CAL_rawXY2mm(xRaw, yRaw, cc2X, cc2Y, cc2Z, coordSysType=CAMCOORD2)
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, localTanThetaArcsec&
      , localPhi)
    
    tanThetaArcsec = localTanThetaArcsec
    phi            = localPhi

    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)
    call CAL_releaseMemory(localTanThetaArcsec)
    call CAL_releaseMemory(localPhi)

  end subroutine rawToTelVector

  !---------------------------------------------------------------------
  subroutine rawToTelArray(xRaw, yRaw, tanThetaArcsec, phi)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in)  :: xRaw(:,:),&
                                   yRaw(size(xRaw,1),size(xRaw,2))
    real(double),   intent(out) :: tanThetaArcsec(size(xRaw,1),size(xRaw,2)),&
                                   phi(size(xRaw,1),size(xRaw,2))

    integer(int32) :: xSize, ySize, vSize
    integer(int16), allocatable :: xRawVec(:), yRawVec(:)
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xRaw, 1)
    ySize = size(xRaw, 2)
    vSize = xSize * ySize

    allocate(xRawVec(vSize), yRawVec(vSize))
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    xRawVec = reshape(xRaw, (/vSize/))
    yRawVec = reshape(yRaw, (/vSize/))

    call rawToTelVector(xRawVec, yRawVec, tanThetaArcsecVec, phiVec)

    tanThetaArcsec = reshape(tanThetaArcsecVec, (/xSize, ySize/))
    phi         = reshape(phiVec,         (/xSize, ySize/))

  end subroutine rawToTelArray


  !---------------------------------------------------------------------
  subroutine detToRawScalar(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and (if ccdNum is not requested) ccdChipId and, if instrument is
  ! EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: xDet, yDet
    real(double),  intent(out)           :: xRaw, yRaw
    logical(bool), intent(out)           :: isOffChip
    integer(int8), intent(out), optional :: ccdNum

    real(double)  :: xRawVec(1), yRawVec(1)
    logical(bool) :: isOffChipVec(1)
    integer(int8) :: ccdNumVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(ccdNum)) then
      call detToRawVector((/xDet/), (/yDet/), xRawVec, yRawVec, isOffChipVec&
        , ccdNumVec)

      ccdNum = ccdNumVec(1)
    else
      call detToRawVector((/xDet/), (/yDet/), xRawVec, yRawVec, isOffChipVec)
    end if

    xRaw  = xRawVec(1)
    yRaw  = yRawVec(1)
    isOffChip = isOffChipVec(1)

  end subroutine detToRawScalar

  !---------------------------------------------------------------------
  subroutine detToRawVector(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: xDet(:),&
                                            yDet(size(xDet))
    real(double),  intent(out)           :: xRaw(size(xDet)),&
                                            yRaw(size(xDet))
    logical(bool), intent(out)           :: isOffChip(size(xDet))
    integer(int8), intent(out), optional :: ccdNum(size(xDet))

!    real(double), dimension(:), pointer :: cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z
real(double), dimension(:), allocatable :: cc2X, cc2Y, cc2Z
real(double), dimension(:), pointer :: cc1X, cc1Y, cc1Z
    real(double) :: mmToDetUnits
    integer(int32) :: numPoints

    !-------------------------------------------------------------------
    ! Start

    numPoints = size(xDet)
    mmToDetUnits = calcMmToDetConversionFactor()
    allocate(cc2X(numPoints), cc2Y(numPoints), cc2Z(numPoints))
    cc2X = xDet / mmToDetUnits
    cc2Y = yDet / mmToDetUnits
    cc2Z = 0.0d0
    call CAL_camCoord2ToCamCoord1(cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z)
!    deallocate(cc2X, cc2Y, cc2Z)

    if (present(ccdNum)) then
      call cc1ToRawMasterCcd(cc1X, cc1Y, xRaw, yRaw, isOffChip&
        , ccdNum)
    else
      call cc1ToRawMaster(cc1X, cc1Y, xRaw, yRaw, isOffChip)
    end if

    call CAL_releaseMemory(cc1X, cc1Y, cc1Z)

  end subroutine detToRawVector

  !---------------------------------------------------------------------
  subroutine detToRawArray(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: xDet(:,:),&
                                            yDet(size(xDet,1),size(xDet,2))
    real(double),  intent(out)           :: xRaw(size(xDet,1),size(xDet,2)),&
                                            yRaw(size(xDet,1),size(xDet,2))
    logical(bool), intent(out)           :: isOffChip(size(xDet,1),size(xDet,2))
    integer(int8), intent(out), optional :: ccdNum(size(xDet,1),size(xDet,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable ::&
      xDetVec(:), yDetVec(:), xRawVec(:), yRawVec(:)
    logical(bool), allocatable ::&
      isOffChipVec(:)
    integer(int8), allocatable :: ccdNumVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xDet, 1)
    ySize = size(xDet, 2)
    vSize = xSize * ySize

    allocate(xDetVec(vSize), yDetVec(vSize))
    allocate(xRawVec(vSize), yRawVec(vSize), isOffChipVec(vSize))
    xDetVec = reshape(xDet, (/vSize/))
    yDetVec = reshape(yDet, (/vSize/))

    if (present(ccdNum)) then
      allocate(ccdNumVec(vSize))
      call detToRawVector(xDetVec, yDetVec, xRawVec, yRawVec, isOffChipVec&
        , ccdNumVec)

      ccdNum = reshape(ccdNumVec, (/xSize, ySize/))
    else
      call detToRawVector(xDetVec, yDetVec, xRawVec, yRawVec, isOffChipVec)
    end if

    xRaw      = reshape(xRawVec,      (/xSize, ySize/))
    yRaw      = reshape(yRawVec,      (/xSize, ySize/))
    isOffChip = reshape(isOffChipVec, (/xSize, ySize/))

  end subroutine detToRawArray


  !---------------------------------------------------------------------
  subroutine telToRawScalar(tanThetaArcsec, phi, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: tanThetaArcsec, phi
    real(double),  intent(out)           :: xRaw, yRaw
    logical(bool), intent(out)           :: isOffChip
    integer(int8), intent(out), optional :: ccdNum

    real(double)  :: xRawVec(1), yRawVec(1)
    logical(bool) :: isOffChipVec(1)
    integer(int8) :: ccdNumVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(ccdNum)) then
      call telToRawVector((/tanThetaArcsec/), (/phi/), xRawVec, yRawVec&
        , isOffChipVec, ccdNumVec)

      ccdNum = ccdNumVec(1)
    else
      call telToRawVector((/tanThetaArcsec/), (/phi/), xRawVec, yRawVec&
        , isOffChipVec)
    end if

    xRaw  = xRawVec(1)
    yRaw  = yRawVec(1)
    isOffChip = isOffChipVec(1)

  end subroutine telToRawScalar

  !---------------------------------------------------------------------
  subroutine telToRawVector(tanThetaArcsec, phi, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: tanThetaArcsec(:),&
                                            phi(size(tanThetaArcsec))
    real(double),  intent(out)           :: xRaw(size(tanThetaArcsec)),&
                                            yRaw(size(tanThetaArcsec))
    logical(bool), intent(out)           :: isOffChip(size(tanThetaArcsec))
    integer(int8), intent(out), optional :: ccdNum(size(tanThetaArcsec))

    real(double), dimension(:), pointer :: cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z

    !-------------------------------------------------------------------
    ! Start

    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)
    call CAL_camCoord2ToCamCoord1(cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z)
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

    if (present(ccdNum)) then
      call cc1ToRawMasterCcd(cc1X, cc1Y, xRaw, yRaw, isOffChip, ccdNum)
    else
      call cc1ToRawMaster(cc1X, cc1Y, xRaw, yRaw, isOffChip)
    end if

    call CAL_releaseMemory(cc1X, cc1Y, cc1Z)

  end subroutine telToRawVector

  !---------------------------------------------------------------------
  subroutine telToRawArray(tanThetaArcsec, phi, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: tanThetaArcsec(:,:),&
                                            phi(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    real(double),  intent(out)           :: xRaw(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2)),&
                                            yRaw(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    logical(bool), intent(out)           :: isOffChip(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    integer(int8), intent(out), optional :: ccdNum(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable ::&
      tanThetaArcsecVec(:), phiVec(:), xRawVec(:), yRawVec(:)
    logical(bool), allocatable ::&
      isOffChipVec(:)
    integer(int8), allocatable :: ccdNumVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanThetaArcsec, 1)
    ySize = size(tanThetaArcsec, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    allocate(xRawVec(vSize), yRawVec(vSize), isOffChipVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi,            (/vSize/))

    if (present(ccdNum)) then
      allocate(ccdNumVec(vSize))
      call telToRawVector(tanThetaArcsecVec, phiVec, xRawVec, yRawVec&
        , isOffChipVec, ccdNumVec)

      ccdNum = reshape(ccdNumVec, (/xSize, ySize/))
    else
      call telToRawVector(tanThetaArcsecVec, phiVec, xRawVec, yRawVec&
        , isOffChipVec)
    end if

    xRaw      = reshape(xRawVec,      (/xSize, ySize/))
    yRaw      = reshape(yRawVec,      (/xSize, ySize/))
    isOffChip = reshape(isOffChipVec, (/xSize, ySize/))

  end subroutine telToRawArray


  !---------------------------------------------------------------------
  subroutine cc1ToRawMaster(xCc1, yCc1, xRaw, yRaw, isOffChip)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)  :: xCc1(:),&
                                  yCc1(size(xCc1))
    real(double),  intent(out) :: xRaw(size(xCc1)),&
                                  yRaw(size(xCc1))
    logical(bool), intent(out) :: isOffChip(size(xCc1))

    integer(int32) :: instrumentId
    integer(int16) :: chipXsize, chipYsize

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Get max rawx and rawy and construct chip corners pseudo-vector:

    call CAL_getState(instrument=instrumentId)
    if (instrumentId == EPN) then
      chipXsize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYsize = EPN_MAX_Y

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXsize = EMOS_MAX_X
      chipYsize = EMOS_MAX_Y

    else
call error('badInstrument','ssclib:cc1ToRawMaster: Only EPIC instruments permitted.')
    end if

    call cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, xRaw, yRaw)

    ! Now check to see which values are off chip:
    isOffChip = .false. ! default
    where(nint(xRaw) < 1 .or. nint(xRaw) > chipXsize&
    .or.  nint(yRaw) < 1 .or. nint(yRaw) > chipYsize)&
      isOffChip = .true.

  end subroutine cc1ToRawMaster

  !---------------------------------------------------------------------
  subroutine cc1ToRawMasterCcd(xCc1, yCc1, xRaw, yRaw, isOffChip, ccdNums)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument.

    implicit none

    real(double),  intent(in)  :: xCc1(:),&
                                  yCc1(size(xCc1))
    real(double),  intent(out) :: xRaw(size(xCc1)),&
                                  yRaw(size(xCc1))
    logical(bool), intent(out) :: isOffChip(size(xCc1))
    integer(int8), intent(out) :: ccdNums(size(xCc1))

    real(double), allocatable ::&
      tempCc1X(:), tempCc1Y(:), tempCc1Z(:),&
!zCc1(:), 
      trialXRaw(:), trialYRaw(:), minXRaw(:), minYRaw(:)
    integer(int32) ::&
      instrumentId, numPoints, trialCcdNum, previousCcdNum, numCcds
    integer(int16) :: chipXsize, chipYsize
    integer(int16), pointer :: chipX(:), chipY(:)
    integer(int8), pointer :: onCcdNums(:)
    real(single), allocatable ::&
      minSeps(:), trialMinSeps(:)
    real(single) :: chipXCentre, chipYCentre
    integer(int8), allocatable :: minCcdNums(:)

    !-------------------------------------------------------------------
    ! Start

    numPoints = size(xCc1)

!    allocate(zCc1(numPoints))
!    zCc1 = 0.0d0
allocate(tempCc1X(numPoints), tempCc1Y(numPoints), tempCc1Z(numPoints))
tempCc1X = xCc1
tempCc1Y = yCc1
tempCc1Z = 0.0d0
!    call CAL_camCoord1ToChipCoord(xCc1, yCc1, zCc1, chipX, chipY, onCcdNums)
call CAL_camCoord1ToChipCoord(tempCc1X, tempCc1Y, tempCc1Z&
 , chipX, chipY, onCcdNums)

    call CAL_releaseMemory(chipX, chipY)

    ! Note that if position (xCc1, yCc1) does not lie on any CCD, the
    ! returned values of chipX, chipY and onCcdNums will all be zero.

    where(onCcdNums .ne. 0) isOffChip = .false.

    !...................................................................
    ! Get chip limits:

    call CAL_getState(instrument=instrumentId)
    if (instrumentId == EPN) then
      chipXsize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYsize = EPN_MAX_Y
      numCcds = 12

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXsize = EMOS_MAX_X
      chipYsize = EMOS_MAX_Y
      numCcds = 7

    else
call error('badInstrument','ssclib:cc1ToRawMasterCcd: Only EPIC instruments permitted.')
    end if

    chipXCentre = (1 + chipXsize) / 2.0
    chipYCentre = (1 + chipYsize) / 2.0

    !...................................................................
    ! Now calculate RAWXY values and the CCDNR of the nearest chip:

    call CAL_getState(ccdChipId=previousCcdNum)

    trialCcdNum = 1
    call CAL_setState(ccdChipId=trialCcdNum)
!***** what about the node????


    allocate(minCcdNums(numPoints))
    allocate(trialXRaw(numPoints), trialYRaw(numPoints))
    allocate(minXRaw(numPoints), minYRaw(numPoints))
    allocate(minSeps(numPoints), trialMinSeps(numPoints))

    call cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, trialXRaw, trialYRaw)

    minSeps = max(abs(trialXRaw - chipXcentre)&
      , abs(trialYRaw - chipYcentre))

    where(.not. isOffChip .and. onCcdNums==1)
      xRaw = trialXRaw
      yRaw = trialYRaw
    elsewhere
      minXRaw = trialXRaw
      minYRaw = trialYRaw
      minCcdNums = 1
    end where

    do trialCcdNum = 2, numCcds
      call CAL_setState(ccdChipId=trialCcdNum)
      call cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, trialXRaw, trialYRaw)

      trialMinSeps = max(abs(trialXRaw - chipXcentre)&
        , abs(trialYRaw - chipYcentre))

      where(.not. isOffChip .and. onCcdNums==trialCcdNum)
        xRaw = trialXRaw
        yRaw = trialYRaw
      elsewhere(trialMinSeps < minSeps)
        minSeps = trialMinSeps
        minXRaw = trialXRaw
        minYRaw = trialYRaw
        minCcdNums = trialCcdNum
      end where
    end do

    where(isOffChip)
      xRaw = minXRaw
      yRaw = minYRaw
      ccdNums = minCcdNums
    elsewhere
      ccdNums = onCcdNums
    end where

    call CAL_setState(ccdChipId=previousCcdNum)

  end subroutine cc1ToRawMasterCcd

  !---------------------------------------------------------------------
  subroutine cc1ToRawSlave(chipXsize, chipYsize&
  , cc1X, cc1Y, trialXRaw, trialYRaw)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    integer(int16), intent(in) :: chipXsize, chipYsize
    real(double), intent(in) :: cc1X(:), cc1Y(size(cc1X))
!    type(Point2dDbleT), intent(out) :: trialRawReal(size(cc1X))
    real(double), intent(out) :: trialXRaw(size(cc1X)), trialYRaw(size(cc1X))

    real(double) :: mat(2,2), offsetCc1X, offsetCc1Y

    !-------------------------------------------------------------------
    ! Start

    !.................................................................
    ! Calculate the transform matrix and offset to enable the following
    ! transform:
    !
    !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
    !	(      ) = (                    ) * (                   )
    !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )

    call calcTransformAndOffset(chipXsize, chipYsize&
      , mat, offsetCc1X, offsetCc1Y)

    !.................................................................
    ! Use this matrix to do all the conversions:

    call linearTransform(mat, cc1X, cc1Y, offsetCc1X, offsetCc1Y&
      , trialXRaw, trialYRaw)

  end subroutine cc1ToRawSlave


  !---------------------------------------------------------------------
  subroutine calcTransformAndOffset(chipXsize, chipYsize&
  , mat, offsetCc1X, offsetCc1Y)
  !---------------------------------------------------------------------
  ! The two big problems with the present call CAL_camCoord1ToChipCoord()
  ! are (i) the returned CHIPCOORD values are integer; (ii) off-chip
  ! positions are not defined. The next two subroutines are designed to
  ! work around this.
  !
  ! The present subroutine calculates the transform matrix and offset to
  ! enable the following transform:
  !
  !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
  !	(      ) = (                    ) * (                   )	(1)
  !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )
  !
  ! This is done by calculating the elements of the inverse matrix tam()
  ! which satisfies the relation
  !
  !	( cc1X )   ( tam(1,1)  tam(1,2) )   ( rawX )   ( offsetCc1X )
  !	(      ) = (                    ) * (      ) + (            ),	(2)
  !	( cc1Y )   ( tam(2,1)  tam(2,2) )   ( rawY )   ( offsetCc1Y )
  !
  ! then by inverting to get mat() = tam()^-1.
  !
  ! The matrix tam() and the offset components are obtained as follows.
  ! Four pairs of rawX,Y values are taken from the corners of the CCD. The
  ! X and Y components are loaded into 4-element vectors for better ease
  ! of use of the CAL coordinate conversion routine(s). The order of the
  ! corners in these pseudo-vectors is as follows:
  !
  !	      -----------
  !	     |2         3|
  !	     |           |
  !	  ^  |           |
  !	  |  |           |
  !	R |  |           |
  !	A |  |1         4|
  !	W |   -----------
  !	Y |
  !	   -------->
  !	      RAWX
  !
  ! The CAL is used to convert these rawX,Y values to components in the
  ! CAMCOORD1 system. At the end of this step we have 4 sets of
  ! equation (2), one each for the four corners of the CCD.
  !
  ! Note now that if we subtract equation (2) for point 2 from that for
  ! point 3, then divide by (chipXsize - 1), the RHS is
  !
  !	( tam(1,1)  tam(1,2) )   ( 1 )   ( tam(1,1) )
  !	(                    ) * (   ) = (          ).	(3)
  !	( tam(2,1)  tam(2,2) )   ( 0 )   ( tam(2,1) )
  !
  ! Subtraction of equation (2) for point 1 from 4 gives the same result.
  ! Thus these two subtractions can be used to calculate tam(1,1) and
  ! tam(2,1) by averaging, viz:
  !
  !	(tam(1,1))   (cc1X(3)-cc1X(2)+cc1X(4)-cc1X(1))       0.5
  !	(        ) = (                               ) * -------------.	(4)
  !	(tam(2,1))   (cc1Y(3)-cc1Y(2)+cc1Y(4)-cc1Y(1))    chipXsize-1
  !
  ! (There is no real need to average, it just makes full (and
  ! symmetrical) use of the available data.) A similar equation yields
  ! the other two tam() components.
  !
  ! The offset vector can in principle be obtained from any of the four
  ! sets of corner values by doing
  !
  !	( offsetCc1X )   ( cc1X )   ( tam(1,1)  tam(1,2) )   ( rawX )
  !	(            ) = (      ) - (                    ) * (      ).	(5)
  !	( offsetCc1Y )   ( cc1Y )   ( tam(2,1)  tam(2,2) )   ( rawY )
  !
  ! However again it is good to average, hence in this routine the offset
  ! vector is computed for each of the four corners and the results averaged.
  !
  ! Note that the matrix indices are ordered in the (row, col) sense not the
  ! (x, y) sense.
  !
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument,
  ! ccdChipId and (if instrument is EMOS1 or EMOS2) ccdNodeId.
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: chipXsize, chipYsize
    real(double), intent(out) :: mat(2,2), offsetCc1X, offsetCc1Y

    integer(int16) ::&
      chipCornersRawX(4), chipCornersRawY(4), i
    real(double), pointer ::&
      chipCornersCc1X(:), chipCornersCc1Y(:), chipCornersCc1Z(:)
    real(double) :: tam(2,2), det
    logical :: oldRandomize

    !-------------------------------------------------------------------
    ! Start

    chipCornersRawX(1) = 1
    chipCornersRawX(2) = 1
    chipCornersRawX(3) = chipXsize
    chipCornersRawX(4) = chipXsize

    chipCornersRawY(1) = 1
    chipCornersRawY(2) = chipYsize
    chipCornersRawY(3) = chipYsize
    chipCornersRawY(4) = 1

    call CAL_getState(randomize=oldRandomize)
    call CAL_setState(randomize=.false.)
    call CAL_rawXY2mm(chipCornersRawX, chipCornersRawY&
      , chipCornersCc1X, chipCornersCc1Y, chipCornersCc1Z&
      , coordSysType=CAMCOORD1)

    call CAL_setState(randomize=oldRandomize)

    !...................................................................
    ! Set up CAMCOORD1 -> PIXCOORD linear transform matrix and offset vector:

    tam(1,1) = (chipCornersCc1X(3) - chipCornersCc1X(0)&
             +  chipCornersCc1X(2) - chipCornersCc1X(1))&
             * 0.5 / (real(chipXsize) - 1.0)

    tam(2,1) = (chipCornersCc1Y(3) - chipCornersCc1Y(0)&
             +  chipCornersCc1Y(2) - chipCornersCc1Y(1))&
             * 0.5 / (real(chipXsize) - 1.0)

    tam(1,2) = (chipCornersCc1X(1) - chipCornersCc1X(0)&
             +  chipCornersCc1X(2) - chipCornersCc1X(3))&
             * 0.5 / (real(chipYsize) - 1.0)

    tam(2,2) = (chipCornersCc1Y(1) - chipCornersCc1Y(0)&
             +  chipCornersCc1Y(2) - chipCornersCc1Y(3))&
             * 0.5 / (real(chipYsize) - 1.0)

    ! Invert:

    det = tam(1,1) * tam(2,2) - tam(1,2) * tam(2,1)
    mat(1,1) =  tam(2,2) / det
    mat(1,2) = -tam(1,2) / det
    mat(2,1) = -tam(2,1) / det
    mat(2,2) =  tam(1,1) / det

    !...................................................................
    ! Now calculate the offset vector:

    offsetCc1X = 0.0
    offsetCc1Y = 0.0
    do i = 1, 4
      offsetCc1X = offsetCc1X + chipCornersCc1X(i-1)&
        - (tam(1,1) * chipCornersRawX(i) + tam(1,2) * chipCornersRawY(i))
      offsetCc1Y = offsetCc1Y + chipCornersCc1Y(i-1)&
        - (tam(2,1) * chipCornersRawX(i) + tam(2,2) * chipCornersRawY(i))
    end do
    offsetCc1X = offsetCc1X / 4.0
    offsetCc1Y = offsetCc1Y / 4.0

    call CAL_releaseMemory(chipCornersCc1X, chipCornersCc1Y, chipCornersCc1Z)

  end subroutine calcTransformAndOffset

  !---------------------------------------------------------------------
!  subroutine linearTransform(mat, cc1X, cc1Y, offsetCc1X, offsetCc1Y, raw)
subroutine linearTransform(mat, cc1X, cc1Y, offsetCc1X, offsetCc1Y, xRaw, yRaw)
  !---------------------------------------------------------------------
  ! The two big problems with the present call CAL_camCoord1ToChipCoord()
  ! are (i) the returned CHIPCOORD values are integer; (ii) off-chip
  ! positions are not defined. The present and previous subroutines are
  ! designed to work around this.
  !
  ! The present subroutine calculates rawX,Y from the following relation:
  !
  !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
  !	(      ) = (                    ) * (                   )	(1)
  !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )
  !

    implicit none

    real(double),       intent(in)  :: mat(2,2),&
                                       cc1X(:),&
                                       cc1Y(size(cc1X)),&
                                       offsetCc1X,&
                                       offsetCc1Y
!    type(Point2dDbleT), intent(out) :: raw(size(cc1X))
!    ! type defined in ssclib/geometric_types.
real(double), intent(out) :: xRaw(size(cc1X)), yRaw(size(cc1X))

    !-------------------------------------------------------------------
    ! Start

!    raw%x = mat(1,1) * (cc1X - offsetCc1X) + mat(1,2) * (cc1Y - offsetCc1Y)
!    raw%y = mat(2,1) * (cc1X - offsetCc1X) + mat(2,2) * (cc1Y - offsetCc1Y)
xRaw = mat(1,1) * (cc1X - offsetCc1X) + mat(1,2) * (cc1Y - offsetCc1Y)
yRaw = mat(2,1) * (cc1X - offsetCc1X) + mat(2,2) * (cc1Y - offsetCc1Y)

  end subroutine linearTransform



  !---------------------------------------------------------------------
  subroutine raDecToTanScalar(ra, dec, refRa, refDec, xTan, yTan)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: refRa, refDec, ra, dec
    real(double), intent(out) :: xTan, yTan

    real(double) :: xTanVec(1), yTanVec(1)

    !-------------------------------------------------------------------
    ! Start

    call raDecToTanVector((/ra/), (/dec/), refRa, refDec&
      , xTanVec, yTanVec)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine raDecToTanScalar

  !---------------------------------------------------------------------
  subroutine raDecToTanVector(ra, dec, refRa, refDec, xTan, yTan)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs to the image or tangent plane
  ! coordinate system.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.

    implicit none

    real(double), intent(in)  :: refRa,&
                                 refDec,&
                                 ra(:),&
                                 dec(size(ra))
    real(double), intent(out) :: xTan(size(ra)),&
                                 yTan(size(ra))

    real(double) ::&
      srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)
    integer(int32) :: numPositions, i

    !-------------------------------------------------------------------
    ! Start

    numPositions = size(ra)

    refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)

    do i = 1, numPositions
      srcVector_sky = skyToCartesian(ra(i), dec(i))
      srcVector_ref = matmul(refMatrix_sky, srcVector_sky)

      if (srcVector_ref(1) <= 0.0)&
call error('angleGreaterThan90',"ssclib:raDecToTanVector: The input &
&position is further than 90 deg from the tangent (reference) direction.")

      xTan(i) = -srcVector_ref(2) / srcVector_ref(1)
      yTan(i) =  srcVector_ref(3) / srcVector_ref(1)
    end do

  end subroutine raDecToTanVector

  !---------------------------------------------------------------------
  subroutine raDecToTanArray(ra, dec, refRa, refDec, xTan, yTan)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: refRa,&
                                 refDec,&
                                 ra(:,:),&
                                 dec(size(ra,1),size(ra,2))
    real(double), intent(out) :: xTan(size(ra,1),size(ra,2)),&
                                 yTan(size(ra,1),size(ra,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(ra, 1)
    ySize = size(ra, 2)
    vSize = xSize * ySize

    allocate(raVec(vSize), decVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    raVec  = reshape(ra,  (/vSize/))
    decVec = reshape(dec, (/vSize/))

    call raDecToTanVector(raVec, decVec, refRa, refDec, xTanVec, yTanVec)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine raDecToTanArray

  !---------------------------------------------------------------------
  subroutine raDecToTanScalarWcs(ra, dec, wcs, xTan, yTan)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: ra, dec
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: xTan, yTan

    real(double) :: xTanVec(1), yTanVec(1)

    !-------------------------------------------------------------------
    ! Start

    call raDecToTanVector((/ra/), (/dec/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
      , xTanVec, yTanVec)

    xTan = xTanVec(1)
    yTan = yTanVec(1)

  end subroutine raDecToTanScalarWcs

  !---------------------------------------------------------------------
  subroutine raDecToTanVectorWcs(ra, dec, wcs, xTan, yTan)
  !---------------------------------------------------------------------
  ! This subroutine projects RAs and decs to the image or tangent plane
  ! coordinate system.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.

    implicit none

    real(double), intent(in)  :: ra(:),&
                                 dec(size(ra))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: xTan(size(ra)),&
                                 yTan(size(ra))

    real(double) ::&
      srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)
    integer(int32) :: numPositions, i

    !-------------------------------------------------------------------
    ! Start

    numPositions = size(ra)

    refMatrix_sky = calcRefMatrixInSkyBasis(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180)

    do i = 1, numPositions
      srcVector_sky = skyToCartesian(ra(i), dec(i))
      srcVector_ref = matmul(refMatrix_sky, srcVector_sky)

      if (srcVector_ref(1) <= 0.0)&
call error('angleGreaterThan90',"ssclib:raDecToTanVectorWcs: The input &
&position is further than 90 deg from the tangent (reference) direction.")

      xTan(i) = -srcVector_ref(2) / srcVector_ref(1)
      yTan(i) =  srcVector_ref(3) / srcVector_ref(1)
    end do

  end subroutine raDecToTanVectorWcs

  !---------------------------------------------------------------------
  subroutine raDecToTanArrayWcs(ra, dec, wcs, xTan, yTan)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: ra(:,:),&
                                 dec(size(ra,1),size(ra,2))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: xTan(size(ra,1),size(ra,2)),&
                                 yTan(size(ra,1),size(ra,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: raVec(:), decVec(:)
    real(double), allocatable :: xTanVec(:), yTanVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(ra, 1)
    ySize = size(ra, 2)
    vSize = xSize * ySize

    allocate(raVec(vSize), decVec(vSize))
    allocate(xTanVec(vSize), yTanVec(vSize))
    raVec  = reshape(ra,  (/vSize/))
    decVec = reshape(dec, (/vSize/))

    call raDecToTanVector(raVec, decVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, xTanVec, yTanVec)

    xTan = reshape(xTanVec, (/xSize, ySize/))
    yTan = reshape(yTanVec, (/xSize, ySize/))

  end subroutine raDecToTanArrayWcs


  !---------------------------------------------------------------------
  subroutine tanToRaDecScalar(xTan, yTan, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTan, yTan, refRa, refDec
    real(double), intent(out) :: ra, dec

    real(double) :: raVec(1), decVec(1)

    !-------------------------------------------------------------------
    ! Start

    call tanToRaDecVector((/xTan/), (/yTan/), refRa, refDec, raVec, decVec)

    ra  = raVec(1)
    dec = decVec(1)

  end subroutine tanToRaDecScalar

  !---------------------------------------------------------------------
  subroutine tanToRaDecVector(xTan, yTan, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------
  ! This subroutine calculates RAs and decs which correspond to positions
  ! in the image or tangent plane coordinate system.
  !
  ! The tangent plane is defined as follows. For a direction of view that
  ! faces outwards, with declination increasing upward, the x-axis of the
  ! tangent plane points to the right (in the direction of decreasing ra)
  ! and the y-axis points up.

    implicit none

    real(double), intent(in)  :: refRa,&
                                 refDec,&
                                 xTan(:),&
                                 yTan(size(xTan))
    real(double), intent(out) :: ra(size(xTan)),&
                                 dec(size(xTan))

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)
!,&
!      norm
    integer(int32) :: numPositions, i

    !-------------------------------------------------------------------
    ! Start

    numPositions = size(xTan)

    refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)

    do i = 1, numPositions
!      norm = sqrt(1.0d0 + xTan(i) * xTan(i) + yTan(i) * yTan(i))
!      srcVector_ref(1) = 1.0d0 / norm
!      srcVector_ref(2) = -xTan(i) / norm
!      srcVector_ref(3) =  yTan(i) / norm
srcVector_ref(1) = 1.0d0
srcVector_ref(2) = -xTan(i)
srcVector_ref(3) =  yTan(i)
! NOTE! srcVector_ref is not normalized to length 1. cartesianToSky should be able to handle this though.

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky, ra(i), dec(i))
    end do

  end subroutine tanToRaDecVector

  !---------------------------------------------------------------------
  subroutine tanToRaDecArray(xTan, yTan, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: refRa,&
                                 refDec,&
                                 xTan(:,:),&
                                 yTan(size(xTan,1),size(xTan,2))
    real(double), intent(out) :: ra(size(xTan,1),size(xTan,2)),&
                                 dec(size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: raVec(:), decVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(raVec(vSize), decVec(vSize))
    xTanVec = reshape(xTan, (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToRaDecVector(xTanVec, yTanVec, refRa, refDec, raVec, decVec)

    ra  = reshape(raVec,  (/xSize, ySize/))
    dec = reshape(decVec, (/xSize, ySize/))

  end subroutine tanToRaDecArray

  !---------------------------------------------------------------------
  subroutine tanToRaDecScalarWcs(xTan, yTan, wcs, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTan, yTan
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: ra, dec

    real(double) :: raVec(1), decVec(1)
!, refRa, refDec

    !-------------------------------------------------------------------
    ! Start

    call tanToRaDecVector((/xTan/), (/yTan/)&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, raVec, decVec)

    ra  = raVec(1)
    dec = decVec(1)

  end subroutine tanToRaDecScalarWcs

  !---------------------------------------------------------------------
  subroutine tanToRaDecVectorWcs(xTan, yTan, wcs, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTan(:),&
                                 yTan(size(xTan))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: ra(size(xTan)),&
                                 dec(size(xTan))

!    real(double) :: refRa, refDec

    !-------------------------------------------------------------------
    ! Start

    call tanToRaDecVector(xTan, yTan&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, ra, dec)

  end subroutine tanToRaDecVectorWcs

  !---------------------------------------------------------------------
  subroutine tanToRaDecArrayWcs(xTan, yTan, wcs, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTan(:,:),&
                                 yTan(size(xTan,1),size(xTan,2))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: ra(size(xTan,1),size(xTan,2)),&
                                 dec(size(xTan,1),size(xTan,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: xTanVec(:), yTanVec(:)
    real(double), allocatable :: raVec(:), decVec(:)
!    real(double) :: refRa, refDec

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTan, 1)
    ySize = size(xTan, 2)
    vSize = xSize * ySize

    allocate(xTanVec(vSize), yTanVec(vSize))
    allocate(raVec(vSize), decVec(vSize))
    xTanVec = reshape(xTan, (/vSize/))
    yTanVec = reshape(yTan, (/vSize/))

    call tanToRaDecVector(xTanVec, yTanVec&
      , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, raVec, decVec)

    ra  = reshape(raVec,  (/xSize, ySize/))
    dec = reshape(decVec, (/xSize, ySize/))

  end subroutine tanToRaDecArrayWcs


  !---------------------------------------------------------------------
  subroutine polarsToRaDecScalar(theta, phi, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)            :: refRa,&
                                           refDec,&
                                           theta,&
                                           phi
    real(double), intent(out)           :: ra,&
                                           dec

    real(double) :: raVec(1), decVec(1)

    !-------------------------------------------------------------------
    ! Start

    call polarsToRaDecVector((/theta/), (/phi/), refRa, refDec, raVec, decVec)

    ra  = raVec(1)
    dec = decVec(1)

  end subroutine polarsToRaDecScalar

  !---------------------------------------------------------------------
  subroutine polarsToRaDecVector(theta, phi, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image-plane coordinate
  ! system into ra and dec. Here the image-plane coordinates are given
  ! in polar form: theta is the radial angle whereas phi is the azimuth.
  ! For a direction of view that is outwards, with declination increasing
  ! upward, the phi=0 axis of the tangent plane points to the right,
  ! and phi increases anticlockwise.

    implicit none

    real(double), intent(in)            :: refRa,&
                                           refDec,&
                                           theta(:),&
                                           phi(size(theta))
    real(double), intent(out)           :: ra(size(theta)),&
                                           dec(size(theta))

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3),&
      semiAxis
    integer(int32) :: numPositions, i

    !-------------------------------------------------------------------
    ! Start

    numPositions = size(theta)
    refMatrix_sky = calcRefMatrixInSkyBasis(refRa, refDec)

    do i = 1, numPositions
      srcVector_ref(1) = cos(theta(i))
      semiAxis = sin(theta(i))
      srcVector_ref(2) = -semiAxis * cos(phi(i))
      srcVector_ref(3) =  semiAxis * sin(phi(i))

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky, ra(i), dec(i))
    end do

  end subroutine polarsToRaDecVector

  !---------------------------------------------------------------------
  subroutine polarsToRaDecArray(theta, phi, refRa, refDec, ra, dec)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from the image-plane coordinate
  ! system into ra and dec. Here the image-plane coordinates are given
  ! in polar form: theta is the radial angle whereas phi is the azimuth.
  ! For a direction of view that is outwards, with declination increasing
  ! upward, the phi=0 axis of the tangent plane points to the right,
  ! and phi increases anticlockwise.

    implicit none

    real(double), intent(in)            :: refRa,&
                                           refDec,&
                                           theta(:,:),&
                                           phi(size(theta,1),size(theta,2))
    real(double), intent(out)           :: ra(size(theta,1),size(theta,2)),&
                                           dec(size(theta,1),size(theta,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable :: thetaVec(:), phiVec(:)
    real(double), allocatable :: raVec(:), decVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(theta, 1)
    ySize = size(theta, 2)
    vSize = xSize * ySize

    allocate(thetaVec(vSize), phiVec(vSize))
    allocate(raVec(vSize), decVec(vSize))
    thetaVec = reshape(theta, (/vSize/))
    phiVec   = reshape(phi,   (/vSize/))

    call polarsToRaDecVector(thetaVec, phiVec, refRa, refDec, raVec, decVec)

    ra  = reshape(raVec,  (/xSize, ySize/))
    dec = reshape(decVec, (/xSize, ySize/))

  end subroutine polarsToRaDecArray


  !---------------------------------------------------------------------
  subroutine telToDetScalar(tanThetaArcsec, phi, xDet, yDet)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in) :: phi, tanThetaArcsec
    real(double), intent(out) :: xDet, yDet

    real(double) ::&
!      tanThetaArcsecVec(1), phiVec(1),&
      mmToDetUnits
    real(double), pointer ::&
      cc2X(:), cc2Y(:), cc2Z(:)

    !-------------------------------------------------------------------
    ! Start

!    tanThetaArcsecVec(1) = tanThetaArcsec
!    phiVec(1)            = phi
!    call CAL_telCoordToCamCoord2(tanThetaArcsecVec, phiVec, cc2X, cc2Y, cc2Z)
    call CAL_telCoordToCamCoord2((/tanThetaArcsec/), (/phi/), cc2X, cc2Y, cc2Z)
    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X(0) * mmToDetUnits
    yDet = cc2Y(0) * mmToDetUnits
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine telToDetScalar

  !---------------------------------------------------------------------
  subroutine telToDetVector(tanThetaArcsec, phi, xDet, yDet)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in)  :: phi(:),&
                                 tanThetaArcsec(size(phi))
    real(double), intent(out) :: xDet(size(phi)),&
                                 yDet(size(phi))

    real(double), pointer :: cc2X(:), cc2Y(:), cc2Z(:)
    real(double) :: mmToDetUnits

    !-------------------------------------------------------------------
    ! Start

    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)
    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine telToDetVector

  !---------------------------------------------------------------------
  subroutine telToDetArray(tanThetaArcsec, phi, xDet, yDet)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in)  :: phi(:,:),&
                                 tanThetaArcsec(size(phi,1),size(phi,2))
    real(double), intent(out) :: xDet(size(phi,1),size(phi,2)),&
                                 yDet(size(phi,1),size(phi,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable ::&
      phiVec(:), tanThetaArcsecVec(:)
    real(double), pointer ::&
      cc2X(:), cc2Y(:), cc2Z(:)
    real(double) :: mmToDetUnits

    !-------------------------------------------------------------------
    ! Start

    xSize = size(phi, 1)
    ySize = size(phi, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi, (/vSize/))

    call CAL_telCoordToCamCoord2(tanThetaArcsecVec, phiVec, cc2X, cc2Y, cc2Z)
    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = reshape(cc2X * mmToDetUnits, (/xSize, ySize/))
    yDet = reshape(cc2Y * mmToDetUnits, (/xSize, ySize/))
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine telToDetArray


  !---------------------------------------------------------------------
  subroutine detToTelScalar(xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in) :: xDet, yDet
    real(double), intent(out) :: phi, tanThetaArcsec

    real(double) ::&
      cc2X(1), cc2Y(1), cc2Z(1), mmToDetUnits
    real(double), pointer ::&
      tanThetaArcsecPtr(:), phiPtr(:)

    !-------------------------------------------------------------------
    ! Start

    mmToDetUnits = calcMmToDetConversionFactor()

    cc2X(1) = xDet / mmToDetUnits
    cc2Y(1) = yDet / mmToDetUnits
    cc2Z(1) = 0.0d0
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, tanThetaArcsecPtr, phiPtr)
    tanThetaArcsec = tanThetaArcsecPtr(0)
    phi            = phiPtr(0)
    call CAL_releaseMemory(tanThetaArcsecPtr, phiPtr)

  end subroutine detToTelScalar

  !---------------------------------------------------------------------
  subroutine detToTelVector(xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in) :: xDet(:), yDet(size(xDet))
    real(double), intent(out) :: phi(size(xDet)), tanThetaArcsec(size(xDet))

    real(double) :: mmToDetUnits
    real(double), allocatable :: cc2X(:), cc2Y(:), cc2Z(:)
    real(double), pointer ::&
      tanThetaArcsecPtr(:), phiPtr(:)
    integer :: vSize

    !-------------------------------------------------------------------
    ! Start

    vSize = size(xDet)
    allocate(cc2X(size(xDet)), cc2Y(size(xDet)), cc2Z(size(xDet)))

    mmToDetUnits = calcMmToDetConversionFactor()

    cc2X = xDet / mmToDetUnits
    cc2Y = yDet / mmToDetUnits
    cc2Z = 0.0d0
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, tanThetaArcsecPtr, phiPtr)
    tanThetaArcsec = tanThetaArcsecPtr
    phi            = phiPtr
    call CAL_releaseMemory(tanThetaArcsecPtr, phiPtr)

  end subroutine detToTelVector

  !---------------------------------------------------------------------
  subroutine detToTelArray(xDet, yDet, tanThetaArcsec, phi)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(in)  :: xDet(:,:),&
                                 yDet(size(xDet,1),size(xDet,2))
    real(double), intent(out) :: phi(size(xDet,1),size(xDet,2)),&
                                 tanThetaArcsec(size(xDet,1),size(xDet,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable ::&
      cc2X(:), cc2Y(:), cc2Z(:)
    real(double) :: mmToDetUnits
    real(double), pointer ::&
      tanThetaArcsecPtr(:), phiPtr(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xDet, 1)
    ySize = size(xDet, 2)
    vSize = xSize * ySize

    allocate(cc2X(vSize), cc2Y(vSize), cc2Z(vSize))

    mmToDetUnits = calcMmToDetConversionFactor()

    cc2X = reshape(xDet / mmToDetUnits, (/vSize/))
    cc2Y = reshape(yDet / mmToDetUnits, (/vSize/))
    cc2Z = 0.0d0
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, tanThetaArcsecPtr, phiPtr)
    tanThetaArcsec = reshape(tanThetaArcsecPtr, (/xSize, ySize/))
    phi            = reshape(phiPtr,            (/xSize, ySize/))
    call CAL_releaseMemory(tanThetaArcsecPtr, phiPtr)

  end subroutine detToTelArray


  !---------------------------------------------------------------------
  subroutine getPsfImagePixelCorners(psfCentreTanThetaArcsec, psfCentrePhi&
  , psfPixelSizeMm, psfImage, psfImagePixelCorners)
  !---------------------------------------------------------------------
  ! This routine returns an array of type Point2dT which records the
  ! positions in the polar TELCOORD system of the corners of the input
  ! PSF image array pixels. Eg if the input array is a 3x3 array as follows:
  !
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,1  |  2,1  |  3,1  |
  !	 |       |       |       |
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,2  |  2,2  |  3,2  |
  !	 |       |       |       |
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,3  |  2,3  |  3,3  |
  !	 |       |       |       |
  !	  -----------------------
  !
  ! then the 4x4 elements of the output array correspond as follows:
  !
  !	1,1-----2,1-----3,1-----4,1
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,2-----2,2-----3,2-----4,2
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,3-----2,3-----3,3-----4,3
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,4-----2,4-----3,4-----4,4
  !
  ! The values of the returned array are tangent values in the TELCOORD
  ! system and thus have no units.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !---------------------------------------------------------------------
  ! The PSF array returned by the cal _appears_ (who knows for sure!) to be oriented such that the first index of the array increases with _decrease_ in TELCOORD X coordinate, and the second index increases with _increase_ in the TELCOORD Y coordinate. TELCOORD X and Y coordinates for each of the XMM EPIC instruments are here defined as that set of orthogonal axes in a projection plane normal to the optical axis of the telescope such that
  !
  !	(TELCOORD X) = (TELCOORD theta) * cos(TELCOORD phi)
  ! and
  !	(TELCOORD Y) = (TELCOORD theta) * sin(TELCOORD phi).
  !
  ! (Note that 'theta' as returned by cal calls appears to actually be the tangent of the angle theta which is described in the cal docs. Hence the lack of a tangent function in the above equations.)
  !---------------------------------------------------------------------

    implicit none

    real(double),       intent(in)  :: psfCentreTanThetaArcsec, psfCentrePhi
    type(PsfBinSizeT),  intent(in)  :: psfPixelSizeMm
    real(single),       intent(in)  :: psfImage(:,:)
    type(Point2dDbleT), intent(out) :: psfImagePixelCorners(&
                                         size(psfImage,1)+1&
                                       , size(psfImage,2)+1)
    ! type defined in geometric_types.f90.

    integer(int16) :: xSize, ySize, xi, yi
    real(double) ::&
      arrayCentreX, arrayCentreY, psfCentreRadiusTel, psfCentreXTel,&
      psfCentreYTel, focalLengthMm
    type(Point2dDbleT) :: psfPixelSizeTel

    !-------------------------------------------------------------------
    ! Start

    xSize = size(psfImage, 1)
    ySize = size(psfImage, 2)
!*** error/warning if not odd??

    arrayCentreX = xSize / 2.0
    arrayCentreY = ySize / 2.0

!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
focalLengthMm = ims_getFocalLength()
!**** need to get a different FL correction for each inst until SOC SPR 2518 is solved.
    psfPixelSizeTel%x = psfPixelSizeMm%binX / focalLengthMm
    psfPixelSizeTel%y = psfPixelSizeMm%binY / focalLengthMm

!    psfCentreRadiusTel = tan(psfCentreThetaArcsec / 3600.0d0 * PiOn180)
psfCentreRadiusTel = psfCentreTanThetaArcsec / 3600.0d0 * PiOn180
    psfCentreXTel = psfCentreRadiusTel * cos(psfCentrePhi)
    psfCentreYTel = psfCentreRadiusTel * sin(psfCentrePhi)

    do xi = 1, xSize + 1
!      psfImagePixelCorners(xi,:)%x = psfCentreXTel + (xi - 1.0 - arrayCentreX)&
psfImagePixelCorners(xi,:)%x = psfCentreXTel - (xi - 1.0 - arrayCentreX)&
!*** change made 20040524 IMS
        * psfPixelSizeTel%x
    end do

    do yi = 1, ySize + 1
      psfImagePixelCorners(:,yi)%y = psfCentreYTel + (yi - 1.0 - arrayCentreY)&
        * psfPixelSizeTel%y
    end do

  end subroutine getPsfImagePixelCorners

  !---------------------------------------------------------------------
  subroutine getPsfPixelCnrsThetaPhi(psfCentreTanThetaArcsec, psfCentrePhi&
  , psfPixelSizeMm, psfImage, psfPixelCnrsTanThetaArcsec, psfPixelCnrsPhi)
  !---------------------------------------------------------------------
  ! This routine returns an array of type Point2dT which records the
  ! positions in the polar TELCOORD system of the corners of the input
  ! PSF image array pixels. Eg if the input array is a 3x3 array as follows:
  !
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,1  |  2,1  |  3,1  |
  !	 |       |       |       |
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,2  |  2,2  |  3,2  |
  !	 |       |       |       |
  !	  -----------------------
  !	 |       |       |       |
  !	 |  1,3  |  2,3  |  3,3  |
  !	 |       |       |       |
  !	  -----------------------
  !
  ! then the 4x4 elements of the output array correspond as follows:
  !
  !	1,1-----2,1-----3,1-----4,1
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,2-----2,2-----3,2-----4,2
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,3-----2,3-----3,3-----4,3
  !	 |       |       |       |
  !	 |       |       |       |
  !	 |       |       |       |
  !	1,4-----2,4-----3,4-----4,4
  !
  ! The values of the returned array are tangent values in the TELCOORD
  ! system and thus have no units.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: psfCentreTanThetaArcsec, psfCentrePhi
    type(PsfBinSizeT), intent(in)  :: psfPixelSizeMm
    real(single),      intent(in)  :: psfImage(:,:)
    real(double),      intent(out) :: psfPixelCnrsTanThetaArcsec(&
                                        size(psfImage,1)+1,&
                                        size(psfImage,2)+1&
                                      ),&
                                      psfPixelCnrsPhi(&
                                        size(psfImage,1)+1,&
                                        size(psfImage,2)+1&
                                      )

    integer(int16) :: xSize, ySize, xi, yi
    real(double) ::&
      arrayCentreX, arrayCentreY, psfCentreRadiusTel, psfCentreXTel,&
      psfCentreYTel, focalLengthMm
    type(Point2dDbleT) :: psfPixelSizeTel
    ! type Point2dDbleT is defined in geometric_types.f90.
    real(double), allocatable :: tanXArray(:,:), tanYArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(psfImage, 1)
    ySize = size(psfImage, 2)
!*** error/warning if not odd??

    allocate(tanXArray(xSize+1,ySize+1), tanYArray(xSize+1,ySize+1))

    arrayCentreX = xSize / 2.0
    arrayCentreY = ySize / 2.0

!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
focalLengthMm = ims_getFocalLength()
!**** need to get a different FL correction for each inst until SOC SPR 2518 is solved.
    psfPixelSizeTel%x = psfPixelSizeMm%binX / focalLengthMm
    psfPixelSizeTel%y = psfPixelSizeMm%binY / focalLengthMm

!    psfCentreRadiusTel = tan(psfCentreThetaArcsec / 3600.0d0 * PiOn180)
psfCentreRadiusTel = psfCentreTanThetaArcsec / 3600.0d0 * PiOn180
    psfCentreXTel = psfCentreRadiusTel * cos(psfCentrePhi)
    psfCentreYTel = psfCentreRadiusTel * sin(psfCentrePhi)

    do xi = 1, xSize + 1
!      tanXArray(xi,:) = psfCentreXTel + (xi - 1.0 - arrayCentreX)&
tanXArray(xi,:) = psfCentreXTel - (xi - 1.0 - arrayCentreX)&
!*** change made 20040524 IMS
        * psfPixelSizeTel%x
    end do

    do yi = 1, ySize + 1
      tanYArray(:,yi) = psfCentreYTel + (yi - 1.0 - arrayCentreY)&
        * psfPixelSizeTel%y
    end do

!    psfPixelCnrsThetaArcsec = atan(sqrt(tanXArray * tanXArray&
!      + tanYArray * tanYArray)) * 3600.0d0 / PiOn180
psfPixelCnrsTanThetaArcsec = sqrt(tanXArray * tanXArray&
  + tanYArray * tanYArray) * 3600.0d0 / PiOn180

    psfPixelCnrsPhi = atan2(tanYArray, tanXArray)

  end subroutine getPsfPixelCnrsThetaPhi

  !---------------------------------------------------------------------
  subroutine loadXYindices(xSize, ySize, rawX, rawY)
  !---------------------------------------------------------------------
  ! This is sort of a 2-dimensional version of indicizeInt16(). It is
  ! designed to work in coordination with the cal routines which convert
  ! RAWXY (ie CHIPCOORD) indices to other coordinate systems. The returned
  ! arrays rawX() and rawY() contain all the pixel-center coordinates for
  ! a 2-dimensional array of size defined by the input parameters xSize
  ! and ySize. Note however that rawX() and rawY() are 1-dimensional. This
  ! is because CAL_chipCoordToCamCoord2() etc take only 1-dimensional
  ! arguments. If, as here, we want to process an entire CCD in one call,
  ! then the 2-dimensional array of chip coordinates must either be split
  ! or (as here) unpacked into 1 or more 1-dimensional arrays.
  !
  ! An example of what goes on here: say xSize=4 and ySize=5 (this would
  ! be a small CCD however!): the RAW or CHIPCOORD values for the centre
  ! of each pixel of an array of these dimensions can be displayed as
  ! follows:
  !
  !	RAWX:
  !		   2nd index
  !	1 2 3 4		^
  !	1 2 3 4		|
  !	1 2 3 4		|
  !	1 2 3 4		|
  !	1 2 3 4		 --------> 1st index
  !
  !	RAWY:
  !		   2nd index
  !	5 5 5 5		^
  !	4 4 4 4		|
  !	3 3 3 3		|
  !	2 2 2 2		|
  !	1 1 1 1		 --------> 1st index
  !
  ! The values returned in rawX() and rawY() would in this case be:
  !
  ! rawX = (1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4)
  ! rawY = (1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5)
  !
  ! NOTE that rawX() and rawY() should be deallocated after use.

    implicit none

    integer(int16), intent(in) :: xSize, ySize
    integer(int16), pointer    :: rawX(:), rawY(:)

    integer(int32) ::&
      vectorShape(1)
    integer(int16) :: rawXVector(xSize), rawYVector(ySize), i

    !-------------------------------------------------------------------
    ! Start

    if (xSize <= 0 .or. ySize <= 0)&
call error('chipArrayTooSmall',"ssclib:loadXYindices: The routine cannot return a zero-size array.")

    vectorShape(1) = xSize * ySize
    allocate(rawX(vectorShape(1)), rawY(vectorShape(1)))
    rawXVector = (/(i, i=1,xSize)/)
    rawYVector = (/(i, i=1,ySize)/)
    rawX = reshape(spread(rawXVector, 2, xSize), vectorShape)
    rawY = reshape(spread(rawYVector, 1, ySize), vectorShape)

  end subroutine loadXYindices


  !---------------------------------------------------------------------
  subroutine getThetaPhiMaps(wcs, scAttitude, timeStamp, thetaMap, phiMap)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up for the correct instrument before this
  ! is called.

    implicit none

    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: thetaMap(:,:),&
                                                 phiMap(size(thetaMap,1)&
                                                      , size(thetaMap,2))

    integer(int16) :: xSize, ySize, i
    integer(int16), allocatable :: xPixels(:), yPixels(:)
    real(double), allocatable :: tanX(:), tanY(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(thetaMap, 1)
    ySize = size(thetaMap, 2)
    allocate(xPixels(xSize), yPixels(ySize))
    allocate(tanX(xSize), tanY(ySize))

    xPixels = (/(i, i=1,xSize)/)
    yPixels = (/(i, i=1,ySize)/)
    call pixelsToTan(real(xPixels, single), real(yPixels, single), wcs&
      , tanX, tanY) ! in ssclib/wcs_aux

!    call tanToInst(spread(tanX, 2, ySize), spread(tanY, 1, xSize)&
!      , wcs%x%refWorld, wcs%y%refWorld, scAttitude&
!      , timeStamp, thetaArcsec=thetaMap, phi=phiMap) ! in ssclib/coordinate

call tanToTel(spread(tanX, 2, ySize), spread(tanY, 1, xSize)&
  , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180, scAttitude&
  , timeStamp, thetaMap, phiMap)

  end subroutine getThetaPhiMaps

  !---------------------------------------------------------------------
  function skyToCartesian(ra, dec) result(vector)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: ra, dec 
    real(double)             :: vector(3)

    real (double) :: cosDec 

    !-------------------------------------------------------------------
    ! Start

    cosDec = cos(dec)
    vector(1) = cos(ra) * cosDec
    vector(2) = sin(ra) * cosDec
    vector(3) = sin(dec)

  end function skyToCartesian

  !---------------------------------------------------------------------
  subroutine cartesianToSky(vector, ra, dec)
  !---------------------------------------------------------------------
  ! Note that this doesn't require the input vector to be of length 1.

    implicit none

    real(double), intent(in)  :: vector(3) 
    real(double), intent(out) :: ra, dec

    !-------------------------------------------------------------------
    ! Start

    if (vector(1) .ne. 0.0 .and. vector(2) .ne. 0.0) then
      ra = atan2(vector(2), vector(1))
      do while(ra < 0.0)
        ra = ra + TwoPi
      end do
    else
      ra = 0.0d0
    end if
    dec = atan2(vector(3), sqrt(vector(1)*vector(1)+vector(2)*vector(2)))

  end subroutine cartesianToSky

  !---------------------------------------------------------------------
  subroutine psfToDetScalar(xTel, yTel, xDet, yDet)
  !---------------------------------------------------------------------
  ! The input arrays contains (x,y) coordinates in a cartesianized TELCOORD
  ! system - ie the x values are tan(theta) * cos(phi) whereas the y values
  ! are tan(theta) * sin(phi). The purpose of the present subroutine is to
  ! transform these positions to the DETXY system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTel,&
                                 yTel
    real(double), intent(out) :: xDet,&
                                 yDet

    real(double) :: xDetVec(1), yDetVec(1)

    !-------------------------------------------------------------------
    ! Start

    call psfToDetVector((/xTel/), (/yTel/), xDetVec, yDetVec)
    xDet = xDetVec(1)
    yDet = yDetVec(1)

  end subroutine psfToDetScalar

  !---------------------------------------------------------------------
  subroutine psfToDetVector(xTel, yTel, xDet, yDet)
  !---------------------------------------------------------------------
  ! The input arrays contains (x,y) coordinates in a cartesianized TELCOORD
  ! system - ie the x values are tan(theta) * cos(phi) whereas the y values
  ! are tan(theta) * sin(phi). The purpose of the present subroutine is to
  ! transform these positions to the DETXY system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTel(:),&
                                 yTel(size(xTel))
    real(double), intent(out) :: xDet(size(xTel)),&
                                 yDet(size(xTel))

    real(double) :: mmToDetUnits
    real(double), allocatable :: tanThetaArcsecVec(:), phiVec(:)
    real(double), pointer :: cc2X(:), cc2Y(:), cc2Z(:)
    integer :: vSize

    !-------------------------------------------------------------------
    ! Start

    vSize = size(xTel)
    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))

    ! Convert the TELCOORD positions from cartesian to polar:
!    tanThetaArcsecVec = atan(sqrt(telXVector * telXVector&
!      + telYVector * telYVector)) * 3600.0d0 / PiOn180
tanThetaArcsecVec = sqrt(xTel * xTel + yTel * yTel) * 3600.0d0 / PiOn180

    phiVec = atan2(yTel, xTel)

    ! Convert to CAMCOORD2, then to DETXY:
    call CAL_telCoordToCamCoord2(tanThetaArcsecVec, phiVec, cc2X, cc2Y, cc2Z)

    mmToDetUnits = calcMmToDetConversionFactor()
    xDet = cc2X * mmToDetUnits
    yDet = cc2Y * mmToDetUnits

    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

  end subroutine psfToDetVector

  !---------------------------------------------------------------------
  subroutine psfToDetArray(xTel, yTel, xDet, yDet)
  !---------------------------------------------------------------------
  ! The input arrays contains (x,y) coordinates in a cartesianized TELCOORD
  ! system - ie the x values are tan(theta) * cos(phi) whereas the y values
  ! are tan(theta) * sin(phi). The purpose of the present subroutine is to
  ! transform these positions to the DETXY system.
  !
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: xTel(:,:),&
                                 yTel(size(xTel,1),size(xTel,2))
    real(double), intent(out) :: xDet(size(xTel,1),size(xTel,2)),&
                                 yDet(size(xTel,1),size(xTel,2))

    integer :: xSize, ySize, vSize
    real(double), allocatable :: xTelVec(:), yTelVec(:), xDetVec(:), yDetVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xTel, 1)
    ySize = size(xTel, 2)
    vSize = xSize * ySize
    allocate(xTelVec(vSize), yTelVec(vSize), xDetVec(vSize), yDetVec(vSize))

    xTelVec = reshape(xTel, (/vSize/))
    yTelVec = reshape(yTel, (/vSize/))
    call psfToDetVector(xTelVec, yTelVec, xDetVec, yDetVec)
    xDet = reshape(xDetVec, (/xSize, ySize/))
    yDet = reshape(yDetVec, (/xSize, ySize/))

  end subroutine psfToDetArray


  !---------------------------------------------------------------------
  subroutine psfToTanArray(psfCtrThetaArcsec, psfCtrPhi, psfPixelSizeMm&
  , psfImageXSize, psfImageYSize, refRa, refDec&
  , instrumentId, timeStamp, scAttitude, tanXArray, tanYArray)
  !---------------------------------------------------------------------
  ! The PSF array returned by CAL_getPsfImage is in the TELCOORD system,
  ! which is aligned with the DETXY/CAMCOORD2 coordinate system, EXCEPT...
  ! for a mirror reflection and N x 90deg rotations, N being different for
  ! MOS1, MOS2 and PN. Hence the need for this special routine.
  !---------------------------------------------------------------------

!**** obsolescent. Should be replaced by
!****   call getPsfImagePixelCorners
!****   call psfToDet
!****   call instToTan

    implicit none

    integer(int16),               intent(in)  :: psfImageXSize,&
                                                 psfImageYSize
    type(PsfBinSizeT),            intent(in)  :: psfPixelSizeMm
    integer(int32),               intent(in)  :: instrumentId
    real(double),                 intent(in)  :: psfCtrThetaArcsec,&
                                                 psfCtrPhi,&
                                                 refRa,&
                                                 refDec,&
                                                 timeStamp
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(out) :: tanXArray(&
                                                   psfImageXSize,&
                                                   psfImageYSize&
                                                 ),&
                                                 tanYArray(&
                                                   psfImageXSize,&
                                                   psfImageYSize&
                                                 )

    real(double) :: thetaVec(1), phiVec(1), mmToDetUnits
    real(double), pointer :: psfCtrXmm(:), psfCtrYmm(:), dummyZmm(:)
    real(double) :: psfCtrDetX, psfCtrDetY
    integer(int16) :: xIndices(psfImageXSize), yIndices(psfImageYSize), i
    real(double), allocatable :: detXVec(:), detYVec(:)

    !-------------------------------------------------------------------
    ! Start

    ! Get the centre of the PSF array in CAMCOORD2:

    thetaVec(1) = psfCtrThetaArcsec
    phiVec(1)   = psfCtrPhi
    call CAL_telCoordToCamCoord2(thetaVec, phiVec, psfCtrXmm, psfCtrYmm&
      , dummyZmm)

    mmToDetUnits = calcMmToDetConversionFactor()
    psfCtrDetX = psfCtrXmm(0) * mmToDetUnits
    psfCtrDetY = psfCtrYmm(0) * mmToDetUnits

    call CAL_releaseMemory(psfCtrXmm, psfCtrYmm, dummyZmm)

    ! There is no need to deal in arrays
    ! until the final conversion from det coord to sky tangent coords,
    ! because the psf image is aligned with the detx/y coordinate axes.
    ! Hence just work with x-axis and y-axis vectors until that point.

    xIndices = (/(i, i = 1, psfImageXSize)/)
    yIndices = (/(i, i = 1, psfImageYSize)/)

    select case(instrumentId)
    case(EMOS1)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			y <-----
    !	  |			        |
    !	  |			        |
    !	   -----> x		        v x
    !
      allocate(detXVec(psfImageYSize), detYVec(psfImageXSize))
      detXVec = -(yIndices - psfImageYSize / 2.0) * psfPixelSizeMm%binY
      detYVec = -(xIndices - psfImageXSize / 2.0) * psfPixelSizeMm%binX

    case(EMOS2)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			   -----> x
    !	  |			  |
    !	  |			  |
    !	   -----> x		y v
    !
      allocate(detXVec(psfImageXSize), detYVec(psfImageYSize))
      detXVec =  (xIndices - psfImageXSize / 2.0) * psfPixelSizeMm%binX
      detYVec = -(yIndices - psfImageYSize / 2.0) * psfPixelSizeMm%binY

    case(EPN)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			        ^ y
    !	  |			        |
    !	  |			        |
    !	   -----> x		x <-----
    !
      allocate(detXVec(psfImageXSize), detYVec(psfImageYSize))
      detXVec = -(xIndices - psfImageXSize / 2.0) * psfPixelSizeMm%binX
      detYVec =  (yIndices - psfImageYSize / 2.0) * psfPixelSizeMm%binY

    case default
call error('badInstrument',"ssclib:psfToTanArray: Only EPIC instruments supported.")
    end select

    detXVec = detXVec * mmToDetUnits + psfCtrDetX
    detYVec = detYVec * mmToDetUnits + psfCtrDetY

    ! Now we have to go from vectors to arrays.

    if (instrumentId .eq. EMOS1) then
!      call instToTan(tanXArray, tanYArray, refRaDeg, refDecDeg, scAttitude&
!        , timeStamp, spread(detXVec, 1, psfImageXSize)&
!        , spread(detYVec, 2, psfImageYSize))
      call detToTan(spread(detXVec, 1, psfImageXSize)&
        , spread(detYVec, 2, psfImageYSize), refRa, refDec, scAttitude&
        , timeStamp, tanXArray, tanYArray)

    else
!      call instToTan(tanXArray, tanYArray, refRaDeg, refDecDeg, scAttitude&
!        , timeStamp, spread(detXVec, 2, psfImageYSize)&
!        , spread(detYVec, 1, psfImageXSize))
      call detToTan(spread(detXVec, 2, psfImageYSize)&
        , spread(detYVec, 1, psfImageXSize), refRa, refDec, scAttitude&
        , timeStamp, tanXArray, tanYArray)
    end if

  end subroutine psfToTanArray

  !---------------------------------------------------------------------
  subroutine tanToPsfArray(tanXArray, tanYArray, psfCtrThetaArcsec, psfCtrPhi&
  , psfPixelSizeMm, refRa, refDec, instrumentId, timeStamp, scAttitude&
  , psfXArray, psfYArray)
  !---------------------------------------------------------------------
  ! The PSF array returned by CAL_getPsfImage is in the TELCOORD system,
  ! which is aligned with the DETXY/CAMCOORD2 coordinate system, EXCEPT...
  ! for a mirror reflection and N x 90deg rotations, N being different for
  ! MOS1, MOS2 and PN. Hence the need for this special routine.
  !---------------------------------------------------------------------

    implicit none

    real(double),                 intent(in)  :: tanXArray(:,:),&
                                                 tanYArray(&
                                                   size(tanXArray,1),&
                                                   size(tanXArray,2)&
                                                 )
    type(PsfBinSizeT),            intent(in)  :: psfPixelSizeMm
    integer(int32),               intent(in)  :: instrumentId
    real(double),                 intent(in)  :: psfCtrThetaArcsec,&
                                                 psfCtrPhi,&
                                                 refRa,&
                                                 refDec,&
                                                 timeStamp
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(out) :: psfXArray(&
                                                   size(tanXArray,1),&
                                                   size(tanXArray,2)&
                                                 ),&
                                                 psfYArray(&
                                                   size(tanXArray,1),&
                                                   size(tanXArray,2)&
                                                 )

    integer(int16) :: xSize, ySize
    real(double) ::&
      thetaVec(1), phiVec(1), mmToDetUnits, psfCtrDetX, psfCtrDetY
    real(double), pointer :: psfCtrXmm(:), psfCtrYmm(:), dummyZmm(:)
    real(double), allocatable ::  detXArray(:,:), detYArray(:,:)

    !-------------------------------------------------------------------
    ! Start

!**** obsolescent.

    xSize = size(tanXArray,1)
    ySize = size(tanXArray,2)
    allocate(detXArray(xSize,ySize), detYArray(xSize,ySize))

    ! Get the centre of the PSF array in CAMCOORD2:

    thetaVec(1) = psfCtrThetaArcsec
    phiVec(1)   = psfCtrPhi
    call CAL_telCoordToCamCoord2(thetaVec, phiVec, psfCtrXmm, psfCtrYmm&
      , dummyZmm)

    mmToDetUnits = calcMmToDetConversionFactor()
    psfCtrDetX = psfCtrXmm(0) * mmToDetUnits
    psfCtrDetY = psfCtrYmm(0) * mmToDetUnits

    call CAL_releaseMemory(psfCtrXmm, psfCtrYmm, dummyZmm)

    ! Convert the array of tan locations into DETXY:

!    call tanToInst(tanXArray, tanYArray, refRaDeg, refDecDeg, scAttitude&
!      , timeStamp, detXArray, detYArray)
    call tanToDet(tanXArray, tanYArray, refRa, refDec, scAttitude&
      , timeStamp, detXArray, detYArray)

    ! Now subtract the DETXY position of the PSF centre and then rotate
    ! and rescale:

    detXArray = (detXArray - psfCtrDetX) / mmToDetUnits
    detYArray = (detYArray - psfCtrDetY) / mmToDetUnits
    ! These arrays are now in mm and are centred on the PSF centre.

    select case(instrumentId)
    case(EMOS1)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			y <-----
    !	  |			        |
    !	  |			        |
    !	   -----> x		        v x
    !
      psfXArray = -detYArray / psfPixelSizeMm%binX
      psfYArray = -detXArray / psfPixelSizeMm%binX

    case(EMOS2)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			   -----> x
    !	  |			  |
    !	  |			  |
    !	   -----> x		y v
    !
      psfXArray =  detXArray / psfPixelSizeMm%binX
      psfYArray = -detYArray / psfPixelSizeMm%binX

    case(EPN)
    !
    !	    PSF:		  DETXY:
    !
    !	y ^			        ^ y
    !	  |			        |
    !	  |			        |
    !	   -----> x		x <-----
    !
      psfXArray = -detXArray / psfPixelSizeMm%binX
      psfYArray =  detYArray / psfPixelSizeMm%binX

    case default
call error('badInstrument',"ssclib:tanToPsfArray: Only EPIC instruments supported.")
    end select

  end subroutine tanToPsfArray

  !---------------------------------------------------------------------
  subroutine calcAitoff(raDeg, decDeg, x, y)
  !---------------------------------------------------------------------
!*** Could make the aitoff calc quicker by precalculating sin and cos dec etc.
!*** Taken from swcat.
!*** Include also drawAitoffGrid?

    implicit none 

    real(single), intent(in) :: raDeg, decDeg
    real(single), intent(out) :: x, y
    real(single) :: ra, dec, r1, denom, cosDec

    !-------------------------------------------------------------------
    ! Start:

    ra  = raDeg  * PiOn180
    dec = decDeg * PiOn180

    r1 = (ra - Pi) / 2.0
    cosDec = cos(dec)
    denom = sqrt(1.0 + cosDec * cos(r1))
    x = 2.0 * cosDec * sin(r1) / denom
    y = sin(dec) / denom

  end subroutine calcAitoff

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module coordinate
